unit Core;

interface
uses
        FMX.Types,
        MCMT32_ASCOM_TLB,
        HiResTim,ShlObj;
Function IsMCMTSlewing(Axis: integer): boolean;
function IsAxisSlewing_Backlash(Axis: integer): boolean;
function MoveOffsetDirectStep(AlphaDSteps, DeltaDSteps: Integer): Boolean;
Function RetrieveDataAlphaDelta: boolean;
Function InitTelescopeSeeker(var Alpha, Delta, Latitude, Longitude: Double): Boolean;
function STOPTelescope(Alpha, Delta: Boolean): Boolean;
function STOP_ALL_Telescope: Boolean;
procedure TelescopeMoveAxe(MoveSteps, Axe: integer);
Function  TelescopeMove(MoveStepsAlpha, MoveStepsDelta: integer): Boolean;
Procedure CalculePositionMonture;
function TelescopeOuest : boolean;
function ReturnToSideralSpeed: Boolean;
function ParkTel_at(AHPark, Dec2000: Double): Boolean;
procedure ASK_FortelescopeOrientation(Latitude:Double);
Procedure CreateAllTimers;
Procedure DisableTimers;
Procedure FreeAllTimers;
function Calibrate(Alpha, Delta: Double): boolean; stdcall;
function Get_AtPark_Core: WordBool;
function Get_CanPark_Core: WordBool;
function Get_CanSetPark_Core: WordBool;
function Get_Connected_Core: WordBool;
function Get_Declination_Core: Double;
function Get_IsPulseGuiding_Core: WordBool;
function Get_RightAscension_Core: Double;
function Get_SiderealTime_Core: Double;
function Get_Slewing_Core: WordBool;
function Get_TargetDeclination_Core: Double;
function Get_TargetRightAscension_Core: Double;
function Get_Tracking_Core: WordBool;
procedure AbortSlew_Core;
procedure MoveAxis_Core(Axe: TelescopeAxes; Rate: Double);
procedure Park_Core;
procedure Set_Connected_Core(Value: WordBool);
procedure Set_TargetDeclination_Core(Value: Double);
procedure Set_TargetRightAscension_Core(Value: Double);
procedure SetPark_Core;
function SetFreeSpeed(Speed_rad:double;device:integer): Boolean;
function ReturnVersionDLL_str(Index:Integer): string;
procedure SetupDialog_Core;
procedure SlewToCoordinates_Core(RightAscension, Declination: Double);
procedure SlewToCoordinatesAsync_Core(RightAscension, Declination: Double);
procedure SlewToTarget_Core;
procedure SlewToTargetAsync_Core;
procedure SyncToCoordinates_Core(RightAscension, Declination: Double);
procedure SyncToTarget_Core;
procedure Unpark_Core;
procedure WriteToTrace(Data: string);

type
  TTimerBacklash = class(TTimer)
    procedure TimerEvent(Sender: TObject);
  private
    Flag: integer;
    Alpha, Delta: boolean;
    { Private declarations }
  public
    Finished: boolean;
    { Public declarations }
  end;

  TTimerPark = class(TTimer)
    procedure TimerEvent(Sender: TObject);
  private
    { Private declarations }
    Flag: integer;
    AngleHour, Delta: double;
  public
    Finished: boolean;
    { Public declarations }
  end;

  TTimerRetardateur = class(TTimer)
    procedure TimerEvent(Sender: TObject);
  private
    { Private declarations }
    Declanchement: double;
    DabordAlpha, DabordDelta : Boolean ;
    AlphaCroissant, DeltaCroissant : Boolean;
    MoveSteps : integer;
  public
    Finished: boolean;
    { Public declarations }
  end;

  TTimerPulse = class(THiResTimer)
    procedure TimerEvent(Sender: TObject);
  private
    StopOrderAlpha,
    StopOrderDecli: boolean;
    { Private declarations }
  public
    IsPulse: boolean;
    { Public declarations }
  end;



var
    fConnected,fParked,
    fTracking,
    fTargetDecOK, fTargetAscOK: boolean;
    fDeclination,fRightAscension,
    fTargetDeclination, fTargetRightAscension,
    fParkAH,fParkDelta,
    fDeclinationRate,fRightAscensionRate    :double;
    fApertureArea,fApertureDiameter :double;
      TimerAlphaPulse   : TTimerPulse;
  TimerDecliPulse   : TTimerPulse;

  TimerBacklash     : TTimerBacklash;
  TimerPark         : TTimerPark;
  TimerRetardateur  : TTimerRetardateur;
Const
  Stopping:Integer=0;
  _ERROR_            =-1;
  ErrorCountRAce:Word= 0;


implementation
Uses
      StdVcl,Windows,
      System.Win.Registry,
      System.Win.ComObj,
      Console,
      ComServ,
      COMM,
      IdGlobal,
      FeuRouge,
      FeuVert,
      scan_mcmt,
      ProcessBox_mcmtII,
{$IFDEF MCMT32}
      Setup32,
{$ELSE}
      SetupII,
{$ENDIF}
      Etat_Telescope,
      System.SysUtils,
      System.IniFiles,
      System.Classes,
      System.Variants,
      System.Math,
      Vcl.Dialogs,
      Vcl.Controls,
      Vcl.Forms,
      VarGlobal,
      Utils,
      MCMT_Main,
      Simulateur,
      Vcl.graphics;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Procedure DisableTimers;
begin
  Try
  If (TimerAlphaPulse  <> nil) then TimerAlphaPulse.Enabled :=False;
  if (TimerDecliPulse  <> nil) then TimerDecliPulse.Enabled :=False;
  if (TimerBacklash    <> nil) then TimerBacklash.Enabled   :=False;
  if (TimerPark        <> nil) then TimerPark.Enabled       :=False;
  if (TimerRetardateur <> nil) then TimerRetardateur.Enabled:=False;
  Except
  End;
end;
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Procedure TTimerPulse.TimerEvent(Sender: TObject);  // Arrive apres X ms
begin
  Self.Enabled := false;
  STOPTelescope(StopOrderAlpha, StopOrderDecli);
  IsPulse := false;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Procedure TTimerPark.TimerEvent(Sender: TObject);
var
  Alpha_Temp,
  Delta_Temp,
  Accuracy       : Double;

const
  procName='TTimerPark.TimerEvent';

  (* Main thread *)

begin
   // Une fois ce timer demarr pour le park, il n'est jamais stopp ! (!)
  if ((IsAxisSlewing_Backlash(C_Alpha) = false) and (IsAxisSlewing_Backlash(C_Delta) = false)) then
  begin
    case flag of
      1:
        begin
          Alpha_Temp:=CalculeTSL(Now+SpeedIntMCMT.GMT/24,Lieu.Long)-AngleHour;
          Delta_Temp := Delta;
          if BackLash.Saved_Park then
          begin
            Alpha_Temp := Alpha_Temp + 0.002; // To avoid backlash correction in the last seeker
            if BackLash.North then
            begin
              Delta_Temp := Delta_Temp + 0.002; // To avoid backlash correction in the last seeker
              if Delta_Temp > Pi_d_2 then Delta_Temp := Pi_d_2;
            end
            else
            begin
              Delta_Temp := Delta_Temp - 0.002; // To avoid backlash correction in the last seeker
              if Delta_Temp < -Pi_d_2 then Delta_Temp := -Pi_d_2;
            end;
          end;
          Remet2Pi(Alpha_Temp);
          InitTelescopeSeeker(Alpha_Temp, Delta_Temp, Lieu.Lat, Lieu.Long);
          Flag := 0;
        end;
      0:
        begin
          Alpha_Temp:=CalculeTSL(Now+SpeedIntMCMT.GMT/24,Lieu.Long)-AngleHour;
          Remet2Pi(Alpha_Temp);
          Delta_Temp := Delta;
          Enabled := false;
          InitTelescopeSeeker(Alpha_Temp, Delta_Temp, Lieu.Lat, Lieu.Long);
          while ((IsAxisSlewing_backlash(C_Alpha) = True) or (IsAxisSlewing_backlash(C_Delta) = True)) do WaitMainthread(procName);
          STOP_ALL_Telescope;
          is_Parking        := False;
          Park              := True;
          BackLash.Activated:= BackLash.Saved_Park;
      end;
    end;
  end;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Procedure TTimerRetardateur.TimerEvent(Sender: TObject);
begin

  Try
    CalculePositionMonture;

    if DabordAlpha
           {Le mouvement Alpha est dj dclanch, le mouvement Delta est retard}
       and ((GerMountSetup.Monture_Allemande
           {test d'attente du dclenchement Delta pour une monture allemande}
              and (DeltaCroissant xor (GerMountSetup.LongitudeHorairePointage<Declanchement))
              and (GerMountSetup.LongitudeHorairePointage< Pi_d_2+GerMountSetup.Angle_horaire_critique_Ouest)
              and (GerMountSetup.LongitudeHorairePointage>-Pi_d_2+GerMountSetup.Angle_horaire_critique_Est))
           or
           (not(GerMountSetup.Monture_Allemande)
           {test d'attente du dclanchement Delta pour une monture  fourche}
            and (AlphaCroissant xor (GerMountSetup.AngleHorairePointage<Declanchement))))

      then  {Dclenchement du mouvement Delta et arrt du timer}
        begin
          TelescopeMoveAxe(MoveSteps, C_Delta);
          Enabled:=false
        end;

    if DabordDelta
          { Le mouvement Delta est dj dclanch, le mouvement Alpha est retard
             Cela ne peut se produire que dans une procdure de retournement
             d'une monture allemande}
       and (DeltaCroissant xor (GerMountSetup.AnglePolairePointage<Declanchement))

      then  {Dclanchement du mouvement Alpha et arrt du timer}
        begin
          TelescopeMoveAxe(MoveSteps, C_Alpha);
          Enabled:=false
        end;

  Finally
  End;
end;

//////////////////////////////////////////////////////////////////////////
procedure WriteToTrace(Data: string);
var Heure:string;
begin
  if TraceEnabled then
    Begin
      Heure:=TimeToStr(Now);
      if Form2.CheckBoxScroll.Checked then
      begin
        Form2.MemoTrace.Lines.Add(Heure+' '+Data);
      end
      else
      begin
        Form2.MemoTrace.Lines.beginupdate;
        Form2.MemoTrace.Lines.Add(Heure+' '+Data);
        Form2.MemoTrace.Lines.endupdate;
      end;

    End;
end;




///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Stops completely the telescope
// If AbleTo_Stop_All is false, leave it as it is.

function STOP_ALL_Telescope: Boolean;
Const
  procName='STOP_ALL_Telescope';
var
  astring : ShortString;
begin
  WriteToTrace('Stop All');
  Result := False;
  while IsAxisSlewing_BackLash(C_Alpha) do WaitMainthread(procName);
  astring := 'P' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0); // Commande Park pour AD
  sendcommand(astring, 0, 0);
  while IsAxisSlewing_BackLash(C_Delta) do WaitMainthread(procName);
  astring := 'R' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0); // Commande vitesse = 0 en DEC;
  sendcommand(astring, 0, 1);
  Result := True;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Procedure TTimerBacklash.TimerEvent(Sender: TObject);
var
  MoveStepsAlpha,
  MoveStepsDelta : integer;
begin
  Try
    if flag <> 0 then
    begin
      dec(Flag);
      if Flag = 0 then
      begin
        MoveStepsAlpha := 0;
        MoveStepsDelta := 0;
        if Alpha then MoveStepsAlpha := BackLash.StepAD;
        if Delta then
          if BackLash.North then
            MoveStepsDelta := -BackLash.StepDEC
          else
            MoveStepsDelta :=  BackLash.StepDEC;
          TelescopeMove(MoveStepsAlpha,MoveStepsDelta);
          Enabled := false;
        end;
      end
      else
      begin
        if ((IsMCMTSlewing(C_Alpha) = false) and (IsMCMTSlewing(C_Delta) = false)) then
        begin
          Flag := BackLash.Pause div 1000;
          if Flag = 0 then
          begin
            MoveStepsAlpha := 0;
            MoveStepsDelta := 0;
            if Alpha then MoveStepsAlpha := BackLash.StepAD;
            if Delta then
              if BackLash.North then
                MoveStepsDelta := -BackLash.StepDEC
              else
                MoveStepsDelta :=  BackLash.StepDEC;
            TelescopeMove (MoveStepsAlpha, MoveStepsDelta);
            Enabled := false;
          end;
        end;
      end;
    Finally
    End;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Procedure CreateAllTimers;
Begin
    TimerAlphaPulse                := TTimerPulse.Create(nil);
    TimerAlphaPulse.Enabled        := False;
    TimerAlphaPulse.Resolution     := 50;
    TimerAlphaPulse.OnTimer        := TimerAlphaPulse.TimerEvent;
    TimerAlphaPulse.StopOrderAlpha := True;  // for the stoptelescope order
    TimerAlphaPulse.StopOrderDecli := False; // for the stoptelescope order
    TimerAlphaPulse.IsPulse        := False;

    TimerDecliPulse                := TTimerPulse.Create(nil);
    TimerDecliPulse.Enabled        := False;
    TimerAlphaPulse.Resolution     := 50;
    TimerDecliPulse.OnTimer        := TimerDecliPulse.TimerEvent;
    TimerDecliPulse.StopOrderAlpha := False; // for the stoptelescope order
    TimerDecliPulse.StopOrderDecli := True;  // for the stoptelescope order
    TimerDecliPulse.IsPulse        := False;

    TimerRetardateur               := TTimerRetardateur.Create(nil);
    TimerRetardateur.Enabled       := False;
    TimerRetardateur.OnTimer       := TimerRetardateur.TimerEvent;

    TimerBacklash                  := TTimerBacklash.Create(nil);
    TimerBacklash.Enabled          := False;
    TimerBacklash.OnTimer          := TimerBacklash.TimerEvent;

    TimerPark                      := TTimerPark.Create(nil);
    TimerPark.Enabled              := False;
    TimerPark.OnTimer              := TimerPark.TimerEvent;
End;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Procedure FreeAllTimers;
Begin
 TimerBacklash   .Free;
 TimerDecliPulse .Free;
 TimerAlphaPulse .Free;
 TimerPark       .Free;
 TimerRetardateur.Free;
End;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// A function that tell you if an axis is moving, but only for the point of vue
// MCMT

Function IsMCMTSlewing(Axis: integer): boolean;
Const
  procName='IsMCMTSlewing';

var
  RecBuff : array[0..7] of byte;
  SendBuff: array[0..3] of byte;
  abyte,
  numbyte,
  retry   : byte;
  NbOfBytesWrite,
  NbOfBytesRead : LongWord;
  deviceready   : integer ;
  HPort:THandle;
begin
  if not(SimulatorEnabled) then
  begin
    Result      := False;
    numbyte     := 2;
    SendBuff[0] := AxeZero + Axis;
    SendBuff[1] := 255;
    retry       := 3;
    result      := False;
    // Read a single byte from Interface
    Repeat
      FlushWriteRead(SendBuff, numbyte, NbOfBytesWrite, RecBuff, 1, NbOfBytesRead);
      abyte := RecBuff[0];
      retry := retry - 1;
    until (retry = 0) or (NbOfBytesRead <> 0);
    if (abyte=1) then deviceready := 1
                 else deviceready := 0;
    If (NbOfBytesRead=0) then raise Exception.Create('Erreur lecture tat mouvement '+Version+', pas de rponse !');
    Result:=(deviceready=0); // 0 = pas de mouvements
  end
  else
    Result:=(DeviceReadySimulator(Axis)=0);
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

function IsAxisSlewing_Backlash(Axis: integer): boolean;
begin
 result := IsMCMTSlewing(Axis) or TimerBacklash.Enabled;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


Function RetrieveDataAlphaDelta: boolean; // NOT IN VCL THREAD
var
  AlphaTemp, DeltaTemp  : Double;
  i                     : Integer;
  astring,receivestring : ShortString;
  SIGNE, MMSB,MSB, LSB, LLSB : Byte;
  tempPEC,temp,
  U                     : Double;
  FactVitesse           : Double;
  DeltaA                : Double;
  DeltaD                : Double;
  Readtemp              : Integer;
  H__, M__, S__         : Double ;
  Index_PecC            : Integer;
  CasePanel_info_pec    : Integer;
  CasePecState2         : Integer;
  CasePecState1         : Integer;
  position_PECD         : Double ;
  delta_time            : Double ;
  AH_Park               : Double ;
  LTelescopeZenithOuest : Boolean;
  L_DeltaRaw            : Longint;
  L_AlphaRaw            : Longint;
  L_AlphaRawPrevious    : Longint;
  L_DeltaRawPrevious    : Longint;
  L_ATime               : Double ;
  L_Alpha               : Double ;
  L_Delta               : Double ;
  L_AlphaInit           : Double ;
  L_AlphaCodeurInit     : Double ;
  L_DeltaInit           : Double ;
  L_DeltaCodeurInit     : Double ;
  S_err                 : String ;
Const
  procName='RetrieveDataAlphaDelta_Internal';

begin
  Try
    Try
      Result:=False;
      // delta_time = variation du temps sidral depuis l'initialisation des codeurs.
      // En JOURS
      SectionSTSTel.Enter;
      Try
        STSTel.ATime      := Now;
        L_AlphaRawPrevious:= STSTel.AlphaRawPrevious;
        L_DeltaRawPrevious:= STSTel.DeltaRawPrevious;
        L_ATime           := STSTel.ATime;
        L_AlphaInit       := STSTel.AlphaInit;
        L_AlphaCodeurInit := STSTel.AlphaCodeurInit;
        L_DeltaInit       := STSTel.DeltaInit;
        L_DeltaCodeurInit := STSTel.DeltaCodeurInit;
        Delta_time        := (STSTel.ATime - STSTel.FirstTime) * OmegaSid; // Bug corrig 06 MARS 08, bug de minuit!
      Finally
        SectionSTSTel.Leave;
      End;
      // Delta Time en RADIANS  2Pi = 23h56m04s Temps Local
      // Lecture encodeurs
      // AlphaRaw et DeltaRow : position des codeurs sur quatre octets, en micropas

      If (ReadAlphaDeltaEncoder(L_AlphaRaw,L_DeltaRaw)=_ERROR_) then //MCMT toujours branch?
      begin
        Try
          Result :=True ; // Important sinon mets un message generique
          DisableTimers ;
        Finally
          Raise Exception.create('Le port COM ne rpond plus, pas de communication avec MCMT....');
        End;
      end;
      DeltaA:=L_AlphaRaw - L_AlphaRawPrevious;
      DeltaD:=L_DeltaRaw - L_DeltaRawPrevious;
      // Si c'est une diffrence trop forte
      // avec la dernire fois ==> erreur  (plus d'un dizime de tour, soit 36)
      If (abs(DeltaA) > (SpeedIntMCMT.Resol_AD  / 10)) or
         (abs(DeltaD) > (SpeedIntMCMT.Resol_DEC / 10)) then
      begin
        If (ErrorCountRAce>5) then
        Begin
          DisableTimers ;
          S_err:=Format('Erreur critique de lecture des codeurs '+Version+', la diffrence entre 5 lectures conscutives est trop importante '+
                                       ':  RAi=%d,RAf=%d <-> DECi=%d,DECf=%d',
                                       [L_AlphaRawPrevious,L_AlphaRaw,L_DeltaRawPrevious,L_DeltaRaw]);
          Raise Exception.create(S_err)
        end
        else
        Begin
          Inc(ErrorCountRAce);
          Result        := True; // Ok fine BUT
          Exit;                  // Warning so Exit this proc for another trial
        End;
      End;
      ErrorCountRAce   := 0; // Ok reset le compteur
      Try
        ConsoleDataSem.Enter;
        Try
          InfosConsoleVar.NbStepPer_sec_meas_Alpha := (L_AlphaRaw - InfosConsoleVar.AlphaRaw_Previous) / (frac(L_atime - InfosConsoleVar.previous_Time) * 24 * 3600);
          InfosConsoleVar.AlphaRaw_Previous        :=  L_AlphaRaw;
          InfosConsoleVar.NbStepPer_sec_meas_Delta := (L_DeltaRaw - InfosConsoleVar.DeltaRaw_Previous) / (frac(L_atime - InfosConsoleVar.previous_Time) * 24 * 3600);
          InfosConsoleVar.DeltaRaw_Previous        :=  L_DeltaRaw;
        Finally
          ConsoleDataSem.Leave;
        End;
      Except
        // Sans importance...
      End;
      Try (* Trace d'une erreur vue par Ren Roy, assez dur a voir...  *)
        // RA
        AlphaTemp := 0;
        AlphaTemp := T_pi * Frac(L_AlphaRaw / SpeedIntMCMT.Resol_AD);
        If not SpeedIntMCMT.Inv_Codeur_AD then
          L_Alpha := T_pi * Frac((L_AlphaInit + AlphaTemp - L_AlphaCodeurInit + delta_time) / T_pi)
        else
          L_Alpha := T_pi * Frac((L_AlphaInit - AlphaTemp + L_AlphaCodeurInit + delta_time) / T_pi);
        While (L_Alpha >= T_pi) do
          L_Alpha := L_Alpha - T_pi; // G_Alpha est alors compris dans [0,T_pi]
        // DEC
        DeltaTemp := 0;
        DeltaTemp := T_pi*Frac(L_DeltaRaw / SpeedIntMCMT.Resol_DEC);
      Except
        On E:Exception do
        begin
          S_err :=Format('Erreur de calcul : L_AlphaRaw=%d, '            +
                                           'SpeedIntMCMT.Resol_AD=%d, ' +
                                           'AlphaTemp=%1.5f, '          +
                                           'L_AlphaInit=%1.5f, '        +
                                           'L_AlphaCodeurInit=%1.5f, '  +
                                           'delta_time=%1.5f, '         +
                                           'L_Alpha=%1.5f, '            +

                                           'L_DeltaRaw=%d, '            +
                                           'SpeedIntMCMT.Resol_DEC=%d, '+
                                           'DeltaTemp=%1.5f, '          +
                                           'Exception=%s',

                          [L_AlphaRaw             ,
                           SpeedIntMCMT.Resol_AD  ,
                           AlphaTemp              ,
                           L_AlphaInit            ,
                           L_AlphaCodeurInit      ,
                           delta_time             ,
                           L_Alpha                ,

                           L_DeltaRaw             ,
                           SpeedIntMCMT.Resol_DEC ,
                           DeltaTemp              ,
                           E.Message              ]);

          raise Exception.Create(S_err);
        end;
      End;
      If SpeedIntMCMT.Inv_Codeur_DEC then
      begin
        L_Delta := L_DeltaInit + DeltaTemp - L_DeltaCodeurInit;
        Logproc(procName,'GDelta1 ' + FloatToStr(L_DeltaInit) + ' ' + FloatToStr(DeltaTemp)+ ' ' + FloatToStr(L_DeltaCodeurInit));
      end
      else
      begin
        L_Delta := -DeltaTemp + L_DeltaInit + L_DeltaCodeurInit;
        Logproc(procName,'GDelta2 ' + FloatToStr(L_DeltaInit) + ' ' + FloatToStr(DeltaTemp)+ ' ' + FloatToStr(L_DeltaCodeurInit));
      end;
      L_Delta := (T_pi) * frac(L_Delta / (T_pi));
      Remet2Pi(L_Delta);
      if (L_Delta >Pi_d_2) and (L_Delta <= 3*Pi_d_2)  then
      begin
        if GerMountSetup.Monture_Allemande then LTelescopeZenithOuest:=True;
        L_Delta := Pi - L_Delta;
        L_Alpha := L_Alpha + Pi;
      end
      else
      begin
        if L_Delta >= (3 * Pi / 2) then L_Delta := L_Delta - T_pi;
        if GerMountSetup.Monture_Allemande then
        begin
          LTelescopeZenithOuest:=false;
          Logproc(procName,'Met  est');
        end;
      end;
      // Sauve l'etat du telescope
      if GerMountSetup.Monture_Allemande then
        WriteBoolToReg('TelescopeZenithOuest',LTelescopeZenithOuest);
      GerMountSetupSem.enter;
      GerMountSetup.TelescopeZenithOuest:=LTelescopeZenithOuest;
      GerMountSetupSem.Leave;
      Remet2Pi(L_Alpha);  // G_Alpha est alors compris dans [0;T_pi]
      FactVitesse                         := SpeedIntMCMT.V_Guidage_AD;
      ConsoleDataSem.Enter;
      InfosConsoleVar .FactVitesseDisplay := FactVitesse ;
      ConsoleDataSem.Leave;
      //AH_Park     := AH(long, G_Alpha, GMT);
      //GAlpha_Park := G_Alpha;
      //GDelta_Park := G_Delta;
      If Not IsAxisSlewing_BackLash(C_Alpha) then
      begin
       Logproc(procName,'No Slewing...');
       astring       := 'r' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0);
{$IFDEF MCMT32}
        receivestring := sendcommand(astring, 11, 0);
        If (receivestring<>'') then
          if (ord(receivestring[4]) <> 0) or (ord(receivestring[3]) <> 0) or
                            (ord(receivestring[2]) <> 0) or (ord(receivestring[1]) <> 0) then
          begin
            ConsoleDataSem.Enter;
            InfosConsoleVar.VitesseCurStepPer_sec := 80000000 / ((ord(receivestring[1])*256+ord(receivestring[2]))*256+
                                                                     ord(receivestring[3]))*256 + ord(receivestring[4]);
            ConsoleDataSem.Leave;
            PECSection.Enter;
            Try
              Index_PecC     := ord(receivestring[5]);
              PEC.PEC_Step   := ord(receivestring[7]) + 256 * ord(receivestring[6]);
              if not SpeedIntMCMT.Inv_Codeur_AD then
                PEC.PEC_Step := 25600 - PEC.PEC_Step;
            Finally
              PECSection.Leave;
            End;
            readtemp := 2 * (ord(receivestring[9]) + 256 * ord(receivestring[8]));
            if not SpeedIntMCMT.Inv_Codeur_AD then
              readtemp := 25600 - readtemp;
          end;
          CasePanel_info_pec := (ord(receivestring[10 ]) and 15);
          CasePecState1      :=  ord(receivestring[10]);
          CasePecState2      :=  ord(receivestring[11]);
{$ELSE}
        receivestring := sendcommand(astring, 10, 0);
        If (receivestring<>'') then
          if (ord(receivestring[2]) <> 0) or (ord(receivestring[1]) <> 0) then
          begin
            ConsoleDataSem.Enter;
            InfosConsoleVar.VitesseCurStepPer_sec := 625000 / (ord(receivestring[3]) / 10 +
                                                                     ord(receivestring[2]) +
                                                                     256 * ord(receivestring[1]));
            ConsoleDataSem.Leave;
            PECSection.Enter;
            Try
              Index_PecC     := ord(receivestring[4]);
              PEC.PEC_Step   := ord(receivestring[6]) + 256 * ord(receivestring[5]);
              if not SpeedIntMCMT.Inv_Codeur_AD then
                PEC.PEC_Step := 25600 - PEC.PEC_Step;
            Finally
              PECSection.Leave;
            End;
            readtemp := 2 * (ord(receivestring[8]) + 256 * ord(receivestring[7]));
            if not SpeedIntMCMT.Inv_Codeur_AD then
               readtemp := 25600 - readtemp;
          end;
          CasePanel_info_pec := (ord(receivestring[9 ]) and 15);
          CasePecState1      :=  ord(receivestring[9 ]);
          CasePecState2      :=  ord(receivestring[10]);
{$ENDIF}
          temp := AH(Lieu.long, L_Alpha, SpeedIntMCMT.GMT) * 12 / pi;
          H__ := int(temp);
          M__ := int((abs(temp) - abs(H__)) * 60);
          S__ := 60 * frac((abs(temp) - abs(H__)) * 60);
          if ((KING_Activated) or (PEC_Activated)) and (IsAxisSlewing_BackLash(C_Alpha) = false) then
          begin
            if King_Activated then
            begin
              FactVitesse := FactVitesse / getKingSpeedCorrection(AH_Park, L_Delta,Lieu.Lat);
              Logproc(procName,Format('King Update : %1.10f',[FactVitesse]));
            end;
            if PEC_Activated then
            begin
              PECSection.Enter;
              Try
                position_PECD := PEC.Periode_PEC[1] * frac(PEC.PEC_Step / 25600);
                // This is the derivative function of X + A1*cos(2Pi/Per*(X + Phi1)) + A2*cos(2Pi/Per*(X + Phi2))
                tempPEC := 0;
                For i := 1 to MaxPECHarmos do
                begin
                  U := (Pi / PEC.Periode_PEC[i]) * (2 * PEC.Amplitude_PEC[i]) *
                        sin(((T_pi) / PEC.Periode_PEC[i]) * (i * (position_PECD + PEC.Phase_PEC[i]) +
                        (15.041 * PEC.FCorrectionPeriodeMs / 1000)));
                  tempPEC := tempPEC + U;
                end;
              Finally
                PECSection.Leave;
              End;
              // Sinus derivative is negative !
              // 1 is derivatif of X
              tempPEC := 1 - tempPEC;
              FactVitesse := FactVitesse * tempPEC;
            end;
            ConsoleDataSem.Enter;
            InfosConsoleVar.FactVitesseDisplay := FactVitesse; // For Display
            ConsoleDataSem.Leave;
{$IFDEF MCMT32}
            FactVitesse := 80000000 / FactVitesse;
            MMSB :=HH(round(abs(FactVitesse)));
            MSB  :=LH(round(abs(FactVitesse)));
            LSB  :=HL(round(abs(FactVitesse)));
            LLSB :=LL(round(abs(FactVitesse)));
            SIGNE := 0;
            if (FactVitesse>0) then SIGNE := 1;
            astring := 'R' + AnsiChar(MMSB) + AnsiChar(MSB) + AnsiChar(LSB) + AnsiChar(LLSB) + AnsiChar(SIGNE);
{$ELSE}
            FactVitesse := 625000 / FactVitesse;
            MSB   := (round(abs(FactVitesse)) shr 8) and $FF;
            LSB   := (round(abs(FactVitesse))) and $FF;
            LLSB  := round(abs(10 * frac(FactVitesse))) and $FF;
            SIGNE := 0;
            if (FactVitesse>0) then SIGNE := 1;
            astring := 'R' + AnsiChar(MSB) + AnsiChar(LSB) + AnsiChar(LLSB) + AnsiChar(SIGNE);
{$ENDIF}
            sendcommand(astring, 0, 0);
          end; // KING_Activated PEC_Activated
          McmtPecSem.Enter;
          Try
            if InternalMCMT_PEC.ActiverPEC then
            begin
              InternalMCMT_PEC.ActiverPEC    := false;
              astring       := 'L' + AnsiChar(47) + AnsiChar(1) + AnsiChar(0) + AnsiChar(0);
              receivestring := sendcommand(astring, 0, 0);
            end;
            if InternalMCMT_PEC.DesactiverPEC then
            begin
              InternalMCMT_PEC.DesactiverPEC := false;
              astring       := 'L' + AnsiChar(47) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0);
              receivestring := sendcommand(astring, 0, 0);
            end;
            if InternalMCMT_PEC.EffacerPEC then
            begin
              InternalMCMT_PEC.EffacerPEC    := false;
              astring       := 'e' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0);
              receivestring := sendcommand(astring, 0, 0);
            end;

            if InternalMCMT_PEC.EnregistrerPEC then
            begin
              InternalMCMT_PEC.EnregistrerPEC := false;
              astring        := 'E' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0);
              receivestring  := sendcommand(astring, 0, 0);
            end;
          Finally
            McmtPecSem.Leave;
          End;
      end; // RA moves
      SectionSTSTel.Enter;
      STSTel.Delta            := L_Delta;
      STSTel.Alpha            := L_Alpha;
      STSTel.AlphaRaw         := L_AlphaRaw;
      STSTel.DeltaRaw         := L_DeltaRaw;
      STSTel.AlphaPointage    := L_Alpha;
      STSTel.DeltaPointage    := L_Delta;
      STSTel.AlphaRawPrevious := L_AlphaRaw; // Previous data
      STSTel.DeltaRawPrevious := L_DeltaRaw;
      SectionSTSTel.Leave;
      ConsoleDataSem.Enter;
      InfosConsoleVar.previous_Time       :=L_Atime;
      InfosConsoleVar .AlphaRaw           :=L_AlphaRaw;
      InfosConsoleVar .DeltaRaw           :=L_DeltaRaw;
      InfosConsoleVar .M__                :=M__     ;
      InfosConsoleVar .H__                :=H__     ;
      InfosConsoleVar .S__                :=S__     ;
      InfosConsoleVar .position_PECD      :=position_PECD      ;
      InfosConsoleVar .Index_PecC         :=Index_PecC         ;
      InfosConsoleVar .Readtemp           :=Readtemp           ;
      InfosConsoleVar .CasePanel_info_pec :=CasePanel_info_pec ;
      InfosConsoleVar .CasePecState2      :=CasePecState2      ;
      InfosConsoleVar .CasePecState1      :=CasePecState1      ;
      PECSection.Enter;
      InfosConsoleVar .PEC_Step           :=PEC.PEC_Step       ;
      PECSection.Leave;
      ConsoleDataSem.Leave;
      Result        := True; // Ok fine
    Finally
    End;
  Except
    On E:Exception do
    begin
      Raise Exception.create('Erreur '+Version+', lecture priodique des codeurs -> '+E.Message);
    end;
  End;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function TelescopeOuest : boolean;
begin
  GerMountSetupSem.Enter;
  Result:=GerMountSetup.TelescopeZenithOuest;
  GerMountSetupSem.leave;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

function ParkTel_at(AHPark, Dec2000: Double): Boolean;
var
  Accuracy  : Double;
  Reg       : TRegistry;
  AlphaPark : Double;

Const
  procName='ParkTel_at';

begin
  is_Parking := True;
  TimerPark.AngleHour := AHPark;
  TimerPark.Delta     := Dec2000;
  BackLash.Saved_Park := BackLash.Activated; // We desactivate backlash, but keep the state in mind
  BackLash.Activated  := False;

  AlphaPark:=CalculeTSL(Now+SpeedIntMCMT.GMT/24,Lieu.Long)-AHPark;
  Result   :=InitTelescopeSeeker(AlphaPark,Dec2000,Lieu.Lat,Lieu.Long);

  TimerPark.Interval := 1000;
  TimerPark.Flag     := 1;
  TimerPark.Enabled  := True;


  Reg         := TRegistry.Create;
  Reg.RootKey := HKEY_CURRENT_USER;

  if not (Reg.OpenKey('SOFTWARE\'+Version, False)) then
    Reg.OpenKey('SOFTWARE\'+Version, True);

  Reg.WriteFloat('Park_AH',  AHPark );
  Reg.WriteFloat('Park_Dec', Dec2000);

  Reg.CloseKey;
  Reg.Free;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function Calibrate(Alpha, Delta: Double): boolean; stdcall;

var
  AlphaRaw,
  DeltaRaw: longInt;
begin
  result    := ERROR;
  If (IsAxisSlewing_BackLash(C_Alpha) = false) then
  begin
    If GerMountSetup.Monture_Allemande then
    begin
      // Cannot be put like that here because this can be asked during image recentering for instance
      // Put something more intelligent...
      // ASK_FortelescopeOrientation(Nil);
      If TelescopeOuest then
      begin
        Delta:=Pi-Delta;
        Alpha:=Alpha+Pi;
      end;
    end;
    Remet2Pi(Alpha);
    ReadAlphaDeltaEncoder(AlphaRaw, DeltaRaw);
    SectionSTSTel.Enter;
    Try
      STSTel.AlphaRawPrevious := AlphaRaw;
      STSTel.DeltaRawPrevious := DeltaRaw;
      STSTel.AlphaCodeurInit  := T_pi * frac(AlphaRaw / SpeedIntMCMT.Resol_AD);
      STSTel.DeltaCodeurInit  := T_pi * frac(DeltaRaw / SpeedIntMCMT.Resol_DEC);
      STSTel.AlphaInit := Alpha;
      STSTel.DeltaInit := Delta;
      if (STSTel.DeltaInit<0) then STSTel.DeltaInit:=T_pi + STSTel.DeltaInit;
      STSTel.Alpha := Alpha;
      STSTel.Delta := Delta;
      STSTel.FirstTime := Now;
      STSTel.atime     := Now;
    Finally
      SectionSTSTel.Leave;
    end;
    result    := NOERROR;
  end;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

function MoveOffsetDirectStep(AlphaDSteps, DeltaDSteps: Integer): Boolean;
// NE SUPPORTE Qu'une SEULE vitesse !

Const
  procName='MoveOffsetDirectStep';

var
  NumOfSteps: longint;
  astring   : ShortString;
  MMSB, MSB,
  LSB, LLSB : byte;
begin
  Try

  NumOfSteps := AlphaDSteps;

  if abs(NumOfSteps) > 0 then
  begin
    MMSB := (abs(NumOfSteps) shr 24) and $FF;
    MSB  := (abs(NumOfSteps) shr 16) and $FF;
    LSB  := (abs(NumOfSteps) shr 8 ) and $FF;
    LLSB :=  abs(NumOfSteps) and $FF;
    // Bit de signe
    if NumOfSteps < 0 then MMSB := MMSB or $80;

    while (IsAxisSlewing_BackLash(C_Alpha)) do WaitMainthread(procName);
    astring := 'p' + AnsiChar(MMSB) + AnsiChar(MSB) + AnsiChar(LSB) + AnsiChar(LLSB);
    sendcommand(astring, 0, 0);
  end;


  NumOfSteps := DeltaDSteps;

  if abs(NumOfSteps) > 0 then
  begin
    MMSB := (abs(NumOfSteps) shr 24) and $FF;
    MSB  := (abs(NumOfSteps) shr 16) and $FF;
    LSB  := (abs(NumOfSteps) shr 8)  and $FF;
    LLSB :=  abs(NumOfSteps) and $FF;
    // Bit de signe
    if NumOfSteps < 0 then MMSB := MMSB or $80;

    while (IsAxisSlewing_BackLash(C_Delta)) do WaitMainthread(procName);
    astring := 'p' + AnsiChar(MMSB) + AnsiChar(MSB) + AnsiChar(LSB) + AnsiChar(LLSB);
    sendcommand(astring, 0, 1);
  end;

  finally
  end;

  result := NOERROR;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
procedure ASK_FortelescopeOrientation(Latitude:Double);
var
  Form_Etat_Telescope: TForm_Etat_Telescope;
  Reg                : TRegistry;
  ShowInterFace      : Boolean;
begin
  Try
    ShowInterFace:=True;
    Reg         := TRegistry.Create;
    Reg.RootKey := HKEY_CURRENT_USER;
    If not (Reg.OpenKey('SOFTWARE\'+Version, False)) then
                  Reg.OpenKey('SOFTWARE\'+Version, True);
    Try
      If ShowInterFace then
      begin
        Form_Etat_Telescope   := TForm_Etat_Telescope.create(Nil,Latitude); //XE3
        Try
          if latitude>0 then
          begin
            Form_Etat_Telescope.Image_Pied.Visible:=True;
            Form_Etat_Telescope.Image_Pied_HemisphereSud.Visible  :=false;
          end
          else
          begin
            Form_Etat_Telescope.Image_Pied.Visible:=false;
            Form_Etat_Telescope.Image_Pied_HemisphereSud.Visible  :=true;
          end;
          If TelescopeOuest then
          begin
            Form_Etat_Telescope.RadiogroupEtatTelescope.ItemIndex:=1;
            Form_Etat_Telescope.Image_Zenith_Ouest.Visible:=True;
            Form_Etat_Telescope.Image_Zenith_Est.Visible  :=false;
          end
          else
          begin
            Form_Etat_Telescope.RadiogroupEtatTelescope.ItemIndex:=0;
            Form_Etat_Telescope.Image_Zenith_Ouest.Visible:=false;
            Form_Etat_Telescope.Image_Zenith_Est.Visible  :=True;
          end;
          Form_Etat_Telescope.Top:=Form_Etat_Telescope.Top-Form_Etat_Telescope.height;
          Form_Etat_Telescope.BringTofront;
          Form_Etat_Telescope.ShowModal;
          GerMountSetupSem.Enter;
          If (Form_Etat_Telescope.RadiogroupEtatTelescope.ItemIndex=1)
                then GerMountSetup.TelescopeZenithOuest:=True
                else GerMountSetup.TelescopeZenithOuest:=False;
          GerMountSetupSem.Leave;
          // Save
          Reg.WriteBool   ('TelescopeZenithOuest',TelescopeOuest);
        Finally
          Form_Etat_Telescope.free;
        End;
      end;
      Reg.CloseKey;
    Finally
      Reg.Free;
    end;
  Except
    GerMountSetup.TelescopeZenithOuest:=True;
  End;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Function InitTelescopeSeeker(var Alpha, Delta, Latitude, Longitude: Double): Boolean;
var
  MoveStepsAlpha, MoveStepsDelta        : Integer;

  AlphaDiff, DeltaDiff, DeltaTempLat    : Double ;
  L_AlphaPointage,L_DeltaPointage       : Double ;
  Rayon_Pointage                        : Double ;

  // Test de retardement du mouvement delta
  // Montures allemandes
  Test                                  : Boolean;
  Ecart                                 : double;
  Omega_DEC_sur_Omega_AD                : double;
  LongitudeHoraire, AnglePolaire        : double;
  LongitudeHoraireCritiqueEst,
  AnglePolaireCritiqueEst               : double;
  LongitudeHoraireCritiqueOuest,
  AnglePolaireCritiqueOuest             : double;

const
  UneSecondesDArc:Double = 0.000005; //une seconde d'arc = 5 10-6 rad
  procName               = '';

Begin

  (* ALWAYS IN MAIN THREAD *)


  // Alpha     : RA Distination coordinates
  // Delta     : DEC Distination coordinates
  // Longitude : observation Longitude place coordinates
  // Latitude  : observation Latitude place coordinates
  // Accuracy  : Desired pointing accuracy

  LogEnterProc_DLL(ProcName);
  Logproc(procName,'InitTelescopeSeeker ' + FloatToStr(Alpha) + ' ' + FloatToStr(Delta)+' '+FloatToStr(Latitude)+' '+FloatToStr(Longitude));

  SectionSTSTel.Enter;
  L_AlphaPointage:=STSTel.AlphaPointage;
  L_DeltaPointage:=STSTel.DeltaPointage;
  SectionSTSTel.Leave;


  // Par ce test, AlphaPointage et DeltaPointage sont dtermins
  if not RetrieveDataAlphaDelta then
  begin
    result := false;
    Logproc(procName,'End Slew with error!');
    Exit;
  end;

  GerMountSetup.AngleHoraire:=AH(Longitude,Alpha,SpeedIntMCMT.GMT);
    // Angle horaire de la cible

  Logproc(procName,'AH='+FloatToStr(Longitude)+' '+FloatToStr(Alpha)+' '+FloatToStr(SpeedIntMCMT.GMT));
  Logproc(procName,'Angle Horaire ='+FloatToStr(GerMountSetup.AngleHoraire));

  TimerRetardateur.DabordAlpha  := False;
  TimerRetardateur.DabordDelta  := False;
  TimerRetardateur.Declanchement:= 0;
  // Rapport des vitesses
  Omega_DEC_sur_Omega_AD:=(SpeedIntMCMT.V_Point_R_DEC/SpeedIntMCMT.Resol_DEC)
                         /(SpeedIntMCMT.V_Point_R_AD /SpeedIntMCMT.Resol_AD);

  // pour viter les obstacles il faut tenir compte des vitesses angulaire
  // alpha et delta ventuellement diffrentes

  /////

  // *********************** Monture allemande *********************************
  // ***************************************************************************


  if GerMountSetup.Monture_Allemande then
  begin // Monture_Allemande

        Lieu.Long:=Longitude;

        CalculePositionMonture;

        if L_AlphaPointage-alpha > Pi then
          L_AlphaPointage:=L_AlphaPointage-T_Pi ;

        if L_AlphaPointage-alpha <-Pi then
          L_AlphaPointage:=L_AlphaPointage+T_Pi ;

        Rayon_Pointage:=arccos(sin(Delta)*sin(L_DeltaPointage)
                              +cos(Delta)*cos(L_DeltaPointage)*cos(L_AlphaPointage-Alpha));

        if L_AlphaPointage-Alpha > Pi then
          Rayon_Pointage:=T_Pi - Rayon_Pointage;

        LongitudeHoraireCritiqueEst  := -Pi_d_2+GerMountSetup.Angle_horaire_critique_Est;
        LongitudeHoraireCritiqueOuest:= +Pi_d_2+GerMountSetup.Angle_horaire_critique_Ouest;
        AnglePolaireCritiqueEst      := -Pi_d_2*Sign(Lieu.Lat)+GerMountSetup.Declinaison_critique_Est  ;
        AnglePolaireCritiqueOuest    := +Pi_d_2*Sign(Lieu.Lat)-GerMountSetup.Declinaison_critique_Ouest;

        // test de pointage (avec ou sans retournement ?) ***********************
        // Ici on peut savoir si la monture va retourner

        if Rayon_pointage > UneSecondesDArc then          // s'il ne s'agit pas d'un retournement sur ordre
        begin

              if (Rayon_pointage > GerMountSetup.Rayon_minimal)  // si le retournement est permis
              then
              begin
                  // alors le test de retournement se fait par rfrence 
                  // l'angle horaire de retournement au pointage dclar dans le setup.

                  if GerMountSetup.AngleHoraire < GerMountSetup.Angle_horaire_Retournement
                    then      {le tube sera  l'Ouest du pied : non retourn}
                    begin
                        LongitudeHoraire:=GerMountSetup.AngleHoraire+Pi_d_2;
                        if LongitudeHoraire>Pi then LongitudeHoraire:=LongitudeHoraire-T_pi;
                        AnglePolaire:=Sign(Lieu.Lat)*Pi_d_2-Delta;               //Hmisphre Sud
                    end
                    else     {le tube sera  l'Est du pied : retourn}
                    begin
                        LongitudeHoraire:=GerMountSetup.AngleHoraire-Pi_d_2;
                        if LongitudeHoraire<-Pi then LongitudeHoraire:=LongitudeHoraire+T_pi;
                        AnglePolaire:=Delta-Sign(Lieu.Lat)*Pi_d_2;               //Hmisphre Sud
                    end;

              end
              else                                   // si le retournement n'est pas permis
              begin

                  if TelescopeOuest
                    then      {Si le tube est  l'Ouest (non retourn), il doit rester  l'Ouest}
                    begin
                        LongitudeHoraire:=GerMountSetup.AngleHoraire+Pi_d_2;
                        if LongitudeHoraire > Pi then LongitudeHoraire:=LongitudeHoraire-T_pi;
                        AnglePolaire:=Sign(Lieu.Lat)*Pi_d_2-Delta;               //Hmisphre Sud
                    end
                    else     {Si le tube est  l'Est (retourn), il doit rester  l'Est}
                    begin
                        LongitudeHoraire:=GerMountSetup.AngleHoraire-Pi_d_2;
                        if LongitudeHoraire<-Pi then LongitudeHoraire:=LongitudeHoraire+T_pi;
                        AnglePolaire:=Delta-Sign(Lieu.Lat)*Pi_d_2;               //Hmisphre Sud
                    end;
              end;

              //si la cible est en zone critique, alors l'tat du tlescope
              //(retourn ou non retourn) doit tre permut.
              if ((LongitudeHoraire > LongitudeHoraireCritiqueOuest)
                     and (Sign(Lieu.Lat)*AnglePolaire <
                           Sign(Lieu.Lat)*AnglePolaireCritiqueOuest))
                    //   si la cible est en zone critique Ouest
                  or ((LongitudeHoraire < LongitudeHoraireCritiqueEst)
                     and (Sign(Lieu.Lat)*AnglePolaire >
                           Sign(Lieu.Lat)*AnglePolaireCritiqueEst))
                       // ou si la cible est en zone critique Est
                then   // retournement
                begin
                    LongitudeHoraire:=LongitudeHoraire-Pi;
                    if LongitudeHoraire<-Pi then LongitudeHoraire:=LongitudeHoraire+T_pi;
                    AnglePolaire:=-AnglePolaire
                end;

        end   // fin de test de pointage sans retournement sur ordre ******************
              // Prparation du retournement sur ordre *******************************
        else  // Si la cible est  moins d'une seconde d'arc
        begin

              if GerMountSetup.Retournement_Autorise  //et que le Retournement est autoris
              then
              begin
                    if not(is_Parking)                //et que l'on ne se trouve pas en procdure de Park
                      then
                        begin
                          AnglePolaire:=-GerMountSetup.AnglePolairePointage;
                          if TelescopeOuest
                            then LongitudeHoraire:=GerMountSetup.LongitudeHorairePointage-Pi
                            else LongitudeHoraire:=GerMountSetup.LongitudeHorairePointage+Pi ;

                        {il est impratif de quitter la section critique,
                        sinon cela produit tt ou tard un plantage gnral de prism}

                          if ((Sign(Lieu.Lat)*AnglePolaire >
                                   -Pi_d_2+Sign(Lieu.Lat)*GerMountSetup.Declinaison_Critique_Est)
                                 and (LongitudeHoraire<-Pi_d_2+GerMountSetup.Angle_horaire_critique_Est))
                             or ((Sign(Lieu.Lat)*AnglePolaire <
                                     +Pi_d_2-Sign(Lieu.Lat)*GerMountSetup.Declinaison_Critique_Ouest)
                                 and (LongitudeHoraire>+Pi_d_2+GerMountSetup.Angle_horaire_critique_Ouest))
                            then
                              begin  // Retournement a partir de cette position = risqu

                                Form_FeuRouge := TForm_FeuRouge.create(nil);
                                Form_FeuRouge.show;

                                Daccord_Renoncer  :=false;
                                Daccord_Perseverer:=false;

                                while not(Daccord_Renoncer) and not(Daccord_Perseverer)
                                   do Application.ProcessMessages;

                                if Daccord_Renoncer
                                  then
                                    begin
                                      result := True;
                                      Form_FeuRouge.free;
                                      Form_FeuRouge:=Nil;
                                      Exit;                {c'est plus sage en effet}
                                    end;

                                Form_FeuRouge.free;
                                Form_FeuRouge:=Nil;
                              end
                              else       //retournement sans problme
                              begin
                                Form_FeuVert := TForm_FeuVert.create(nil);
                                Form_FeuVert.show;

                                Daccord_Confirmer:=false;
                                Daccord_Annuler  :=false;

                                while not(Daccord_Confirmer) and not(Daccord_Annuler)
                                  do Application.ProcessMessages;

                                if Daccord_Annuler then
                                begin
                                    result := True;
                                    Form_FeuVert.free;
                                    Form_FeuVert:=Nil;
                                    Exit;
                                end;

                                Form_FeuVert.free;
                                Form_FeuVert:=Nil;
                              end;
                        end;
              end
              else    // si le retournement sur ordre n'est pas autoris
              begin // alors AlphaTemp et DeltaTemp seront nuls
                    LongitudeHoraire:=GerMountSetup.LongitudeHorairePointage;
                    AnglePolaire    :=GerMountSetup.AnglePolairePointage;
              end;

        end;        // fin de prparation du retournement sur ordre  *******


        AlphaDiff   :=LongitudeHoraire-GerMountSetup.LongitudeHorairePointage;
        DeltaDiff   :=AnglePolaire    -GerMountSetup.AnglePolairePointage;

        DeltaTempLat:=DeltaDiff*Sign(Lieu.Lat);


        //rglage des timers **************************************************

        //dans certains cas il est impratif de commencer par dgager le
        //tlescope pour un mouvement sur un seul axe : soit alpha, soit delta

        (* premier test : s'agit-il de sortir de la zone des positions basses Est ?
         Si oui, il faut imprativement tourner en alpha dans le sens horaire
         avant de dclancher le mouvement delta,
         afin d'viter de traverser la zone critique Est  *)

        Ecart:= abs(AnglePolaireCritiqueEst - GerMountSetup.AnglePolairePointage)
                  /Omega_DEC_sur_Omega_AD;

        if (GerMountSetup.LongitudeHorairePointage < LongitudeHoraireCritiqueEst - Ecart)
            and (LongitudeHoraire > LongitudeHoraireCritiqueEst)
          then
            begin
              TimerRetardateur.DabordAlpha:=True;
              TimerRetardateur.Declanchement:=LongitudeHoraireCritiqueEst - Ecart
            end;

        (* Deuxime test : s'agit-il de pntrer dans la zone des positions basses Est ?
         Si oui, il faudra terminer par un mouvement alpha dans le sens direct
         en limitant le temps de fonctionnement simultan des deux moteurs.
         Pour cela, selon o on se trouve, on retarde le dclanchement alpha ou delta
         afin d'viter de traverser la zone critique Est  *)

        Ecart:= abs(AnglePolaireCritiqueEst - AnglePolaire)/Omega_DEC_sur_Omega_AD;

        if (GerMountSetup.LongitudeHorairePointage > LongitudeHoraireCritiqueEst)
                             and (LongitudeHoraire < LongitudeHoraireCritiqueEst - Ecart)
                             and (DeltaTempLat < 0)   //Pas d'activation de timer dans le cas contraire
          then
            if GerMountSetup.LongitudeHorairePointage - LongitudeHoraireCritiqueEst + Ecart
                 < -DeltaTempLat /Omega_DEC_sur_Omega_AD
              then
               begin
                  TimerRetardateur.DabordDelta:=True;
                  TimerRetardateur.Declanchement:=AnglePolaire + Sign(Lieu.Lat)*Omega_DEC_sur_Omega_AD
                    *(GerMountSetup.LongitudeHorairePointage - LongitudeHoraireCritiqueEst + Ecart)
               end
              else
               begin
                  TimerRetardateur.DabordAlpha:=True;
                  TimerRetardateur.Declanchement:= LongitudeHoraireCritiqueEst
                                              -(DeltaTempLat+Ecart)/Omega_DEC_sur_Omega_AD
               end;

        (* Troisime test: s'agit-il de sortir la zone des positions basses Ouest ?
         Si oui, il faut imprativement tourner en alpha dans le sens direct
         avant de dclancher le mouvement delta,
         afin d'viter de traverser la zone critique Ouest  *)

        Ecart:= abs(AnglePolaireCritiqueOuest - GerMountSetup.AnglePolairePointage)
                  /Omega_DEC_sur_Omega_AD;

        If (GerMountSetup.LongitudeHorairePointage > LongitudeHoraireCritiqueOuest + Ecart)
                             and (LongitudeHoraire < LongitudeHoraireCritiqueOuest)
          then
           begin
              TimerRetardateur.DabordAlpha:=True;
              TimerRetardateur.Declanchement:=LongitudeHoraireCritiqueOuest + Ecart
           end;

        (* Quatrime test : s'agit-il de pntrer dans la zone des positions basses Ouest ?
         Si oui, il faudra terminer par un mouvement alpha dans le sens horaire
         en limitant le temps de fonctionnement simultan des deux moteurs.
         Pour cela, selon o on se trouve, on retarde le dclanchement alpha ou delta
         afin d'viter de traverser la zone critique Ouest  *)

        Ecart:= abs(AnglePolaireCritiqueOuest - AnglePolaire)/Omega_DEC_sur_Omega_AD;

        if (GerMountSetup.LongitudeHorairePointage < LongitudeHoraireCritiqueOuest)
                             and (LongitudeHoraire > LongitudeHoraireCritiqueOuest + Ecart)
                             and (DeltaTempLat > 0)
          then
          begin
            if LongitudeHoraireCritiqueOuest - GerMountSetup.LongitudehorairePointage + Ecart
                           < DeltaTempLat /Omega_DEC_sur_Omega_AD
              then
              begin
                  TimerRetardateur.DabordDelta:=True;

                  TimerRetardateur.Declanchement:= AnglePolaire - Sign(Lieu.Lat)*Omega_DEC_sur_Omega_AD
                    *(LongitudeHoraireCritiqueOuest - GerMountSetup.LongitudeHorairePointage + Ecart)
              end
              else
              begin
                  TimerRetardateur.DabordAlpha:=True;

                  TimerRetardateur.Declanchement:=LongitudeHoraireCritiqueOuest
                                          - (DeltaTempLat -Ecart)/Omega_DEC_sur_Omega_AD
              end;
          end;

        (* Cinquime test: en restant dans la zone principale de pointage,
           se rapproche-t-on de l'horizon ?
           et peut-on encore retarder le dclanchement du mouvement delta ?
           si oui, il faudra le faire afin d'viter que le tube ne pointe sous l'horizon *)

        If (GerMountSetup.LongitudeHorairePointage     <  Pi_d_2 + GerMountSetup.Angle_horaire_critique_Ouest)
           and (GerMountSetup.LongitudeHorairePointage > -Pi_d_2 + GerMountSetup.Angle_horaire_critique_Est)
           and (LongitudeHoraire <  Pi_d_2 + GerMountSetup.Angle_horaire_critique_Ouest)
           and (LongitudeHoraire > -Pi_d_2 + GerMountSetup.Angle_horaire_critique_Est)
        then
        begin
          if (TelescopeOuest and (DeltaTempLat > 0)
                             and (AlphaDiff > DeltaTempLat/Omega_DEC_sur_Omega_AD)
                             and (GerMountSetup.LongitudeHorairePointage < 0))
             or
             (not(TelescopeOuest) and (DeltaTempLat < 0)
                                  and (AlphaDiff < DeltaTempLat/Omega_DEC_sur_Omega_AD)
                                  and (GerMountSetup.LongitudeHorairePointage > 0))
            then
              begin
                TimerRetardateur.DabordAlpha:=True;
                TimerRetardateur.Declanchement:=LongitudeHoraire-DeltaTempLat/Omega_DEC_sur_Omega_AD;
              end;

        end;

        //fin de rglage des timers ********************************************


  end  // Monture_Allemande

  // *********************** Monture  fourche *********************************
  // ***************************************************************************

  else
  begin // Monture  fourche

        AnglePolaire:=Delta -  Pi_d_2;
        DeltaDiff   :=Delta -  L_DeltaPointage;

        if GerMountSetup.Passage_Interdit
        then
            (*l'algorithme est tel que le tlescope ne peut pas faire des tours
             complets autour de l'axe horaire.*)

            begin
              if (L_Alphapointage=STSTel.AlphaOldPoint)
                  and (L_DeltaPointage=STSTel.DeltaOldPoint)
                then
                  GerMountSetup.AngleHorairePointage:=GerMountSetup.AngleHorairePointagePrecedent+
                  (now-GerMountSetup.DatePointagePrecedent)*OmegaSid
                else
                  GerMountSetup.AngleHorairePointage:=AH(Longitude,L_AlphaPointage,SpeedIntMCMT.GMT);

              AlphaDiff := GerMountSetup.AngleHoraire-GerMountSetup.AngleHorairePointage;
            end
            else
            (* Algorithme plus rapide autorisant le tlescope  tourner librement
            autour de l'axe alpha, ce qui peut tre un inconvnient. *)
            begin
              AlphaDiff := L_AlphaPointage;

              if Alpha - Pi > L_AlphaPointage
                then AlphaDiff := AlphaDiff +  T_pi
                else
                if L_AlphaPointage - Pi > Alpha then Alpha := Alpha +  T_pi ;

              AlphaDiff := AlphaDiff - Alpha;
            end;

        (* test: se rapproche-t-on de l'horizon ?
           et peut-on retarder le dclanchement du mouvement delta ?
           si oui, il faudra le faire *)

        if (DeltaDiff<0)
            and (abs(AlphaDiff)>abs(DeltaDiff)/Omega_DEC_sur_Omega_AD)
          then
            begin
              TimerRetardateur.DabordAlpha:=True;
              if (alphaDiff>0)
                then TimerRetardateur.Declanchement:=GerMountSetup.AngleHoraire + DeltaDiff/Omega_DEC_sur_Omega_AD
                else TimerRetardateur.Declanchement:=GerMountSetup.AngleHoraire - DeltaDiff/Omega_DEC_sur_Omega_AD;
            end;


  end; // Monture  fourche

  // Steps to MOVE for slew

  MoveStepsDelta := round(SpeedIntMCMT.Resol_DEC * DeltaDiff  / (T_pi ));
  MoveStepsAlpha := round(SpeedIntMCMT.Resol_AD  * AlphaDiff  / (T_pi ));

  Logproc(procName,'Movesteps ' + IntToStr(MoveStepsAlpha) + ' ' + FloatToStr(MoveStepsDelta));

  //*** rattrapage de jeu ********************************************************
  //******************************************************************************

  TimerBacklash.Alpha := false;
  TimerBacklash.Delta := false;

  Try

    if BackLash.Activated then
    begin
        If (MoveStepsAlpha<0) then
        begin
           TimerBacklash.Alpha := True;
           Dec(MoveStepsAlpha, BackLash.StepAD);
        end;

        Case BackLash.Delta_Mode of
        0:  ;          //  Backlash Delta dactiv

        1:  begin      //  Correction lors d'un dplacement vers le Nord
                 if ((MoveStepsDelta > 0)
                 and (L_DeltaPointage < (90 - BackLash.Value / 60) * pi / 180))
                then
                  begin
                    TimerBacklash.Delta := True;
                    Inc(MoveStepsDelta, BackLash.StepDEC);
                    BackLash.North:=True;
                  end;
            end;

        2:  begin      // Correction lors d'un dplacement vers le Sud
              if ((MoveStepsDelta < 0)
                  and (L_DeltaPointage > (-90 + Backlash.value / 60) * pi / 180))
                then
                  begin
                    TimerBacklash.Delta := True;
                    Dec(MoveStepsDelta, BackLash.StepDEC);
                    BackLash.North:=false;
                  end;
            end;

        3:  begin      //  Correction lors d'un dplacement vers le Znith
              Test:=(tan(Latitude)<tan(Delta)*cos(GerMountSetup.AngleHoraire))
                            xor (AnglePolaire<0);
              if (MoveStepsDelta>0) and Test
                then
                  begin
                    TimerBacklash.Delta := True;
                    Inc(MoveStepsDelta,BackLash.StepDEC);
                    BackLash.North:=True;
                  end;
              if (MoveStepsDelta<0) and not(Test)
                then
                  begin
                    TimerBacklash.Delta := True;
                    Dec(MoveStepsDelta,BackLash.StepDEC);
                    BackLash.North:=false;
                  end;
            end;

        4:  begin      //  Correction lors d'un dplacement vers l'Horizon
              Test:=(tan(Latitude)<tan(Delta)*cos(GerMountSetup.AngleHoraire))
                            xor (AnglePolaire<0);
              if (MoveStepsDelta>0) and not(Test)
                then
                  begin
                    TimerBacklash.Delta := True;
                    Inc(MoveStepsDelta,BackLash.StepDEC);
                    BackLash.North:=True;
                  end;
              if (MoveStepsDelta<0) and Test
                then
                  begin
                    TimerBacklash.Delta := True;
                    Dec(MoveStepsDelta,BackLash.StepDEC);
                    BackLash.North:=false;
                  end;
            end;
        end; {case}
    end; // Backslash activated


    //programmation des timers **************************************************
    //***************************************************************************

    if TimerRetardateur.DabordDelta or TimerRetardateur.DabordAlpha then
      begin
          TimerRetardateur.Interval      := 1000;

          TimerRetardateur.AlphaCroissant:=(AlphaDiff>0);
          TimerRetardateur.DeltaCroissant:=(Sign(Lieu.Lat)*DeltaDiff>0);

          if TimerRetardateur.DabordDelta
            then           {Dclenchement mouvement Delta, Alpha en attente}
              begin
                TelescopeMoveAxe(MoveStepsDelta, C_Delta);
                TimerRetardateur.Movesteps := MovestepsAlpha;
                TimerRetardateur.Enabled := True;
              end
              else          {Dclenchement mouvement Alpha, Delta en attente}
              begin
                TelescopeMoveAxe(MoveStepsAlpha, C_Alpha);
                TimerRetardateur.Movesteps := MovestepsDelta;
                TimerRetardateur.Enabled := True;
              end;
      end
      else               {Dclenchement simultan Alpha et Delta}
        TelescopeMove(MoveStepsAlpha,MoveStepsDelta);

  Finally
  End;


  if TimerBacklash.Alpha or TimerBacklash.Delta then
  begin
    TimerBacklash.Interval := 1000;
    TimerBacklash.Flag     := 0;
    TimerBacklash.Enabled  := True;
    Logproc(procName,'Backslash enabled');
  end;

  STSTel.AlphaOldPoint := Alpha;  // always in this proc.
  STSTel.DeltaOldPoint := Delta;


  GerMountSetup.AngleHorairePointagePrecedent:=GerMountSetup.AngleHoraire;
  GerMountSetup.DatePointagePrecedent        :=now;

  if GerMountSetup.monture_allemande then
  begin
    GerMountSetupSem.Enter;
    GerMountSetup.TelescopeZenithOuest:= (AnglePolaire*Sign(Lieu.Lat)>0);         // Hmisphre sud
    GerMountSetupSem.Leave;
    WriteBoolToReg('TelescopeZenithOuest',TelescopeOuest);
  end;


 // Must be send back
//  Accuracy := DegToRad(1 / 30);
  result   := True;

  Logproc(procName,'End Slewing');
  LogEndProc_DLL(ProcName);

end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Procedure CalculePositionMonture;
Const
  procName='CalculePositionMonture';

  (* Main thread only *)

Var
  L_DeltaPointage : Double;
  L_AlphaPointage : Double;

begin

  RetrieveDataAlphaDelta; // get position from hardware

  SectionSTSTel.Enter;
  L_AlphaPointage:=STSTel.AlphaPointage;
  L_DeltaPointage:=STSTel.DeltaPointage;
  SectionSTSTel.Leave;

  GerMountSetup.AngleHorairePointage:=AH(Lieu.Long,L_AlphaPointage,SpeedIntMCMT.GMT);

  Logproc(procName,'CalculePositionMonture->AH                  =' + FloatToStr(Lieu.Long)+' '+FloatToStr(L_AlphaPointage)+' '+FloatToStr(SpeedIntMCMT.GMT));
  Logproc(procName,'CalculePositionMonture->AngleHorairePointage=' + FloatToStr(GerMountSetup.AngleHorairePointage));


  {-Pi < AngleHorairePointage < +Pi}
  if TelescopeOuest
  then   // Tlescope non retourn
  begin
        GerMountSetup.LongitudeHorairePointage:=GerMountSetup.AngleHorairePointage+Pi_d_2;
          {-Pi_d_2 < LongitudeHorairePointage < +3*Pi_d_2}
        if GerMountSetup.LongitudeHorairePointage>Pi
          then GerMountSetup.LongitudeHorairePointage:=GerMountSetup.LongitudeHorairePointage-T_pi;
          {-Pi < LongitudeHorairePointage < +Pi}
        GerMountSetup.AnglePolairePointage:=Sign(Lieu.Lat)*Pi_d_2-L_DeltaPointage;      //Hmisphre Sud
  end
  else  // Tlescope retourn
  begin

        GerMountSetup.LongitudeHorairePointage:=GerMountSetup.AngleHorairePointage-Pi_d_2;
          {-3*Pi_d_2 < LongitudeHorairePointage < Pi_d_2}
        if GerMountSetup.LongitudeHorairePointage<-Pi
          then GerMountSetup.LongitudeHorairePointage:=GerMountSetup.LongitudeHorairePointage+T_pi;
          {-Pi < LongitudeHorairePointage < +Pi}
        GerMountSetup.AnglePolairePointage:=L_DeltaPointage-Sign(Lieu.Lat)*Pi_d_2;      //Hmisphre Sud
  end;
  Logproc(procName,'CalculePositionMonture->LongitudeHorairePointage=' + FloatToStr(GerMountSetup.LongitudeHorairePointage));
  Logproc(procName,'CalculePositionMonture->AnglePolairePointage    =' + FloatToStr(GerMountSetup.AnglePolairePointage));
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// If AbleTo_WidthStand_SpeedChanges is false, leave it as it is.
// Otherwise allows the telescope to come back to sideral speed
// If PEC correction is activated, here you should use this function to desactivate it
// as so as the moon and solar speeds.

function ReturnToSideralSpeed: Boolean;
var
  astring   : shortstring;
Const
  ProcName='ReturnToSideralSpeed';
begin
  Try
    while (IsAxisSlewing_BackLash(C_Alpha)) do WaitMainthread(ProcName);
    astring := 'N' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0); // Commande No Park
    sendcommand(astring, 0, 0);
    while (IsAxisSlewing_BackLash(C_Alpha)) do WaitMainthread(ProcName);
    astring := 'S' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0); // Vitesse Siderale en AD
    sendcommand(astring, 0, 0);
    while (IsAxisSlewing_BackLash(C_Delta)) do WaitMainthread(ProcName);
    astring := 'S' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0);
    sendcommand(astring, 0, 1);
    SpeedIntMCMT.V_Guidage_AD := SpeedIntMCMT.V_Guidage_AD_SID;
  Finally
  End;
  result := True;
end;
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////// ///////// ///////// ///////// ///////// ///////// ///////// /
//////// ///////// ///////// ///////// ///////// ///////// ///////// /
// POINTAGE AUTO
//////// ///////// ///////// ///////// ///////// ///////// ///////// /
//////// ///////// ///////// ///////// ///////// ///////// ///////// /

procedure TelescopeMoveAxe(MoveSteps, Axe: integer);
var
  MMSB, MSB, LSB, LLSB: byte;
  astring             : shortstring;
Const
  procName='TelescopeMoveAxe';
begin
  LogEnterProc(procName);
  LogProc(procName,Format('Axe=%d Step=%d',[Axe,MoveSteps]));
  if MoveSteps=0 then exit;
  If (abs(MoveSteps) > 0) then
  begin
    MMSB := (abs(MoveSteps) shr 24) and $7F;
    MSB  := (abs(MoveSteps) shr 16) and $FF;
    LSB  := (abs(MoveSteps) shr 8)  and $FF;
    LLSB :=  abs(MoveSteps) and $FF;
    If MoveSteps < 0 then MMSB := MMSB or $80;
    While IsMCMTSlewing(Axe) do WaitMainthread(procName);
    astring := 'p' + AnsiChar(MMSB) + AnsiChar(MSB) + AnsiChar(LSB) + AnsiChar(LLSB);
    sendcommand(astring, 0, Axe);
  end;
  LogEndProc(procName);
end;
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Function TelescopeMove(MoveStepsAlpha, MoveStepsDelta: integer): Boolean;
begin
  TelescopeMoveAxe(MoveStepsDelta, C_Delta);
  TelescopeMoveAxe(MoveStepsAlpha, C_Alpha);
  result := True;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Stops the telescope (Alpha;delta : axis) and put it back in tracking mode

function STOPTelescope(Alpha, Delta: Boolean): Boolean;
// Stoppe le tlescope depuis les instructions envoyes par la raquette du tlescope
// une fois la commande stop faite, il faut un certain temps que la rampe descende pour que ca rende la main
// Ca rends la raquette pas bien reactive...
var
  astring : shortstring ;
  LSpeeds : Tspeeds     ;
Const
  procName='STOPTelescope';
begin
  InterlockedIncrement(Stopping);
  SpeedsSection.Enter;
  LSpeeds:=Speeds;
  SpeedsSection.Leave;
  TimerBacklash.Enabled    := False;
  TimerRetardateur.Enabled := False;
  if Alpha then
    if Delta then LogEnterProc(procName+ ' Alpha et Delta')
    else LogEnterProc(procName+ ' Alpha')
  else LogEnterProc(procName+ ' Delta');
  Try
    If Alpha then // Si l'arrt en Alpha a t demand ...
    begin
      case LSpeeds.SpeedAlpha of // With respect to the sideral speed (ie = 1x);
      SMax:
        begin
          if IsAxisSlewing_BackLash(C_Alpha) then
          begin
            StopSlewing(C_Alpha);
            while IsAxisSlewing_BackLash(C_Alpha) do      // Attends la fin de la rampe
              WaitMainthread(procName);
          end;
        end;

      SMedium:
        begin
          if IsAxisSlewing_BackLash(C_Alpha) then
          begin
            StopSlewing(C_Alpha);
            while IsAxisSlewing_BackLash(C_Alpha) do  // Attends la fin de la rampe
              WaitMainthread(procName);
          end;
        end;

      SSideral:
        begin
          //while IsAxisSlewing_OC(hcom,C_Alpha) do WaitMainthread(procName);
          astring := 'S' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0);
          sendcommand(astring, 0, 0);
        end;

      SNone: ;
      end;
    end;
    If Delta then // Si l'arrt en Delta a t demand
    begin
      case LSpeeds.SpeedDelta of // With respect to the sideral speed (ie = 1x);
      SMax:
        begin
          If IsAxisSlewing_BackLash(C_Delta) then
          begin
            StopSlewing(C_Delta);
            while IsAxisSlewing_BackLash(C_Delta) do  // Attends la fin de la rampe
              WaitMainthread(procName);
          end;
        end;

      SMedium:
        begin
          If IsAxisSlewing_BackLash(C_Delta) then
          begin
            StopSlewing(C_Delta);
            while IsAxisSlewing_BackLash(C_Delta) do  // Attends la fin de la rampe
              WaitMainthread(procName);
          end;
        end;

      SSideral:
        begin
          //while IsAxisSlewing_BackLash(Hcom,C_Delta) do WaitMainthread(procName);
          astring := 'S' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0);
          sendcommand(astring, 0, 1);
        end;

      SNone: ;
      end;
    end;
  Finally
    InterlockedDecrement(Stopping);
    LogEndProc(procName);
  End;
  Result := True;
end;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

function Get_AtPark_Core: WordBool;
Const
  procName='Get_AtPark_Core';
Var
  STrace:string;
begin
  LogEnterProc(procName);
  result := fParked;
  LogEndProc(procName);
  if fParked then STrace:='AtPark=true' else STrace:='AtPark=false';
  WriteToTrace(STrace);
end;

function Get_CanPark_Core: WordBool;
Const
  procName='Get_CanPark_Core';
begin
  LogEnterProc(procName);
  result := true;
  LogEndProc(procName);
  WriteToTrace('CanPark=true');
end;

function Get_CanSetPark_Core: WordBool;
Const
  procName='Get_CanSetPark_Core';
begin
  LogEnterProc(procName);
  result := true;
  LogEndProc(procName);
  WriteToTrace('CanSetPark=true');
end;

function Get_Connected_Core: WordBool;
Const
  procName='Get_Connected_Core';
Var
  STrace:string;
begin
  LogEnterProc(procName);
  result :=fConnected;
  LogEndProc(procName);
  if fConnected then STrace:='Connected=true' else STrace:='Connected=false';
  WriteToTrace(STrace);
end;

function Get_Declination_Core: Double;
Const
  procName='Get_Declination_Core';
begin
  LogEnterProc(procName);
  if RetrieveDataAlphaDelta then
  begin
    fDeclination:=STSTel.Delta/PI* 180;
    result := fDeclination;
    WriteToTrace('Declination-->'+FloatToStr(result));
  end
  else
  begin
    WriteToTrace('Declination-->Error');
    raise EoleSysError.Create('Can''t get declination', $80040401, 0);
  end;
  LogEndProc(procName+' '+FloatToStr(result));
end;

function Get_IsPulseGuiding_Core: WordBool;
Const
  procName='Get_IsPulseGuiding_Core';
Var
  STrace:string;
begin
  LogEnterProc(procName);
  result := (TimerAlphaPulse.IsPulse or TimerDecliPulse.IsPulse);
  LogEndProc(procName);
  if result then STrace:='IsPulseGuiding=true' else STrace:='IsPulseGuiding=false';
  WriteToTrace(STrace);
end;

function Get_RightAscension_Core: Double;
Const
  procName='Get_RightAscension_Core';
begin
  LogEnterProc(procName);
  if RetrieveDataAlphaDelta then
  begin
    fRightAscension:=STSTel.Alpha/PI*12;
    result := fRightAscension;
    WriteToTrace('RightAscension-->'+FloatToStr(result));
  end
  else
  begin
    WriteToTrace('RightAscension-->Error');
    raise EoleSysError.Create('Can''t get right ascension', $80040401, 0);
  end;
  LogEndProc(procName);
end;

function Get_SiderealTime_Core: Double;
Const
  procName='Get_SiderealTime_Core';
begin
  LogEnterProc(procName);
  result := CalculeTSL(Now+SpeedIntMCMT.GMT/24,Lieu.Long)*12/Pi;
  WriteToTrace('SiderealTime-->'+FloatToStr(result));
  LogEndProc(procName);
end;

function Get_Slewing_Core: WordBool;
Const
  procName='Get_Slewing_Core';
Var
  STrace:string;
begin
  LogEnterProc(procName);
  if (IsAxisSlewing_backlash(C_Alpha) = True) or (IsAxisSlewing_backlash(C_Delta) = True) then result:=true else result:=false;
  if result then
  begin
//    showmessage('Slewing=true');
    STrace:='Slewing=true';
  end
  else
  begin
//    showmessage('Slewing=false');
    STrace:='Slewing=false';
  end;
  WriteToTrace(STrace);
  LogEndProc(procName);
end;

function Get_TargetDeclination_Core: Double;
Const
  procName='Get_TargetDeclination_Core';
begin
  LogEnterProc(procName);
  if fTargetDecOK then
  begin
    result := fTargetDeclination;
    WriteToTrace('TargetDeclination-->'+FloatToStr(result));
  end
  else
  begin
    WriteToTrace('TargetDeclination-->Error');
    raise EoleSysError.Create('Target declination not available', $80040401, 0);
  end;
  LogEndProc(procName);
end;

function Get_TargetRightAscension_Core: Double;
Const
  procName='Get_TargetRightAscension_Core';
begin
  LogEnterProc(procName);
  if fTargetAscOK then
  begin
    result := fTargetRightAscension;
    WriteToTrace('TargetRightAscension-->'+FloatToStr(result));
  end
  else
  begin
    WriteToTrace('TargetRightAscension-->Error');
    raise EoleSysError.Create('Target right ascension not available', $80040401, 0);
  end;
  LogEndProc(procName);
end;

function Get_Tracking_Core: WordBool;
Const
  procName='Get_Tracking_Core';
Var
  STrace:string;
begin
  LogEnterProc(procName);
  result := fTracking;
  if result then STrace:='Tracking=true' else STrace:='Tracking=false';
  WriteToTrace(STrace);
  LogEndProc(procName);
end;

procedure AbortSlew_Core;
Const
  procName='AbortSlew_Core';
begin
  LogEnterProc(procName);
  TimerBacklash.Enabled    := false;
  TimerRetardateur.Enabled := false;
  Try
  StopSlewing(C_Alpha);
  StopSlewing(C_Delta);
  Finally
  End;
  WriteToTrace('AbortSlew');
  LogEndProc(procName);
end;

procedure MoveAxis_Core(Axe: TelescopeAxes; Rate: Double);
Const
  procName='MoveAxis_Core';
var
  StopAlpha, StopDelta: boolean;
  astring: shortstring ;
  LSpeeds: Tspeeds     ;
  OutC,Axis   : Integer     ;
  Ordre:AnsiChar;
  STrace:string;

  function Pas2DegAD(Pas:double):double;
  begin
    result:=Pas/SpeedIntMCMT.Resol_AD*360;
  end;

  function Pas2DegDEC(Pas:double):double;
  begin
    result:=Pas/SpeedIntMCMT.Resol_DEC*360;
  end;

  function AxeVitesse(Axe:integer;Vitesse:AnsiChar):boolean;
  begin
    try
      while IsAxisSlewing_BackLash(Axe) do WaitMainthread(procName);
      astring := Vitesse + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0);
      sendcommand(astring, 0,Axe);
      result:=true;
    except
      result:=false;
    end;
  end;

begin
  LogEnterProc(procName);
  if fParked then
  begin
    WriteToTrace('MoveAxis error telescope parked');
    raise EoleSysError.Create('Telescope is parked', $80040401, 0);
  end
  else
  If (InterlockedCompareExchange(Stopping,1,1)=1) then   // Do not enter because stopping is pending
  begin
    Exit;
  end;
  SpeedsSection.Enter;
  LSpeeds:=Speeds;
  SpeedsSection.Leave;
  //  Monture allemande ***********************************************
  //  Do not handle for mount correction
  If GerMountSetup.Monture_Allemande and TelescopeOuest and (Axe=axisPrimary) then
    Rate:=-Rate;
  // ***************************
  StopAlpha := false;
  StopDelta := false;
  if Axe= axisPrimary then
  begin
    STrace:='MoveAxis RightAscension';
    Axis:=C_Alpha;
    StopAlpha:=true;  // in case of stop!
    if Rate < 0 then  //OUEST
    begin
      if Rate<>0 then LEDO.Green;
      if Abs(Abs(Rate)-Pas2DegAD(SpeedIntMCMT.V_Point_R_AD))<0.1 then
      begin
        Ordre:='X';
        LSpeeds.SpeedAlpha := SMax;
      end
      else if Abs(Abs(Rate)-Pas2DegAD(SpeedIntMCMT.V_Point_L_AD))<0.01 then
      begin
        Ordre:='G';
        LSpeeds.SpeedAlpha := SMedium;
      end
      else
      begin
        Ordre:='D';
        LSpeeds.SpeedAlpha := SSideral;
      end;
    end
    else  //EST
    begin
      if Rate<>0 then LEDE.Green;
      if Abs(Abs(Rate)-Pas2DegAD(SpeedIntMCMT.V_Point_R_AD))<0.1 then
      begin
        Ordre:='W';
        LSpeeds.SpeedAlpha := SMax;
      end
      else if Abs(Abs(Rate)-Pas2DegAD(SpeedIntMCMT.V_Point_L_AD))<0.01 then
      begin
        Ordre:='F';
        LSpeeds.SpeedAlpha := SMedium;
      end
      else
      begin
        Ordre:='Q';
        LSpeeds.SpeedAlpha := SSideral;
      end;
    end;
  end
  else
  begin
    STrace:='MoveAxis Declination';
    Axis:=C_Delta;
    StopDelta:=true;  // in case of stop!
    if Rate < 0 then  //SUD
    begin
      if Abs(Abs(Rate)-Pas2DegDEC(SpeedIntMCMT.V_Point_R_DEC))<0.1 then
      begin
        Ordre:='W';
        LSpeeds.SpeedDelta := SMax;
      end
      else if Abs(Abs(Rate)-Pas2DegDEC(SpeedIntMCMT.V_Point_L_DEC))<0.01 then
      begin
        Ordre:='F';
        LSpeeds.SpeedDelta := SMedium;
      end
      else
      begin
        Ordre:='Q';
        LSpeeds.SpeedDelta := SSideral;
      end;
    end
    else  //NORD
    begin
      if Abs(Abs(Rate)-Pas2DegDEC(SpeedIntMCMT.V_Point_R_DEC))<0.1 then
      begin
        Ordre:='X';
        LSpeeds.SpeedDelta := SMax;
      end
      else if Abs(Abs(Rate)-Pas2DegDEC(SpeedIntMCMT.V_Point_L_DEC))<0.01 then
      begin
        Ordre:='G';
        LSpeeds.SpeedDelta := SMedium;
      end
      else
      begin
        Ordre:='D';
        LSpeeds.SpeedDelta := SSideral;
      end;
    end;
  end;

  if Rate = 0 then //case of stop!!
  begin
    if not (StopTelescope(StopAlpha, StopDelta)) then
    begin
      WriteToTrace('MoveAxis error Stop impossible');
      raise EoleSysError.Create('Stop Axis impossible', $80040401, 0);
    end;
  end
  else
  begin
//    showmessage('Axe='+inttostr(axis));
    if not (AxeVitesse(Axis,Ordre)) then
    begin
      WriteToTrace('MoveAxis error move impossible');
      raise EoleSysError.Create('Move Axis impossible', $80040401, 0);
    end;
  end;
  WriteToTrace(STrace+' '+FloatToStr(Rate)+'/s');
  SpeedsSection.Enter;
  Speeds:=LSpeeds;
  SpeedsSection.Leave;
  LogEndProc(procName);
end;

procedure Park_Core;
Const
  procName='Park_Core';
var
  AH_Temp,Delta_Temp:double;
begin
  LogEnterProc(procName);
  WriteToTrace('Park');
  if (not (fParked)) then
  begin
    if not (fTracking) then
    begin
      ReturnToSideralSpeed;
      fTracking := true;
    end;
    AH_Temp:=fParkAH*Pi/12;
    Delta_Temp:=fParkDelta*Pi/180;
    if not(ParkTel_At(AH_Temp,Delta_Temp)) then
    begin
      WriteToTrace('Park error');
      raise EoleSysError.Create('Park failed', $80040400, 0);
    end;
    fParked := true;
  end;
  WriteToTrace('Park');
  LogEndProc(procName);
end;

procedure Set_Connected_Core(Value: WordBool);
Const
  procName='Set_Connected_Core';
var
  astring,
  receivestring      : ShortString;
  FileIni            : TIniFile;
  Park_AH,
  Park_Dec,
  valueDouble        : Double;
  AutoComPort,
  found,
  ok,
  ValueBool          : Boolean;
  retour,
  i,
  ValueInt           : Integer;
  Mj_ra, Mn_ra, Mi_ra,
  Mj_dc, Mn_dc, Mi_dc: Byte;
  ValueLong          : longint;
  TimeZone           : TTimeZoneInformation;
  Reg                : TRegistry;
  NumCOMp            : Byte;
  KeyHandle          : HKEY;
  TmpPorts           : TStringList;
  Form_scan_mcmt     : TForm_scan_mcmt;
  ProcessingBox      : TProcessingBox;
  STrace             : String;
            {

  found              : Boolean;
  Mj_dc, Mn_dc, Mi_dc: Byte;
  }


begin
  LogEnterProc(procName);
  if Value then STrace:='Connected<--true' else STrace:='Connected<--false';
  WriteToTrace(STrace);
  if ((Value = true) and (fConnected = false)) then
  begin
    try
      Park_AH                  :=0;
      Park_Dec                 :=0;
      FormatSettings.Decimalseparator         := '.';
      AutoComPort              :=True;
      FormConsole              := Nil;
      Retour := GetTimeZoneInformation(TimeZone);
      case retour of
        TIME_ZONE_ID_UNKNOWN : SpeedIntMCMT.GMT := TimeZone.Bias div 60;
        TIME_ZONE_ID_STANDARD: SpeedIntMCMT.GMT := (TimeZone.Bias + TimeZone.StandardBias) div 60;
        TIME_ZONE_ID_DAYLIGHT: SpeedIntMCMT.GMT := (TimeZone.Bias + TimeZone.DaylightBias) div 60;
      else                   Raise Exception.create('Erreur Fatale : Librairie '+Version+' "GetTimeZoneInformation" a chou !');
      end;
      FileIni     := TIniFile.Create(return_MCMT_StoreIniFilePath);
      Reg         := TRegistry.Create;
      Try
        Reg.RootKey := HKEY_CURRENT_USER;
        if not (Reg.OpenKey('SOFTWARE\'+Version, False)) then Reg.OpenKey('SOFTWARE\'+Version, true);
        // Principe :
        // Si le fichier INI n'existe pas, charge les parametres depuis la base de registres
        LogEnabled:=False;
        if Reg.ValueExists('LogEnabled') then
        LogEnabled:= Reg.ReadBool('LogEnabled');
        SimulatorEnabled:=False;
        if Reg.ValueExists('SimulatorEnabled') then
        SimulatorEnabled:= Reg.ReadBool('SimulatorEnabled');
        TraceEnabled:=False;
        if Reg.ValueExists('TraceEnabled') then
        TraceEnabled:= Reg.ReadBool('TraceEnabled');
        If Reg.ValueExists('Port') then
        begin
          NumCOMp     := Reg.ReadInteger('Port');
          Try
            AutoComPort := Reg.ReadBool   ('FindAutoPort');
          Except
            AutoComPort := True;
          End;
        end
        else
        begin
          NumCOMp:= FileIni.ReadInteger('Configuration', 'Port', 1);
          Reg.WriteInteger('Port', NumCOMp);
        end;
        If Reg.ValueExists('TimeOutMs_ConnectionsWrite') then
        begin
          TimeOutMs_ConnectionsWrite := Reg.ReadInteger('TimeOutMs_ConnectionsWrite');
          If TimeOutMs_ConnectionsWrite<100  then TimeOutMs_ConnectionsWrite:=100 ;
          If TimeOutMs_ConnectionsWrite>5000 then TimeOutMs_ConnectionsWrite:=5000;
        end;
        If Reg.ValueExists('TimeOutMs_ConnectionsRead') then
        begin
          TimeOutMs_ConnectionsRead := Reg.ReadInteger('TimeOutMs_ConnectionsRead');
          If TimeOutMs_ConnectionsRead<100  then TimeOutMs_ConnectionsRead:=100 ;
          If TimeOutMs_ConnectionsRead>5000 then TimeOutMs_ConnectionsRead:=5000;
        end;
        if Reg.ValueExists('LastDelta') then
          STSTel.Delta := Reg.ReadFloat('LastDelta')
        else
        begin
          STSTel.Delta := strtofloat(FileIni.ReadString('Configuration', 'LastDelta', '0'));
          Reg.WriteFloat('LastDelta', STSTel.Delta);
        end;
        if Reg.ValueExists('LastAlpha') then
          STSTel.Alpha := Reg.ReadFloat('LastAlpha')
        else
        begin
          STSTel.Alpha := strtofloat(FileIni.ReadString('Configuration', 'LastAlpha', '0'));
          Reg.WriteFloat('LastAlpha', STSTel.Alpha);
        end;
        if Reg.ValueExists('DIR_AD') then
          SpeedIntMCMT.Inv_Codeur_AD := Reg.ReadBool('DIR_AD')
        else
        begin
          SpeedIntMCMT.Inv_Codeur_AD := FileIni.ReadBool('Parameters', 'DIR_AD', False);
          Reg.WriteBool('DIR_AD', SpeedIntMCMT.Inv_Codeur_AD);
        end;
        if Reg.ValueExists('DIR_DEC') then
          SpeedIntMCMT.Inv_Codeur_DEC := Reg.ReadBool('DIR_DEC')
        else
        begin
          SpeedIntMCMT.Inv_Codeur_DEC := FileIni.ReadBool('Parameters', 'DIR_DEC', False);
          Reg.WriteBool('DIR_DEC', SpeedIntMCMT.Inv_Codeur_DEC);
        end;
        if Reg.ValueExists('Resol_AD') then
          SpeedIntMCMT.Resol_AD := Reg.ReadInteger('Resol_AD')
        else
        begin
          SpeedIntMCMT.Resol_AD := FileIni.ReadInteger('Parameters', 'Resol_AD', 4608000);
          Reg.WriteInteger('Resol_AD',SpeedIntMCMT.Resol_AD);
        end;
        if (SpeedIntMCMT.Resol_AD=0) then raise Exception.create('Fatal : La rsolution en Alpha est nulle ! Verifier dans le panneau de configuration...');
        if Reg.ValueExists('Resol_DEC') then
          SpeedIntMCMT.Resol_DEC := Reg.ReadInteger('Resol_DEC')
        else
        begin
          SpeedIntMCMT.Resol_DEC := FileIni.ReadInteger('Parameters', 'Resol_DEC', 4608000);
          Reg.WriteInteger('Resol_DEC', SpeedIntMCMT.Resol_DEC);
        end;
        if (SpeedIntMCMT.Resol_DEC=0) then raise Exception.create('Fatal : La rsolution en Delta est nulle ! Verifier dans le panneau de configuration...');
        if Reg.ValueExists('V_Guidage_AD') then
          SpeedIntMCMT.V_Guidage_AD := Reg.ReadFloat('V_Guidage_AD')
        else
        begin
          SpeedIntMCMT.V_Guidage_AD := StrToFloat(FileIni.ReadString('Parameters', 'V_Guidage_AD', '1'));
          Reg.WriteFloat('V_Guidage_AD', SpeedIntMCMT.V_Guidage_AD);
        end;
        SpeedIntMCMT.V_Guidage_AD_SID := SpeedIntMCMT.V_Guidage_AD;
        if Reg.ValueExists('V_Guidage_DEC') then
          SpeedIntMCMT.V_Guidage_DEC := Reg.ReadFloat('V_Guidage_DEC')
        else
        begin
          SpeedIntMCMT.V_Guidage_DEC := StrToFloat(FileIni.ReadString('Parameters', 'V_Guidage_DEC', '1'));
          Reg.WriteFloat('V_Guidage_DEC', SpeedIntMCMT.V_Guidage_DEC);
        end;
        if Reg.ValueExists('V_Corr_P_AD') then
          SpeedIntMCMT.V_Corr_P_AD := Reg.ReadFloat('V_Corr_P_AD')
        else
        begin
          SpeedIntMCMT.V_Corr_P_AD := StrToFloat(FileIni.ReadString('Parameters', 'V_Corr_P_AD', '1'));
          Reg.WriteFloat('V_Corr_P_AD', SpeedIntMCMT.V_Corr_P_AD);
        end;
        if Reg.ValueExists('V_Corr_P_DEC') then
          SpeedIntMCMT.V_Corr_P_DEC := Reg.ReadFloat('V_Corr_P_DEC')
        else
        begin
          SpeedIntMCMT.V_Corr_P_DEC := strtofloat(FileIni.ReadString('Parameters', 'V_Corr_P_DEC', '1'));
          Reg.WriteFloat('V_Corr_P_DEC', SpeedIntMCMT.V_Corr_P_DEC);
        end;
        if Reg.ValueExists('V_Corr_M_AD') then
          SpeedIntMCMT.V_Corr_M_AD := Reg.ReadFloat('V_Corr_M_AD')
        else
        begin
          SpeedIntMCMT.V_Corr_M_AD := strtofloat(FileIni.ReadString('Parameters', 'V_Corr_M_AD', '1'));
          Reg.WriteFloat('V_Corr_M_AD', SpeedIntMCMT.V_Corr_M_AD);
        end;
        if Reg.ValueExists('V_Corr_M_DEC') then
          SpeedIntMCMT.V_Corr_M_DEC := Reg.ReadFloat('V_Corr_M_DEC')
        else
        begin
          SpeedIntMCMT.V_Corr_M_DEC := strtofloat(FileIni.ReadString('Parameters', 'V_Corr_M_DEC', '1'));
          Reg.WriteFloat('V_Corr_M_DEC', SpeedIntMCMT.V_Corr_M_DEC);
        end;
        if Reg.ValueExists('V_Point_L_AD') then
          SpeedIntMCMT.V_Point_L_AD := Reg.ReadFloat('V_Point_L_AD')
        else
        begin
          SpeedIntMCMT.V_Point_L_AD := strtofloat(FileIni.ReadString('Parameters', 'V_Point_L_AD', '1'));
          Reg.WriteFloat('V_Point_L_AD', SpeedIntMCMT.V_Point_L_AD);
        end;
        if Reg.ValueExists('V_Point_L_DEC') then
          SpeedIntMCMT.V_Point_L_DEC := Reg.ReadFloat('V_Point_L_DEC')
        else
        begin
          SpeedIntMCMT.V_Point_L_DEC := strtofloat(FileIni.ReadString('Parameters', 'V_Point_L_DEC', '1'));
          Reg.WriteFloat('V_Point_L_DEC', SpeedIntMCMT.V_Point_L_DEC);
        end;
        if Reg.ValueExists('V_Point_R_AD') then
          SpeedIntMCMT.V_Point_R_AD := Reg.ReadFloat('V_Point_R_AD')
        else
        begin
          SpeedIntMCMT.V_Point_R_AD := strtofloat(FileIni.ReadString('Parameters', 'V_Point_R_AD', '1'));
          Reg.WriteFloat('V_Point_R_AD', SpeedIntMCMT.V_Point_R_AD);
        end;
        if Reg.ValueExists('V_Point_R_DEC') then
          SpeedIntMCMT.V_Point_R_DEC := Reg.ReadFloat('V_Point_R_DEC')
        else
        begin
          SpeedIntMCMT.V_Point_R_DEC := strtofloat(FileIni.ReadString('Parameters', 'V_Point_R_DEC', '1'));
          Reg.WriteFloat('V_Point_R_DEC', SpeedIntMCMT.V_Point_R_DEC);
        end;
        if Reg.ValueExists('Accel_AD') then
          SpeedIntMCMT.Accel_AD := Reg.ReadInteger('Accel_AD')
        else
        begin
          SpeedIntMCMT.Accel_AD := strtoint(FileIni.ReadString('Parameters', 'Accel_AD', '0'));
          Reg.WriteInteger('Accel_AD', SpeedIntMCMT.Accel_AD);
        end;
        if Reg.ValueExists('Accel_DEC') then
          SpeedIntMCMT.Accel_DEC := Reg.ReadInteger('Accel_DEC')
        else
        begin
          SpeedIntMCMT.Accel_DEC := strtoint(FileIni.ReadString('Parameters', 'Accel_DEC', '0'));
          Reg.WriteInteger('Accel_DEC', SpeedIntMCMT.Accel_DEC);
        end;
        if Reg.ValueExists('Backlash_Activated') then
          Backlash.Activated := Reg.ReadBool('Backlash_Activated')
        else
        begin
          Backlash.Activated := FileIni.ReadBool('Parameters', 'Backlash_Activated', false);
          Reg.WriteBool('Backlash_Activated', Backlash.Activated);
        end;
        if Reg.ValueExists('Backlash') then
          Backlash.Value := Reg.ReadFloat('Backlash')
        else
        begin
          Backlash.Value := strtofloat(FileIni.ReadString('Parameters', 'Backlash', '20'));
          Reg.WriteFloat('Backlash', Backlash.Value);
        end;
        if Reg.ValueExists('Backlash_Delta_Mode')
         then Backlash.Delta_Mode := Reg.ReadInteger('Backlash_Delta_Mode')
        else
        begin
          Backlash.Delta_Mode := FileIni.ReadInteger('Parameters', 'Backlash_Delta_Mode', 0);
         Reg.WriteInteger('Backlash_Delta_Mode',Backlash.Delta_Mode);
        end;
        if Reg.ValueExists('Backlash_Pause') then
          Backlash.Pause := Reg.ReadInteger('Backlash_Pause')
        else
        begin
          Backlash.Pause := FileIni.ReadInteger('Parameters', 'Backlash_Pause', 0) * 1000;
          Reg.WriteInteger('Backlash_Pause', Backlash.Pause);
        end;
        if Reg.ValueExists('Park_AH') then
          Park_AH := Reg.ReadFloat('Park_AH')
        else
        begin
          Park_AH := FileIni.Readfloat('ParkMode', 'AH', 0);
          Reg.WriteFloat('Park_AH', Park_AH);
        end;
        if Reg.ValueExists('Park_Dec') then
          Park_Dec := Reg.ReadFloat('Park_Dec')
        else
        begin
          Park_Dec := FileIni.Readfloat('ParkMode', 'DEC', 0);
          Reg.WriteFloat('Park_Dec', Park_Dec);
        end;
        if Reg.ValueExists('Passage_Interdit')
          then GerMountSetup.Passage_Interdit := Reg.ReadBool('Passage_Interdit')
        else
        begin
          GerMountSetup.Passage_Interdit := FileIni.ReadBool('Pointage','Passage_Interdit',false);
          Reg.WriteBool('Passage_Interdit',GerMountSetup.Passage_Interdit);
        end;
        if Reg.ValueExists('Monture_Allemande')
          then GerMountSetup.Monture_Allemande := Reg.ReadBool('Monture_Allemande')
        else
        begin
          GerMountSetup.Monture_Allemande := FileIni.ReadBool('Pointage','Monture_Allemande',false);
          Reg.WriteBool('Monture_Allemande',GerMountSetup.Monture_Allemande);
        end;
        if Reg.ValueExists('Angle_horaire_critique_Est')
          then GerMountSetup.Angle_horaire_critique_Est := Reg.ReadFloat('Angle_horaire_critique_Est')
        else
          begin
            GerMountSetup.Angle_horaire_critique_Est := FileIni.Readfloat('Pointage', 'Angle_horaire_critique_Est', 0);
            Reg.WriteFloat('Angle_horaire_critique_Est',GerMountSetup.Angle_horaire_critique_Est);
          end;

        if Reg.ValueExists('Angle_horaire_critique_Ouest')
          then GerMountSetup.Angle_horaire_critique_Ouest := Reg.ReadFloat('Angle_horaire_critique_Ouest')
          else
            begin
              GerMountSetup.Angle_horaire_critique_Ouest := FileIni.Readfloat('Pointage', 'Angle_horaire_critique_Ouest', 0);
              Reg.WriteFloat('Angle_horaire_critique_Ouest',GerMountSetup.Angle_horaire_critique_Ouest);
            end;

        if Reg.ValueExists('Angle_horaire_Retournement')
          then GerMountSetup.Angle_horaire_Retournement := Reg.ReadFloat('Angle_horaire_Retournement')
          else
            begin
              GerMountSetup.Angle_horaire_Retournement := FileIni.Readfloat('Pointage', 'Angle_horaire_Retournement', 0);
              Reg.WriteFloat('Angle_horaire_Retournement',GerMountSetup.Angle_horaire_Retournement);
            end;

        if Reg.ValueExists('Rayon_minimal')
          then GerMountSetup.Rayon_minimal := Reg.ReadFloat('Rayon_minimal')
          else
            begin
              GerMountSetup.Rayon_minimal := FileIni.Readfloat('Pointage', 'Rayon_minimal', 0);
              Reg.WriteFloat('Rayon_minimal',GerMountSetup.Rayon_minimal);
            end;

        if Reg.ValueExists('Retournement_Autorise')
          then GerMountSetup.Retournement_Autorise := Reg.ReadBool('Retournement_Autorise')
          else
            begin
              GerMountSetup.Retournement_Autorise :=
                         FileIni.ReadBool('Pointage','Retournement_Autorise',false);
              Reg.WriteBool('Retournement_Autorise',GerMountSetup.Retournement_Autorise);
            end;

        if Reg.ValueExists('Declinaison_Critique_Est')
          then GerMountSetup.Declinaison_Critique_Est :=
                       Reg.ReadFloat('Declinaison_Critique_Est')
          else
            begin
              GerMountSetup.Declinaison_Critique_Est :=
                FileIni.Readfloat('Pointage', 'Declinaison_Critique_Est', 0);
              Reg.WriteFloat('Declinaison_Critique_Est',GerMountSetup.Declinaison_Critique_Est);
            end;

        if Reg.ValueExists('Declinaison_Critique_Ouest')
          then GerMountSetup.Declinaison_Critique_Ouest :=
                 Reg.ReadFloat('Declinaison_Critique_Ouest')
          else
            begin
              GerMountSetup.Declinaison_Critique_Ouest :=
                FileIni.Readfloat('Pointage', 'Declinaison_Critique_Ouest', 0);
              Reg.WriteFloat('Declinaison_Critique_Ouest', GerMountSetup.Declinaison_Critique_Ouest);
            end;

        if Reg.ValueExists('TelescopeZenithOuest')              // Tlescope non retourn
          then GerMountSetup.TelescopeZenithOuest := Reg.ReadBool('TelescopeZenithOuest')
          else GerMountSetup.TelescopeZenithOuest := False;

        if Reg.ValueExists('Longitude')
          then Lieu.Long:=Reg.ReadFloat('Longitude')
          else Lieu.Long:=0;
        if Reg.ValueExists('Latitude')
          then Lieu.Lat:=Reg.ReadFloat('Latitude')
          else Lieu.Lat:=0;
        if Reg.ValueExists('Altitude')
          then Lieu.Alt:=Reg.ReadFloat('Altitude')
          else Lieu.Alt:=0;
        Backlash.Saved_Park := Backlash.Activated;
        Backlash.StepAD     := round(SpeedIntMCMT.Resol_AD  * Backlash.Value / 60 / 360);
        Backlash.StepDEC    := round(SpeedIntMCMT.Resol_DEC * Backlash.Value / 60 / 360);

        if Reg.ValueExists('King')
          then King_Activated:=Reg.ReadBool('King')
        else
        begin
          King_Activated:= FileIni.ReadBool('Parameters','King',false);
          Reg.WriteBool('King',King_Activated);
        end;

        if Reg.ValueExists('External_PEC')
          then ExternalMCMT_PEC.Activated:=Reg.ReadBool('External_PEC')
        else
        begin
          ExternalMCMT_PEC.Activated:=FileIni.ReadBool('Parameters','External_PEC',false);
          Reg.WriteBool('External_PEC',ExternalMCMT_PEC.Activated);
        end;
        if Reg.ValueExists('Periode0') then
          ExternalMCMT_PEC.Periode[0]:=Reg.ReadFloat('Periode0')
        else
        begin
          ExternalMCMT_PEC.Periode[0]:=FileIni.ReadFloat('Parameters','Periode0',0);
          Reg.WriteFloat('Periode0',ExternalMCMT_PEC.Periode[0]);
        end;
        if Reg.ValueExists('Periode1') then
          ExternalMCMT_PEC.Periode[1]:=Reg.ReadFloat('Periode1')
        else
        begin
          ExternalMCMT_PEC.Periode[1]:=FileIni.ReadFloat('Parameters','Periode1',0);
          Reg.WriteFloat('Periode1',ExternalMCMT_PEC.Periode[1]);
        end;
        if Reg.ValueExists('Periode2') then
          ExternalMCMT_PEC.Periode[2]:=Reg.ReadFloat('Periode2')
        else
        begin
          ExternalMCMT_PEC.Periode[2]:=FileIni.ReadFloat('Parameters','Periode2',0);
          Reg.WriteFloat('Periode2',ExternalMCMT_PEC.Periode[2]);
        end;
        if Reg.ValueExists('Periode3') then
          ExternalMCMT_PEC.Periode[3]:=Reg.ReadFloat('Periode3')
        else
        begin
          ExternalMCMT_PEC.Periode[3]:=FileIni.ReadFloat('Parameters','Periode3',0);
          Reg.WriteFloat('Periode3',ExternalMCMT_PEC.Periode[3]);
        end;
        if Reg.ValueExists('Phase0') then
          ExternalMCMT_PEC.Phase[0]:=Reg.ReadFloat('Phase0')
        else
        begin
          ExternalMCMT_PEC.Phase[0]:=FileIni.ReadFloat('Parameters','Phase0',0);
          Reg.WriteFloat('Phase0',ExternalMCMT_PEC.Phase[0]);
        end;
        if Reg.ValueExists('Phase1') then
          ExternalMCMT_PEC.Phase[1]:=Reg.ReadFloat('Phase1')
        else
        begin
          ExternalMCMT_PEC.Phase[1]:=FileIni.ReadFloat('Parameters','Phase1',0);
          Reg.WriteFloat('Phase1',ExternalMCMT_PEC.Phase[1]);
        end;
        if Reg.ValueExists('Phase2') then
          ExternalMCMT_PEC.Phase[2]:=Reg.ReadFloat('Phase2')
        else
        begin
          ExternalMCMT_PEC.Phase[2]:=FileIni.ReadFloat('Parameters','Phase2',0);
          Reg.WriteFloat('Phase2',ExternalMCMT_PEC.Phase[2]);
        end;
        if Reg.ValueExists('Phase3') then
          ExternalMCMT_PEC.Phase[3]:=Reg.ReadFloat('Phase3')
        else
        begin
          ExternalMCMT_PEC.Phase[3]:=FileIni.ReadFloat('Parameters','Phase3',0);
          Reg.WriteFloat('Phase3',ExternalMCMT_PEC.Phase[3]);
        end;
        if Reg.ValueExists('Amplitude0') then
          ExternalMCMT_PEC.Amplitude[0]:=Reg.ReadFloat('Amplitude0')
        else
        begin
          ExternalMCMT_PEC.Amplitude[0]:=FileIni.ReadFloat('Parameters','Amplitude0',0);
          Reg.WriteFloat('Amplitude0',ExternalMCMT_PEC.Amplitude[0]);
        end;
        if Reg.ValueExists('Amplitude1') then
          ExternalMCMT_PEC.Amplitude[1]:=Reg.ReadFloat('Amplitude1')
        else
        begin
          ExternalMCMT_PEC.Amplitude[1]:=FileIni.ReadFloat('Parameters','Amplitude1',0);
          Reg.WriteFloat('Amplitude1',ExternalMCMT_PEC.Amplitude[1]);
        end;
        if Reg.ValueExists('Amplitude2') then
          ExternalMCMT_PEC.Amplitude[2]:=Reg.ReadFloat('Amplitude2')
        else
        begin
          ExternalMCMT_PEC.Amplitude[2]:=FileIni.ReadFloat('Parameters','Amplitude2',0);
          Reg.WriteFloat('Amplitude2',ExternalMCMT_PEC.Amplitude[2]);
        end;
        if Reg.ValueExists('Amplitude3') then
          ExternalMCMT_PEC.Amplitude[3]:=Reg.ReadFloat('Amplitude3')
        else
        begin
          ExternalMCMT_PEC.Amplitude[3]:=FileIni.ReadFloat('Parameters','Amplitude3',0);
          Reg.WriteFloat('Amplitude3',ExternalMCMT_PEC.Amplitude[3]);
        end;
        Reg.CloseKey;
      Finally
        Reg.Free;
        FileIni.free;
      end;

      If AutoComPort and not(SimulatorEnabled) then
      Begin
        TmpPorts := TStringList.Create;
        Try
          EnumComPorts(TmpPorts);
          if (TmpPorts.Count>0) then
          begin
            For i:= 0 to TmpPorts.Count- 1 do
              Try
                Form_scan_mcmt:=TForm_scan_mcmt.create(Nil,TmpPorts); //XE3
              Except
                // AOwner error
                raise Exception.Create('Erreur interne DLL d''interface sur AOwner !');
              End;
            Form_scan_mcmt._Show_;
            Found         :=Form_scan_mcmt.Found_MCMT;
            NumCOMp       :=Form_scan_mcmt.NumCom;
            Form_scan_mcmt.Free;
          end
          else
          begin
            Found:=False;
            ShowMessage('Aucun port srie n''a t detect automatiquement sur cette machine !');
          end;
        Finally
          TmpPorts.Free;
        End;
      end  // AutoSearch
      else
      Begin
        // Essaie qd meme le port COM
        Found         := GetVersionAxes(NumCOMp, 0, Mj_ra, Mn_ra, Mi_ra);
      End;
      If not Found then
      begin
        If AutoComPort then
        Begin
          If Not User_Has_Stopped_COM_Scans then
            ShowMessage('Le programme n''a pas pu dtecter '+Version+' sur les ports COMs dtects sur cette machine !')
          else
            ShowMessage('La dtection '+Version+' a t stoppe par l''utilisateur !');
        End
        else
          ShowMessage(Format('Le programme n''a pas pu dtecter '+Version+' sur le port COM %d fix par l''utilisateur!',[NumCOMp]));
        fConnected:=false;
      end
      else
      begin
        // COM found  YES !
        // version du Soft
        ProcessingBox:=TProcessingBox.create(Nil,'Lecture version firmware et init '+Version+' ....'); //XE3
        ProcessingBox.Show;
        ProcessingBox.Refresh;
        GetVersionAxes(NumCOMp, 0, Mj_ra, Mn_ra, Mi_ra);
        GetVersionAxes(NumCOMp, 1, Mj_dc, Mn_dc, Mi_dc);
        ProcessingBox.Free;
        if not VersionOk(Mj_ra, Mn_ra, Mi_ra, Mj_dc, Mn_dc, Mi_dc) then
        begin
          fConnected:=false;
          ShowMessage   ('Erreur Fatale, La version firmware de '+Version+' n''est pas la bonne pour cette DLL, mettre  jour le firmware de MCMT !');
          Exit; // Exec finally sections
        end;
        Is_Parking := False;
        // Console creation
        {$DEFINE CONSOLE_OP}
        {$IFDEF CONSOLE_OP}
        FormConsole:= TFormConsole.create(Nil); // XE3
        {$ENDIF}
        Open_Cod_Eng(NumCOMp,ok); // ouvre le port COM ->> Reste OUVERT par la suite
        if Ok then
        begin
          Reg:= TRegistry.Create ;
          Try
            Reg.RootKey := HKEY_CURRENT_USER;
            if not (Reg.OpenKey('SOFTWARE\'+Version, False)) then Reg.OpenKey('SOFTWARE\'+Version, true);
            // Regarde si ce qui a t lu dans la base de registres ou le fichier INI est pareil que
            // l'EEPROM de MCMT -> ALPHA
            astring := 'K' + Ansichar(0) + Ansichar(0) + Ansichar(0) + Ansichar(0);
            receivestring := sendcommand(astring, 48, 0);
            If (receivestring<>'') then   // Ce sont la les donnes VITALES
            begin
{$IFDEF MCMT32}
              If (ord(receivestring[4]) <> 0) or (ord(receivestring[3]) <> 0) or (ord(receivestring[2]) <> 0)
                                or (ord(receivestring[1]) <> 0)then
              begin
                valueDouble:=(80000000 / (((ord(receivestring[4])*256 + ord(receivestring[3])) * 256 +
                  ord(receivestring[2]))*256+ord(receivestring[1])));
{$ELSE}
              If (ord(receivestring[2]) <> 0) or (ord(receivestring[1]) <> 0) then
              begin
                valueDouble:=(625000 / (ord(receivestring[3]) / 10 + ord(receivestring[2]) * 256 + ord(receivestring[1])));
{$ENDIF}
                if (valueDouble<>SpeedIntMCMT.V_Guidage_AD) then
                begin
                  SpeedIntMCMT.V_Guidage_AD:=valueDouble;
                  Reg.WriteFloat('V_Guidage_AD', SpeedIntMCMT.V_Guidage_AD);
                end;
              end;

{$IFDEF MCMT32}
              If (ord(receivestring[8]) <> 0) or (ord(receivestring[7]) <> 0) or (ord(receivestring[6]) <> 0)
                                or (ord(receivestring[5]) <> 0)then
              begin
                valueDouble:=(80000000 / (((ord(receivestring[8])*256 + ord(receivestring[7])) * 256 +
                  ord(receivestring[6]))*256+ord(receivestring[5])));
{$ELSE}
              If (ord(receivestring[5]) <> 0) or (ord(receivestring[4]) <> 0) then
              begin
                valueDouble:=(625000 / (ord(receivestring[5]) * 256 + ord(receivestring[4])));
{$ENDIF}
                if (valueDouble<>SpeedIntMCMT.V_Corr_P_AD) then
                begin
                  SpeedIntMCMT.V_Corr_P_AD:=valueDouble;
                  Reg.WriteFloat('V_Corr_P_AD',SpeedIntMCMT.V_Corr_P_AD);
                end;
              end;

{$IFDEF MCMT32}
              If (ord(receivestring[12]) <> 0) or (ord(receivestring[11]) <> 0) or (ord(receivestring[10]) <> 0)
                                or (ord(receivestring[9]) <> 0)then
              begin
                valueDouble:= (80000000/ (((ord(receivestring[12])*256 + ord(receivestring[11])) * 256 +
                        ord(receivestring[10]))*256+ord(receivestring[9])));
{$ELSE}
              if (ord(receivestring[7]) <> 0) or (ord(receivestring[6]) <> 0) then
              begin
                valueDouble:= (625000 / (ord(receivestring[7]) * 256 + ord(receivestring[6])));
{$ENDIF}
                if (valueDouble<>SpeedIntMCMT.V_Corr_M_AD) then
                begin
                  SpeedIntMCMT.V_Corr_M_AD:=valueDouble;
                  Reg.WriteFloat('V_Corr_M_AD',SpeedIntMCMT.V_Corr_M_AD);
                end;
              end;

{$IFDEF MCMT32}
              If (ord(receivestring[16]) <> 0) or (ord(receivestring[15]) <> 0) or (ord(receivestring[14]) <> 0)
                                or (ord(receivestring[13]) <> 0)then
              begin
                valueDouble:=(80000000 / (((ord(receivestring[16])*256 + ord(receivestring[15])) * 256 +
                        ord(receivestring[14]))*256+ord(receivestring[13])));
{$ELSE}
              if (ord(receivestring[9]) <> 0) or (ord(receivestring[8]) <> 0) then
              begin
                valueDouble:=(625000 / (ord(receivestring[9]) * 256 + ord(receivestring[8])) * 8);
{$ENDIF}
                if (ValueDouble<>SpeedIntMCMT.V_Point_L_AD) then
                begin
                  SpeedIntMCMT.V_Point_L_AD:=ValueDouble;
                  Reg.WriteFloat('V_Point_L_AD',SpeedIntMCMT.V_Point_L_AD);
                end;
              end;

{$IFDEF MCMT32}
              If (ord(receivestring[20]) <> 0) or (ord(receivestring[19]) <> 0) or (ord(receivestring[18]) <> 0)
                                  or (ord(receivestring[17]) <> 0)then
              begin
                ValueDouble:=(80000000 /(((ord(receivestring[20])*256 + ord(receivestring[19])) * 256 +
                     ord(receivestring[18]))*256+ord(receivestring[17])));
{$ELSE}
              if (ord(receivestring[11]) <> 0) or (ord(receivestring[10]) <> 0) then
              begin
                ValueDouble:=(625000 / (ord(receivestring[11]) * 256 + ord(receivestring[10])) * 32);
{$ENDIF}
                if (SpeedIntMCMT.V_Point_R_AD<>ValueDouble) then
                begin
                  SpeedIntMCMT.V_Point_R_AD:=valueDouble;
                  Reg.WriteFloat('V_Point_R_AD',SpeedIntMCMT.V_Point_R_AD);
                end;
              end;

{$IFDEF MCMT32}
              ValueInt:= ord(receivestring[21]);
{$ELSE}
              ValueInt:= ord(receivestring[12]);
{$ENDIF}
              If (ValueInt<>SpeedIntMCMT.Accel_AD) then
              begin
                SpeedIntMCMT.Accel_AD:=ValueInt;
                Reg.WriteInteger('Accel_AD', SpeedIntMCMT.Accel_AD);
              end;

{$IFDEF MCMT32}
              ValueBool:=(ord(receivestring[Dir_Guidage+1])=1);
{$ELSE}
              ValueBool:=(ord(receivestring[13])=1);
{$ENDIF}
              if (ValueBool<>SpeedIntMCMT.Inv_Codeur_AD) then
              begin
                SpeedIntMCMT.Inv_Codeur_AD:=ValueBool;
                Reg.WriteBool('DIR_AD', SpeedIntMCMT.Inv_Codeur_AD);
              end;

{$IFDEF MCMT32}
              ValueLong := 25600 * (ord(receivestring[38]) * 256 + ord(receivestring[39]));
{$ELSE}
              ValueLong := 25600 * (ord(receivestring[29]) * 256 + ord(receivestring[30]));
{$ENDIF}
              if (ValueLong<>SpeedIntMCMT.Resol_AD) and (ValueLong>0) then
              begin
                SpeedIntMCMT.Resol_AD:=ValueLong;
                Reg.WriteInteger('Resol_AD', SpeedIntMCMT.Resol_AD);
              end;
            end; //If (receivestring<>'')

            if not(Reg.ValueExists('Tracking')) then
            begin
              Reg.WriteBool('Tracking',true);
              fTracking := true;
            end
            else
            begin
              if Reg.ReadBool('Tracking') then
              begin
                fTracking := true;
              end
              else
              begin
                fTracking := false;
              end;
            end;

            // l'EEPROM de MCMT -> DELTA
            Receivestring := sendcommand(astring, 48, 1);
            if (receivestring<>'') then   // Ce sont la les donnes VITALES
            begin
{$IFDEF MCMT32}
              If (ord(receivestring[4]) <> 0) or (ord(receivestring[3]) <> 0) or (ord(receivestring[2]) <> 0)
                                  or (ord(receivestring[1]) <> 0)then
              begin
                valueDouble:=(80000000 / (((ord(receivestring[4])*256 + ord(receivestring[3])) * 256 +
                    ord(receivestring[2]))*256+ord(receivestring[1])));
{$ELSE}
              If (ord(receivestring[2]) <> 0) or (ord(receivestring[1]) <> 0) then
              begin
                valueDouble:=(625000 / (ord(receivestring[3]) / 10 + ord(receivestring[2]) * 256 + ord(receivestring[1])));
{$ENDIF}
                if (valueDouble<>SpeedIntMCMT.V_Guidage_DEC) then
                begin
                  SpeedIntMCMT.V_Guidage_DEC:=valueDouble;
                  Reg.WriteFloat('V_Guidage_DEC', SpeedIntMCMT.V_Guidage_DEC);
                end;
              end;

{$IFDEF MCMT32}
              If (ord(receivestring[8]) <> 0) or (ord(receivestring[7]) <> 0) or (ord(receivestring[6]) <> 0)
                                or (ord(receivestring[5]) <> 0)then
              begin
                valueDouble:=(80000000 / (((ord(receivestring[8])*256 + ord(receivestring[7])) * 256 +
                      ord(receivestring[6]))*256+ord(receivestring[5])));
{$ELSE}
              If (ord(receivestring[5]) <> 0) or (ord(receivestring[4]) <> 0) then
              begin
                valueDouble:=(625000 / (ord(receivestring[5]) * 256 + ord(receivestring[4])));
{$ENDIF}
                if (valueDouble<>SpeedIntMCMT.V_Corr_P_DEC) then
                begin

                  SpeedIntMCMT.V_Corr_P_DEC:=valueDouble;
                  Reg.WriteFloat('V_Corr_P_DEC',SpeedIntMCMT.V_Corr_P_DEC);
                end;
              end;

{$IFDEF MCMT32}
              If (ord(receivestring[12]) <> 0) or (ord(receivestring[11]) <> 0) or (ord(receivestring[10]) <> 0)
                                or (ord(receivestring[9]) <> 0)then
              begin
                valueDouble:= (80000000/ (((ord(receivestring[12])*256 + ord(receivestring[11])) * 256 +
                      ord(receivestring[10]))*256+ord(receivestring[9])));
{$ELSE}
              if (ord(receivestring[7]) <> 0) or (ord(receivestring[6]) <> 0) then
              begin
                valueDouble:= (625000 / (ord(receivestring[7]) * 256 + ord(receivestring[6])));
{$ENDIF}
                if (valueDouble<>SpeedIntMCMT.V_Corr_M_DEC) then
                begin
                  SpeedIntMCMT.V_Corr_M_DEC:=valueDouble;
                  Reg.WriteFloat('V_Corr_M_DEC',SpeedIntMCMT.V_Corr_M_DEC);
                end;
              end;

{$IFDEF MCMT32}
              If (ord(receivestring[16]) <> 0) or (ord(receivestring[15]) <> 0) or (ord(receivestring[14]) <> 0)
                                or (ord(receivestring[13]) <> 0)then
              begin
                valueDouble:=(80000000 / (((ord(receivestring[16])*256 + ord(receivestring[15])) * 256 +
                       ord(receivestring[14]))*256+ord(receivestring[13])));
{$ELSE}
              if (ord(receivestring[9]) <> 0) or (ord(receivestring[8]) <> 0) then
              begin
                valueDouble:=(625000 / (ord(receivestring[9]) * 256 + ord(receivestring[8])) * 8);
{$ENDIF}
                if (ValueDouble<>SpeedIntMCMT.V_Point_L_DEC) then
                begin
                  SpeedIntMCMT.V_Point_L_DEC:=ValueDouble;
                  Reg.WriteFloat('V_Point_L_DEC',SpeedIntMCMT.V_Point_L_DEC);
                end;
              end;

{$IFDEF MCMT32}
              If (ord(receivestring[20]) <> 0) or (ord(receivestring[19]) <> 0) or (ord(receivestring[18]) <> 0)
                                or (ord(receivestring[17]) <> 0)then
              begin
                ValueDouble:=(80000000 /(((ord(receivestring[20])*256 + ord(receivestring[19])) * 256 +
                    ord(receivestring[18]))*256+ord(receivestring[17])));
{$ELSE}
              if (ord(receivestring[11]) <> 0) or (ord(receivestring[10]) <> 0) then
              begin
                ValueDouble:=(625000 / (ord(receivestring[11]) * 256 + ord(receivestring[10])) * 32);
{$ENDIF}
                if (SpeedIntMCMT.V_Point_R_DEC<>ValueDouble) then
                begin
                  SpeedIntMCMT.V_Point_R_DEC:=valueDouble;
                  Reg.WriteFloat('V_Point_R_DEC',SpeedIntMCMT.V_Point_R_DEC);
                end;
              end;

{$IFDEF MCMT32}
              ValueInt:= ord(receivestring[VM_Acce+1]);
{$ELSE}
              ValueInt:= ord(receivestring[12]);
{$ENDIF}
              If (ValueInt<>SpeedIntMCMT.Accel_DEC) then
              begin
                SpeedIntMCMT.Accel_DEC:=ValueInt;
                Reg.WriteInteger('Accel_DEC', SpeedIntMCMT.Accel_DEC);
              end;

{$IFDEF MCMT32}
              ValueBool:=(ord(receivestring[Dir_Guidage+1])=1);
{$ELSE}
              ValueBool:=(ord(receivestring[13])=1);
{$ENDIF}
              if (ValueBool<>SpeedIntMCMT.Inv_Codeur_DEC) then
              begin
                SpeedIntMCMT.Inv_Codeur_DEC:=ValueBool;
                Reg.WriteBool('DIR_DEC', SpeedIntMCMT.Inv_Codeur_DEC);
              end;

{$IFDEF MCMT32}
              ValueLong := 25600 * (ord(receivestring[38]) * 256 + ord(receivestring[39]));
{$ELSE}
              ValueLong := 25600 * (ord(receivestring[29]) * 256 + ord(receivestring[30]));
{$ENDIF}
              if (ValueLong<>SpeedIntMCMT.Resol_DEC) and (ValueLong>0) then
              begin
                SpeedIntMCMT.Resol_DEC:=ValueLong;
                Reg.WriteInteger('Resol_DEC', SpeedIntMCMT.Resol_DEC);
              end;
            end;     //if (receivestring<>'')
            Reg.CloseKey;
          Finally
            Reg.Free;
          end;

          // Est ce que le telescope est park ?
          // OU le fait que le telescope est en mode STOP
          astring       := 'r' + Ansichar(0) + Ansichar(0) + Ansichar(0) + Ansichar(0);
{$IFDEF MCMT32}
          receivestring := sendcommand(astring, 11, 0);
          if (ord(receivestring[11]) = 1) then fParked := true else fParked := false;
{$ELSE}
          receivestring := sendcommand(astring, 10, 0);
          if (ord(receivestring[10]) = 1) then fParked := true else fParked := false;
{$ENDIF}
          Form2.ButtonParkUnpark.Enabled:=true;
          if fParked then
          begin
            STSTel.Alpha := CalculeTSL(Now+SpeedIntMCMT.GMT/24,Lieu.Long) - Park_AH;
            STSTel.Delta := Park_Dec;
            Form2.ButtonParkUnpark.Caption:='Unpark';
            If GerMountSetup.Monture_Allemande then
            begin
              ProcessingBox:=TProcessingBox.create(Nil,'Telescope park : orientation tube monture Allemande connue...'); //XE3
              ProcessingBox.Width:=520;
            end
            else
              ProcessingBox:=TProcessingBox.create(Nil,'Telescope park...');  //XE3
            ProcessingBox.Show;
            ProcessingBox.Refresh;
            Sleep(2000);
            ProcessingBox.Free;
          end
          else
          begin
          // Dans le cas d'une monture allemande, il faut savoir, lors de
          // l'initialisation des codeurs, si le tlescope est retourn ou non.
          // Lorsque l'on sort d'un "park", l'tat (retourn ou non) du tlescope
          // est inscrit dans la base de registre, il n'est donc pas ncessaire
          // de poser la question.
            Form2.ButtonParkUnpark.Caption:='Park';
            If GerMountSetup.Monture_Allemande then
            begin
            // De quel cot sommes nous ????
              ASK_FortelescopeOrientation(Lieu.Lat);  // <- Toujours demander a l'INIT
            end;  // German Mount
          end; // Park Active
          /////////////////////////////////
         TimeCheck                      := 1000; // 1s
          PEC.FCorrectionPeriodeMs       := TimeCheck;
          // Cree les 5 TIMERS
          CreateAllTimers;
          Try
            // Initialisation de Delta et Alpha avec dernires valeurs enregistres
            // Lit La valeur des encodeurs, fatal si ca ne repond pas...
            If not Calibrate(STSTel.Alpha,STSTel.Delta) then
               Raise Exception.create(Version+' initialisation -> Erreur fatale, Calibrate a chou !');
          Except
            FreeAllTimers;
            Raise;
          End;
          fConnected:=True;
        end;  // if Ok //Le COM a pu etre OUVERT
      end; // Com found
      Form2.TimerCoord.Enabled:=true;
    Except
      On E:Exception do
      Begin
        raise Exception.create('Erreur d''initialisation de '+Version+' -> '+E.Message);
      End;
    end;
  end;
  if ((Value = false) and (fConnected = true)) then
  begin
    Form2.TimerCoord.Enabled:=false;
    FormConsole.StopTimer;
    DisableTimers;
    CloseCom;
    FreeAllTimers;
    fConnected:=false;
    Form2.ButtonParkUnpark.Enabled:=true;
    Try
      Reg:= TRegistry.Create;
      Try
        Reg.RootKey := HKEY_CURRENT_USER;
        if not (Reg.OpenKey('SOFTWARE\'+Version, False)) then Reg.OpenKey('SOFTWARE\'+Version, True);
        SectionSTSTel.Enter;
        Try
          Reg.WriteFloat('LastAlpha', STSTel.Alpha);
          Reg.WriteFloat('LastDelta', STSTel.Delta);
        Finally
          SectionSTSTel.Leave;
        End;
        Reg.WriteBool ('TelescopeZenithOuest',TelescopeOuest);
        Reg.CloseKey;
      Finally
        Reg.Free;
      End;
      FileIni:= TIniFile.Create(return_MCMT_StoreIniFilePath);
      SectionSTSTel.Enter;
      Try
        with FileIni do
        begin
          WriteString('Configuration', 'LastDelta', floattostr(STSTel.Delta));
          WriteString('Configuration', 'LastAlpha', floattostr(STSTel.Alpha));
        end;
      Finally
        SectionSTSTel.Leave;
        FileIni.free;
      End;
    Except
    End;
  end;
  LogEndProc(procName);
end;

procedure Set_TargetDeclination_Core(Value: Double);
Const
  procName='Set_TargetDeclination_Core';
begin
  LogEnterProc(procName);
  if (Value > 90) or (Value < -90) then
  begin
    raise EoleSysError.Create('Declination out of [-90,90]', $80040401, 0);
  end
  else
  begin
    fTargetDeclination := Value;
    fTargetDecOK := true;
  end;
  LogEndProc(procName+' '+FloatToStr(Value));
  WriteToTrace('TargetDeclination<-- '+FloatToStr(Value));
end;

procedure Set_TargetRightAscension_Core(Value: Double);
Const
  procName='Set_TargetRightAscension_Core';
begin
  LogEnterProc(procName);
  if (Value > 24) or (Value < 0) then
  begin
    raise EoleSysError.Create('Right ascension out of [0,24]', $80040401, 0);
  end
  else
  begin
    fTargetRightAscension := Value;
    fTargetAscOK := true;
  end;
  LogEndProc(procName+' '+FloatToStr(Value));
  WriteToTrace('TargetRightAscension<-- '+FloatToStr(Value));
end;

procedure SetPark_Core;
Const
  procName='SetPark_Core';
var
  Alpha: double;
begin
  LogEnterProc(procName);
  if fConnected then
  begin
    Alpha := Get_RightAscension_Core;
    fParkAH :=Get_SiderealTime_Core- Alpha;
    fParkDelta := Get_Declination_Core;
  end;
  LogEndProc(procName);
  WriteToTrace('SetPark');
end;

function SetFreeSpeed(Speed_rad:double;device:integer): Boolean;
var
  astring : ShortString;
  SIGNE,MMSB,MSB,LSB,LLSB: byte;
  Resolution:integer;
  Speed_trs,aval:double;
Const
  Procname='SetFreeSpeed';

begin
  if device=C_Alpha then
  begin
    Resolution:=SpeedIntMCMT.Resol_AD;
  end
  else
  begin
    Resolution:=SpeedIntMCMT.Resol_DEC;
  end;
  Try
    if Speed_rad <> 0 then
    begin
      Speed_trs:=Speed_rad/T_pi*Resolution;
      aval:=80000000 / aval;
      MMSB :=HH(round(abs(aval)));
      MSB  :=LH(round(abs(aval)));
      LSB  :=HL(round(abs(aval)));
      LLSB :=LL(round(abs(aval)));
      SIGNE := 0;
      if (aval>0) then SIGNE := 1;
      astring := 'R' + AnsiChar(MMSB) + AnsiChar(MSB) + AnsiChar(LSB) + AnsiChar(LLSB) + AnsiChar(SIGNE);
      while (IsAxisSlewing_BackLash(device)) do WaitMainthread(procName);
      sendcommand(astring, 0, device);
    end;
  Finally
  End;
  result := True;
end;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function ReturnVersionDLL_str(Index:Integer): string;
Const
  VersionInfo: array [1..8] of string
    = ('FileDescription', 'CompanyName'   , 'FileVersion',
       'InternalName'   , 'LegalCopyRight', 'OriginalFileName',
       'ProductName'    , 'ProductVersion');
  Version_Inconnue='Version inconnue !';
var
  Handle   : DWord;
  Info     : Pointer;
  InfoData : Pointer;
  InfoSize : LongInt;
  DataLen  : UInt;
  LangPtr  : Pointer;
  InfoType : string;
  pidl       : PItemIDList;
  InFolder   : array[0..MAX_PATH] of Char;
  Chemin:string;
begin
  Result   := Version_Inconnue;
  Try

  SHGetSpecialFolderLocation(Application.Handle, CSIDL_PROGRAM_FILES_COMMON, pidl);
  SHGetPathFromIDList(PIDL, InFolder);
  Chemin:=InFolder;
  Chemin:=Chemin+'\ASCOM\Telescope\MCMT32_ASCOM.exe';
  InfoSize := GetFileVersionInfoSize(PChar(Chemin),Handle);
  If (InfoSize>0) then
    begin
      GetMem(Info, InfoSize);
      Try
        If GetFileVersionInfo(PChar(Chemin),Handle, InfoSize, Info) then
        begin
          InfoType := VersionInfo[index];
          if VerQueryValue(Info,'\VarFileInfo\Translation',LangPtr, DataLen) then
            InfoType:=Format('\StringFileInfo\%0.4x%0.4x\%s'#0,[LoWord(LongInt(LangPtr^)),
                                                              HiWord(LongInt(LangPtr^)), InfoType]);
          if VerQueryValue(Info,@InfoType[1],InfoData,Datalen) then  // Based on FileVersion
            result:=strPas(Pchar(InfoData));
        end;
      Finally
        FreeMem(Info,InfoSize);
      End;
    end;
  except
    Result := Version_Inconnue;
  end;
end;

procedure SetupDialog_Core;
Const
  procName='SetupDialog_Core';
var

  SetupTelescope  : TSetupTelescope;
  Form_Etat_Telescope : TForm_Etat_Telescope;
  Form_scan_mcmt  : TForm_scan_mcmt;
  DeviceIni       : TIniFile;
  Reg             : TRegistry;

  S               : string;
  found           : boolean;
  c,i             : Integer;
  TmpPorts        : TStringList;

  T               : string;

  Mj_ra,
  Mn_ra, Mi_ra    : Byte;
  Mj_dc, Mn_dc,
  Mi_dc           : Byte;

  COM_OK          : Boolean;
  LongSetup,
  LatSetup,
  AltSetup        : Double;
  AutoFindCOM     : Boolean;
  ProcessingBox   : TProcessingBox;
  Latitude        : Double;
  NumCOM          : Byte;
Const
  FileVersionProduct=3;

  Function ReadlatfromRegistry:Double;
  Var
    Reg             : TRegistry;
  Begin
    Result:=Pi/4;  // + 45N
    Reg         := TRegistry.Create;
    Try
    Reg.RootKey := HKEY_CURRENT_USER;
    If Reg.OpenKey('SOFTWARE\'+Version, False) then
    Begin
      If Reg.ValueExists('Latitude') then
        Result:=Reg.ReadFloat('Latitude');
      Reg.Closekey;
    End;
    Finally
      Reg.Free;
    End;
  End;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

begin
  LogEnterProc(procName);
  WriteToTrace('SetupDialog');
  COM_OK           := False;
  AutoFindCOM      := True;

  FormatSettings.decimalseparator := '.';

  Latitude         := ReadlatfromRegistry;

  SetupTelescope   := TSetupTelescope.create(Nil,Latitude); // XE3
  Form_Etat_Telescope := TForm_Etat_Telescope.create(Nil,Latitude); // par analogie ???

  SetupTelescope.caption := Version+' configuration, version : ';

  SetupTelescope.TabSheetASCOM_PEC_KING.TabVisible:=true; // Do not show (only for ASCOM)

  SetupTelescope.caption :=SetupTelescope.caption+ReturnVersionDLL_str(FileVersionProduct);


  DeviceIni   := TIniFile.Create(return_MCMT_StoreIniFilePath);
  Reg         := TRegistry.Create;

  Try
  Reg.RootKey := HKEY_CURRENT_USER;
  if not (Reg.OpenKey('SOFTWARE\'+Version, False)) then Reg.OpenKey('SOFTWARE\'+Version, true);
  NumCOM:=1;
  if Reg.ValueExists('Port') then
  begin
    NumCOM      := Reg.ReadInteger('Port');
    Try
    AutoFindCOM := Reg.ReadBool   ('FindAutoPort');
    Except
    AutoFindCOM := True;
    End;
  end
  else
  begin
    NumCOM := DeviceIni.ReadInteger('Configuration', 'Port', 1);
    Reg.WriteInteger('Port', NumCOM);
  end;
  if Reg.ValueExists('SimulatorEnabled') then
    SimulatorEnabled:= Reg.ReadBool('SimulatorEnabled')
  else
    SimulatorEnabled:=false;
  Reg.CloseKey;
  Finally
  Reg.Free;
  DeviceIni.free;
  End;

  TmpPorts := TStringList.Create;
  EnumComPorts(TmpPorts);

  Try

  if (TmpPorts.Count>0) then
   for I := 0 to TmpPorts.Count - 1 do
    SetupTelescope.ComboBoxNumCOM.Items.Add(TmpPorts[i]);

  SetupTelescope.ComboBoxNumCOM.Visible         :=AutoFindCOM;
  SetupTelescope.NbreEditPortSerieForced.Visible:=Not AutoFindCOM;
  SetupTelescope.NbreEditPortSerieForced.Text   :=Inttostr(NumCOM);


  if AutoFindCOM and (TmpPorts.Count>0) then
  Begin

      Form_scan_mcmt:=TForm_scan_mcmt.create(Nil,TmpPorts); //XE3
      Form_scan_mcmt._Show_;
      Found:= Form_scan_mcmt.Found_MCMT;
      NumCOM:=Form_scan_mcmt.NumCom;
      Form_scan_mcmt.Free;


  End
  Else
  Found:=True;

  ///////////////////

  If not Found then // toujours pas trouv?
  begin
    If AutoFindCOM then
    Begin

      If (TmpPorts.Count>0) then
      begin
          S := '';
          // create message String, COM related !
          c := 0;
          while (c < TmpPorts.Count) do
          begin
            T := TmpPorts[c];
            if (S<>'') then
              S := S + ', COM ' + T
            else
              S := 'COM ' + T;
            inc(c);
          end;

          if User_Has_Stopped_COM_Scans then
           S:='L''utilisateur a abandonn la rechercher des ports COM sur ' + S + ' ! Voulez-vous poursuivre quand mme ?'
          else
           S:='Le programme n''a pas pu dtecter '+Version+' sur ' + S + ' ! Voulez-vous poursuivre quand mme ?'

      end
      else
      S:='Aucun port COM n''a t dtect automatiquement sur ce PC, Voulez-vous poursuivre quand mme et le rentrer manuellement ?';


      If MessageDlg(S,mtConfirmation,[mbYes, mbNo], 0) = mrYes then
      begin
        SetupTelescope.ComboBoxNumCOM.ItemIndex := 0;
      end
      else
      begin
        SetupTelescope.Free; //Destroy...
        Exit; // calls finally
      end;

    End;
  end;

  ////////////////
{$IFDEF MCMT32}
  SetupTelescope.LabelFirmwareAD.Caption:='inconnu';
  SetupTelescope.LabelFirmwareDEC.Caption:='inconnu';
  if GetVersionAxes(NumCOM, 0, Mj_ra, Mn_ra, Mi_ra) then
    SetupTelescope.LabelFirmwareAD.Caption:=IntToStr(Mj_ra)+'.'+IntToStr(Mn_ra)+'.'+IntToStr(Mi_ra);
  if GetVersionAxes(NumCOM, 1, Mj_dc, Mn_dc, Mi_dc) then
    SetupTelescope.LabelFirmwareDEC.Caption:=IntToStr(Mj_dc)+'.'+IntToStr(Mn_dc)+'.'+IntToStr(Mi_dc);
{$ELSE}
{$ENDIF}

  If Not AutoFindCOM then
  begin
    ProcessingBox:=TProcessingBox.create(Nil,'Lecture version firmware....');  //XE3
    ProcessingBox.Show;
    ProcessingBox.Refresh;
    if GetVersionAxes(NumCOM, 0, Mj_ra, Mn_ra, Mi_ra) then
    begin
     COM_OK := True ;
     Found  := False;
     ProcessingBox.Free;
    end
    else
    begin
      ShowMessage(Format('Connexion  '+Version+' impossible sur ce port COM %d !',[NumCOM]));
      COM_OK := False;
      Found  := False;
      ProcessingBox.Free;
    end;
  end;


  if Found then
  Begin
     ProcessingBox:=TProcessingBox.create(Nil,'Lecture version firmware et init '+Version+'....');  //XE3
     ProcessingBox.Show;
     ProcessingBox.Refresh;
     If not GetVersionAxes(NumCOM, 0, Mj_ra, Mn_ra, Mi_ra) or
        not GetVersionAxes(NumCOM, 1, Mj_dc, Mn_dc, Mi_dc) or
        not VersionOk(Mj_ra, Mn_ra, Mi_ra, Mj_dc, Mn_dc, Mi_dc) then
     begin
      ProcessingBox.Free;
      ShowMessage('AVERTISSEMENT : La version firmware de '+Version+' n''est pas la bonne pour ce plugin, mettre  jour le firmware de '+Version+' sous peine d''obtenir un fonctionnement ratique !');
     end
     else
     ProcessingBox.Free;
  end;

  Finally
  TmpPorts.Free;
  end;

  Reg         := TRegistry.Create;
  DeviceIni   := TIniFile.Create(return_MCMT_StoreIniFilePath);

  Try
  Reg.RootKey := HKEY_CURRENT_USER;
  if not (Reg.OpenKey('SOFTWARE\'+Version, False)) then Reg.OpenKey('SOFTWARE\'+Version, true);

  with DeviceIni do // Regarde dans le .INI
  begin

    If AutoFindCOM then
    Begin
      if not Found then
      begin
        SetupTelescope.ComboBoxNumCOM.Enabled := True; // On peut le rentrer a la main
        COM_OK := False; //  Sinon il va tenter de lire les valeurs de MCMT !!
      end
      else
      begin
        SetupTelescope.ComboBoxNumCOM.text    := inttostr(NumCOM);
        SetupTelescope.ComboBoxNumCOM.Enabled := False;
      end;
    End;

    If Reg.ValueExists('LastDelta') then
      STSTel.Delta := Reg.ReadFloat('LastDelta')
    else
    begin
      STSTel.Delta := strtofloat(ReadString('Configuration', 'LastDelta', '0'));
      Reg.WriteFloat('LastDelta', STSTel.Delta);
    end;

    if Reg.ValueExists('LastAlpha') then
      STSTel.Alpha := Reg.ReadFloat('LastAlpha')
    else
    begin
      STSTel.Alpha := strtofloat(ReadString('Configuration', 'LastAlpha', '0'));
      Reg.WriteFloat('LastAlpha', STSTel.Alpha);
    end;

    if Reg.ValueExists('DIR_AD') then
      SetupTelescope.CheckBox_DIR_AD.Checked := Reg.ReadBool('DIR_AD')
    else
    begin
      SetupTelescope.CheckBox_DIR_AD.Checked := ReadBool('Parameters', 'DIR_AD', False);
      Reg.WriteBool('DIR_AD', SetupTelescope.CheckBox_DIR_AD.Checked);
    end;

    if Reg.ValueExists('DIR_DEC') then
      SetupTelescope.CheckBox_DIR_DEC.Checked := Reg.ReadBool('DIR_DEC')
    else
    begin
      SetupTelescope.CheckBox_DIR_DEC.checked := ReadBool('Parameters', 'DIR_DEC', False);
      Reg.WriteBool('DIR_DEC', SetupTelescope.CheckBox_DIR_DEC.Checked);
    end;

    if Reg.ValueExists('PEC_AD') then
      SetupTelescope.CheckBox_PEC.Checked := Reg.ReadBool('PEC_AD')
    else
    begin
      SetupTelescope.CheckBox_PEC.checked := ReadBool('Parameters', 'PEC_AD', False);
      Reg.WriteBool('PEC_AD', SetupTelescope.CheckBox_PEC.Checked);
    end;

    If Reg.ValueExists('TimeOutMs_ConnectionsWrite') then
    begin
     SetupTelescope.NumberEditWriteTimeOutms.Text := Inttostr(Reg.ReadInteger('TimeOutMs_ConnectionsWrite'));
    end
    else
     SetupTelescope.NumberEditWriteTimeOutms.text:=Inttostr(TimeOutMs_ConnectionsWrite);

    If Reg.ValueExists('TimeOutMs_ConnectionsRead') then
    begin
       SetupTelescope.NumberEditReadTimeoutMs.Text  := Inttostr(Reg.ReadInteger('TimeOutMs_ConnectionsRead'));
    end
    else
      SetupTelescope.NumberEditReadTimeoutMs.Text :=Inttostr(TimeOutMs_ConnectionsRead);

    If Reg.ValueExists('PurgeCommMethod') then
     SetupTelescope.CheckBoxPurgCOMM.Checked:= Reg.ReadBool('PurgeCommMethod')
    else
     SetupTelescope.CheckBoxPurgCOMM.Checked:= False;

    if Reg.ValueExists('LogEnabled') then
    begin
      LogEnabled:= Reg.ReadBool('LogEnabled');
    end
    else
      LogEnabled:=False;

    if Reg.ValueExists('SimulatorEnabled') then
    begin
      SimulatorEnabled:= Reg.ReadBool('SimulatorEnabled');
    end
    else
      SimulatorEnabled:=False;

    if Reg.ValueExists('TraceEnabled') then
    begin
      TraceEnabled:= Reg.ReadBool('TraceEnabled');
    end
    else
      TraceEnabled:=False;

    SetupTelescope.CheckBoxLogFile.Checked:=LogEnabled;
    SetupTelescope.CheckBoxTrace.Checked:=TraceEnabled;
{$IFDEF MCMT32}
    if Reg.ValueExists('FreeSpeed') then
      FreeSpeed:=Reg.ReadBool('FreeSpeed')
    else
      FreeSpeed:=False;
    SetupTelescope.CheckBoxFreeSpeed.OnClick:=nil;
    SetupTelescope.CheckBoxFreeSpeed.Checked:=FreeSpeed;
    SetupTelescope.CheckBoxFreeSpeed.OnClick:=SetupTelescope.CheckBoxFreeSpeedClick;
    SetupTelescope.CheckBoxSimulator.Checked:=SimulatorEnabled;
    if FreeSpeed then
    begin
      SetupTelescope.Edit_V_Guidage_AD.Enabled:=false;
      SetupTelescope.Edit_V_Corr_P_AD.Enabled:=false;
      SetupTelescope.Edit_V_Corr_M_AD.Enabled:=false;
      SetupTelescope.Edit_V_Point_L_AD.Enabled:=false;
      SetupTelescope.Edit_V_Guidage_DEC.Enabled:=false;
      SetupTelescope.Edit_V_Corr_P_DEC.Enabled:=false;
      SetupTelescope.Edit_V_Corr_M_DEC.Enabled:=false;
      SetupTelescope.Edit_V_Point_L_DEC.Enabled:=false;
      SetupTelescope.Panel1.Font.Color:=clInactiveCaption;
      SetupTelescope.Panel2.Font.Color:=clInactiveCaption;
      SetupTelescope.Panel25.Font.Color:=clInactiveCaption;
      SetupTelescope.Panel26.Font.Color:=clInactiveCaption;
      SetupTelescope.Panel27.Caption:='Vit. MAX';
      SetupTelescope.Panel30.Font.Color:=clInactiveCaption;
      SetupTelescope.Panel31.Font.Color:=clInactiveCaption;
      SetupTelescope.Panel32.Font.Color:=clInactiveCaption;
      SetupTelescope.Panel33.Font.Color:=clInactiveCaption;
      SetupTelescope.Panel34.Caption:='Vit. MAX';
    end
    else
    begin
      SetupTelescope.Edit_V_Guidage_AD.Enabled:=true;
      SetupTelescope.Edit_V_Corr_P_AD.Enabled:=true;
      SetupTelescope.Edit_V_Corr_M_AD.Enabled:=true;
      SetupTelescope.Edit_V_Point_L_AD.Enabled:=true;
      SetupTelescope.Edit_V_Guidage_DEC.Enabled:=true;
      SetupTelescope.Edit_V_Corr_P_DEC.Enabled:=true;
      SetupTelescope.Edit_V_Corr_M_DEC.Enabled:=true;
      SetupTelescope.Edit_V_Point_L_DEC.Enabled:=true;
      SetupTelescope.Panel1.Font.Color:=clBlack;
      SetupTelescope.Panel2.Font.Color:=clBlack;
      SetupTelescope.Panel25.Font.Color:=clBlack;
      SetupTelescope.Panel26.Font.Color:=clBlack;
      SetupTelescope.Panel27.Caption:='Vit. Pointe';
      SetupTelescope.Panel30.Font.Color:=clBlack;
      SetupTelescope.Panel31.Font.Color:=clBlack;
      SetupTelescope.Panel32.Font.Color:=clBlack;
      SetupTelescope.Panel33.Font.Color:=clBlack;
      SetupTelescope.Panel34.Caption:='Vit. Pointe';
    end;
{$ELSE}
{$ENDIF}

    if Reg.ValueExists('Raq_Can') then
    if Reg.ReadBool('Raq_Can')
         then SetupTelescope.RadioGroup_Raquette.ItemIndex:=1
         else SetupTelescope.RadioGroup_Raquette.ItemIndex:=0
    else
    begin
      if ReadBool('Parameters', 'Raq_Can', False)
            then SetupTelescope.RadioGroup_Raquette.ItemIndex:=1
            else SetupTelescope.RadioGroup_Raquette.ItemIndex:=0;
         Reg.WriteBool('Raq_Can',(SetupTelescope.RadioGroup_Raquette.ItemIndex=1));
    end;

    SpeedIntMCMT.Inv_Codeur_AD  := SetupTelescope.CheckBox_DIR_AD.Checked;
    SpeedIntMCMT.Inv_Codeur_DEC := SetupTelescope.CheckBox_DIR_DEC.checked;

    if Reg.ValueExists('Resol_AD') then
      SpeedIntMCMT.Resol_AD := Reg.ReadInteger('Resol_AD')
    else
    begin
      SpeedIntMCMT.Resol_AD := ReadInteger('Parameters', 'Resol_AD', 4608000);
      Reg.WriteInteger('Resol_AD', SpeedIntMCMT.Resol_AD);
    end;
    SetupTelescope.Edit_Resol_AD.text := inttostr(SpeedIntMCMT.Resol_AD);

    if Reg.ValueExists('Resol_DEC') then
      SpeedIntMCMT.Resol_DEC := Reg.ReadInteger('Resol_DEC')
    else
    begin
      SpeedIntMCMT.Resol_DEC := ReadInteger('Parameters', 'Resol_DEC', 4608000);
      Reg.WriteInteger('Resol_DEC', SpeedIntMCMT.Resol_DEC);
    end;
    SetupTelescope.Edit_Resol_DEC.text := inttostr(SpeedIntMCMT.Resol_DEC);

    if Reg.ValueExists('ReducAlpha') then
      SetupTelescope.EditResolAlpha.text := FloatToStr(Reg.ReadFloat('ReducAlpha'))
    else
    begin
      SetupTelescope.EditResolAlpha.text := ReadString('Parameters', 'ReducAlpha', '360');
      Reg.WriteFloat('ReducAlpha', StrToFloat(SetupTelescope.EditResolAlpha.text));
    end;

    if Reg.ValueExists('ReducDelta') then
      SetupTelescope.EditResolDelta.text := FloatToStr(Reg.ReadFloat('ReducDelta'))
    else
    begin
      SetupTelescope.EditResolDelta.text := ReadString('Parameters', 'ReducDelta', '360');
      Reg.WriteFloat('ReducDelta', StrToFloat(SetupTelescope.EditResolDelta.text));
    end;

    if Reg.ValueExists('V_Guidage_AD') then
      SetupTelescope.Edit_V_Guidage_AD.text := Format('%1.3f',[Reg.ReadFloat('V_Guidage_AD')])
    else
    begin
      SetupTelescope.Edit_V_Guidage_AD.text := ReadString('Parameters', 'V_Guidage_AD', '1');
      Reg.WriteFloat('V_Guidage_AD', StrToFloat(SetupTelescope.Edit_V_Guidage_AD.text));
    end;

    if Reg.ValueExists('V_Guidage_DEC') then
      SetupTelescope.Edit_V_Guidage_DEC.text := Format('%1.2f',[Reg.ReadFloat('V_Guidage_DEC')])
    else
    begin
      SetupTelescope.Edit_V_Guidage_DEC.text := ReadString('Parameters', 'V_Guidage_DEC', '1');
      Reg.WriteFloat('V_Guidage_DEC', StrToFloat(SetupTelescope.Edit_V_Guidage_DEC.text));
    end;

    if Reg.ValueExists('V_Corr_P_AD') then
      SetupTelescope.Edit_V_Corr_P_AD.text := Format('%1.2f',[Reg.ReadFloat('V_Corr_P_AD')])
    else
    begin
      SetupTelescope.Edit_V_Corr_P_AD.text := ReadString('Parameters', 'V_Corr_P_AD', '1');
      Reg.WriteFloat('V_Corr_P_AD', StrToFloat(SetupTelescope.Edit_V_Corr_P_AD.text));
    end;

    if Reg.ValueExists('V_Corr_P_DEC') then
      SetupTelescope.Edit_V_Corr_P_DEC.text := Format('%1.2f',[Reg.ReadFloat('V_Corr_P_DEC')])
    else
    begin
      SetupTelescope.Edit_V_Corr_P_DEC.text := ReadString('Parameters', 'V_Corr_P_DEC', '1');
      Reg.WriteFloat('V_Corr_P_DEC', StrToFloat(SetupTelescope.Edit_V_Corr_P_DEC.text));
    end;

    if Reg.ValueExists('V_Corr_M_AD') then
      SetupTelescope.Edit_V_Corr_M_AD.text := Format('%1.2f',[Reg.ReadFloat('V_Corr_M_AD')])
    else
    begin
      SetupTelescope.Edit_V_Corr_M_AD.text := ReadString('Parameters', 'V_Corr_M_AD', '1');
      Reg.WriteFloat('V_Corr_M_AD', StrToFloat(SetupTelescope.Edit_V_Corr_M_AD.text));
    end;

    if Reg.ValueExists('V_Corr_M_DEC') then
      SetupTelescope.Edit_V_Corr_M_DEC.text := Format('%1.2f',[Reg.ReadFloat('V_Corr_M_DEC')])
    else
    begin
      SetupTelescope.Edit_V_Corr_M_DEC.text := ReadString('Parameters', 'V_Corr_M_DEC', '1');
      Reg.WriteFloat('V_Corr_M_DEC', StrToFloat(SetupTelescope.Edit_V_Corr_M_DEC.text));
    end;

    if Reg.ValueExists('V_Point_L_AD') then
      SetupTelescope.Edit_V_Point_L_AD.text := Format('%1.2f',[Reg.ReadFloat('V_Point_L_AD')])
    else
    begin
      SetupTelescope.Edit_V_Point_L_AD.text := ReadString('Parameters', 'V_Point_L_AD', '1');
      Reg.WriteFloat('V_Point_L_AD', StrToFloat(SetupTelescope.Edit_V_Point_L_AD.text));
    end;

    if Reg.ValueExists('V_Point_L_DEC') then
      SetupTelescope.Edit_V_Point_L_DEC.text := Format('%1.2f',[Reg.ReadFloat('V_Point_L_DEC')])
    else
    begin
      SetupTelescope.Edit_V_Point_L_DEC.text := ReadString('Parameters', 'V_Point_L_DEC', '1');
      Reg.WriteFloat('V_Point_L_DEC', StrToFloat(SetupTelescope.Edit_V_Point_L_DEC.text));
    end;

    if Reg.ValueExists('V_Point_R_AD') then
      SetupTelescope.Edit_V_Point_R_AD.text := Format('%1.2f',[Reg.ReadFloat('V_Point_R_AD')])
    else
    begin
      SetupTelescope.Edit_V_Point_R_AD.text := ReadString('Parameters', 'V_Point_R_AD', '1');
      Reg.WriteFloat('V_Point_R_AD', StrToFloat(SetupTelescope.Edit_V_Point_R_AD.text));
    end;

    if Reg.ValueExists('V_Point_R_DEC') then
      SetupTelescope.Edit_V_Point_R_DEC.text := Format('%1.2f',[Reg.ReadFloat('V_Point_R_DEC')])
    else
    begin
      SetupTelescope.Edit_V_Point_R_DEC.text := ReadString('Parameters', 'V_Point_R_DEC', '1');
      Reg.WriteFloat('V_Point_R_DEC', StrToFloat(SetupTelescope.Edit_V_Point_R_DEC.text));
    end;

    if Reg.ValueExists('Accel_AD') then
      SetupTelescope.ComboBox_Accel_AD.ItemIndex := Reg.ReadInteger('Accel_AD')
    else
    begin
      SetupTelescope.ComboBox_Accel_AD.ItemIndex := strtoint(ReadString('Parameters', 'Accel_AD', '0'));
      Reg.WriteInteger('Accel_AD', SetupTelescope.ComboBox_Accel_AD.ItemIndex);
    end;

    if Reg.ValueExists('Accel_DEC') then
      SetupTelescope.ComboBox_Accel_DEC.ItemIndex := Reg.ReadInteger('Accel_DEC')
    else
    begin
      SetupTelescope.ComboBox_Accel_DEC.ItemIndex := strtoint(ReadString('Parameters', 'Accel_DEC', '0'));
      Reg.WriteInteger('Accel_DEC', SetupTelescope.ComboBox_Accel_DEC.ItemIndex);
    end;

    if Reg.ValueExists('Backlash_Activated') then Backlash.Activated:= Reg.ReadBool('Backlash_Activated')
    else
    begin
      Backlash.Activated:=ReadBool('Parameters', 'Backlash_Activated', false);
      Reg.WriteBool('Backlash_Activated', Backlash.Activated);
    end;
    SetupTelescope.CheckBoxActivateBacklash.Checked           := Backlash.Activated;
    SetupTelescope.GroupBox_Parametres_Correction_Jeux.Visible:= Backlash.Activated;


    if Reg.ValueExists('Backlash') then
      SetupTelescope.NbreEditBacklash.Text := FloatToStr(Reg.ReadFloat('Backlash'))
    else
    begin
      SetupTelescope.NbreEditBacklash.Text := ReadString('Parameters', 'Backlash', '20');
      Reg.WriteFloat('Backlash', StrToFloat(SetupTelescope.NbreEditBacklash.Text));
    end;

    if Reg.ValueExists('Backlash_Delta_Mode')
        then SetupTelescope.RadioGroup_Mode_Correction_Delta.ItemIndex :=
                                        Reg.ReadInteger('Backlash_Delta_Mode')
    else
    begin
      SetupTelescope.RadioGroup_Mode_Correction_Delta.ItemIndex :=ReadInteger('Parameters', 'Backlash_Delta_Mode', 0);
      Reg.WriteInteger('Backlash_Delta_Mode',SetupTelescope.RadioGroup_Mode_Correction_Delta.ItemIndex);
    end;

    if Reg.ValueExists('Backlash_Pause') then
      SetupTelescope.ComboBoxPauseBacklash.ItemIndex := Reg.ReadInteger('Backlash_Pause') div 1000
    else
    begin
      SetupTelescope.ComboBoxPauseBacklash.ItemIndex := ReadInteger('Parameters', 'Backlash_Pause', 0);
      Reg.WriteInteger('Backlash_Pause', SetupTelescope.ComboBoxPauseBacklash.ItemIndex * 1000);
    end;

    // Type de monture *****************************************************
    if Reg.ValueExists('Monture_Allemande')
      then GerMountSetup.Monture_Allemande:=Reg.ReadBool('Monture_Allemande')
    else
    begin
      GerMountSetup.Monture_Allemande:=ReadBool('Pointage', 'Monture_Allemande', false);
      Reg.WriteBool('Monture_Allemande',GerMountSetup.Monture_Allemande);
    end;

    if GerMountSetup.Monture_Allemande then
    begin
      SetupTelescope.RadioGroupTypeMonture.ItemIndex :=1;
      SetupTelescope.GroupBoxMontureAllemande.Visible:=true;
      SetupTelescope.GroupBoxMonturefourche.Visible:=false;
      SetupTelescope.GroupBox_aide_Allemande1.Visible:=true;
      SetupTelescope.GroupBox_aide_Allemande2.Visible:=true;
      SetupTelescope.GroupBox_aide_Fourche.Visible:=False;
      If (Lieu.Lat>0) then
      begin
        SetupTelescope.Image_Pied.Visible     :=true;
        SetupTelescope.Image_Pied_HemisphereSud.Visible     :=false;
        Form_Etat_Telescope.Image_Pied.Visible              :=true;
        Form_Etat_Telescope.Image_Pied_HemisphereSud.Visible:=false;
      end
      else
      begin
        SetupTelescope.Image_Pied_HemisphereSud.Visible     :=true;
        SetupTelescope.Image_Pied.Visible                   :=false;
        Form_Etat_Telescope.Image_Pied_HemisphereSud.Visible:=true;
        Form_Etat_Telescope.Image_Pied.Visible              :=false;
      end;
    end
    else
    begin
      SetupTelescope.RadioGroupTypeMonture.ItemIndex:=0;
      SetupTelescope.GroupBoxMontureAllemande.Visible:=false;
      SetupTelescope.GroupBoxMonturefourche.Visible:=true;
    end;

    // Monture  fourche *****************************************************
    if Reg.ValueExists('Passage_Interdit')
      then GerMountSetup.Passage_Interdit:=Reg.ReadBool('Passage_Interdit')
    else
    begin
      GerMountSetup.Passage_Interdit :=ReadBool('Pointage', 'Passage_Interdit', false);
      Reg.WriteBool('Passage_Interdit',GerMountSetup.Passage_Interdit);
    end;
    if GerMountSetup.Passage_Interdit then
    begin
      SetupTelescope.RadioGroupAlgorithmePointage.ItemIndex:=1;
      SetupTelescope.Image_Antimeridien.Visible:=false;
      SetupTelescope.Image_Meridien.Visible:=true;
    end
    else
    begin
      SetupTelescope.RadioGroupAlgorithmePointage.ItemIndex:=0;
      SetupTelescope.Image_Antimeridien.Visible:=true;
      SetupTelescope.Image_Meridien.Visible:=false;
    end;

    // Monture allemande *****************************************************
      if Reg.ValueExists('TelescopeZenithOuest')
          then GerMountSetup.TelescopeZenithOuest:=Reg.ReadBool('TelescopeZenithOuest')
          else GerMountSetup.TelescopeZenithOuest:=false;

      if GerMountSetup.TelescopeZenithOuest
        then
          begin
            SetupTelescope.RadioGroupEtatTelescope.ItemIndex:=1;
            SetupTelescope.Image_Zenith_Est.Visible:=false;
            SetupTelescope.Image_Zenith_Ouest.Visible:=true;
          end
        else
          begin
            SetupTelescope.RadioGroupEtatTelescope.ItemIndex:=0;
            SetupTelescope.Image_Zenith_Est.Visible:=true;
            SetupTelescope.Image_Zenith_Ouest.Visible:=false;
          end;

      if Reg.ValueExists('Retournement_Autorise')
        then  GerMountSetup.Retournement_Autorise:=Reg.ReadBool('Retournement_Autorise')
        else
          begin
            GerMountSetup.Retournement_Autorise:=
                     ReadBool('Pointage', 'Retournement_Autorise', false);
            Reg.WriteBool('Retournement_Autorise',GerMountSetup.Retournement_Autorise) ;
          end;

      if GerMountSetup.Retournement_Autorise
        then          // Les retournements sont autoriss
                      // alors il faut afficher la fentre d'avertissements
          begin
            SetupTelescope.RadioGroup_retournement.ItemIndex:=0;
            SetupTelescope.GroupBox_Avertissement.Visible:=true;
          end
        else          // Les retournements ne sont pas autoriss
          begin
            SetupTelescope.RadioGroup_retournement.ItemIndex:=1;
            SetupTelescope.GroupBox_Avertissement.Visible:=false;
          end;

     if Reg.ValueExists('Angle_horaire_critique_Ouest')
        then SetupTelescope.SpinEdit_Angle_horaire_critique_Ouest.value :=
                                 Round(Reg.ReadFloat('Angle_horaire_critique_Ouest')/Pi*12*60)
        else
          begin
            SetupTelescope.SpinEdit_Angle_horaire_critique_Ouest.value :=
                  Round(ReadFloat('Pointage', 'Angle_horaire_critique_Ouest', Pi/12));
            Reg.WriteFloat('Angle_horaire_critique_Ouest',
                         SetupTelescope.SpinEdit_Angle_horaire_critique_Ouest.value*Pi/12/60);
          end;

       if Reg.ValueExists('Angle_horaire_critique_Est')
        then SetupTelescope.SpinEdit_Angle_horaire_critique_Est.value :=
                                 Round(Reg.ReadFloat('Angle_horaire_critique_Est')/Pi*12*60)
        else
          begin
            SetupTelescope.SpinEdit_Angle_horaire_critique_Est.value :=
                  Round(ReadFloat('Pointage', 'Angle_horaire_critique_Est', Pi/12));
            Reg.WriteFloat('Angle_horaire_critique_Est',
                         SetupTelescope.SpinEdit_Angle_horaire_critique_Est.value*Pi/12/60);
          end;

       if Reg.ValueExists('Angle_horaire_Retournement')
        then SetupTelescope.SpinEdit_Angle_horaire_Retournement.value :=
                                 Round(Reg.ReadFloat('Angle_horaire_Retournement')/Pi*12*60)
        else
          begin
            SetupTelescope.SpinEdit_Angle_horaire_Retournement.value :=
                  Round(ReadFloat('Pointage', 'Angle_horaire_Retournement', Pi/12));
            Reg.WriteFloat('Angle_horaire_Retournement',
                         SetupTelescope.SpinEdit_Angle_horaire_Retournement.value*Pi/12/60);
          end;

     if Reg.ValueExists('Rayon_minimal')
        then SetupTelescope.SpinEdit_Rayon_minimal.value :=
                                 Round(Reg.ReadFloat('Rayon_minimal')/Pi*180)
        else
          begin
            SetupTelescope.SpinEdit_Rayon_minimal.value :=
                  Round(ReadFloat('Pointage', 'Rayon_minimal', 0));
            Reg.WriteFloat('Rayon_minimal',
                         SetupTelescope.SpinEdit_Rayon_minimal.value*Pi/180);
          end;

      if Reg.ValueExists('Declinaison_critique_Est')
        then SetupTelescope.SpinEdit_Declinaison_critique_Est.value :=
                         Round(Reg.ReadFloat('Declinaison_critique_Est')/Pi*180)
        else
          begin
            SetupTelescope.SpinEdit_Declinaison_critique_Est.value :=
            Round(ReadFloat('Pointage', 'Declinaison_critique_Est', 0));
            Reg.WriteFloat('Declinaison_critique_Est',
              SetupTelescope.SpinEdit_Declinaison_critique_Est.value*Pi/180);
          end;

        if Reg.ValueExists('Declinaison_critique_Ouest')
        then SetupTelescope.SpinEdit_Declinaison_critique_Ouest.value :=
                        Round(Reg.ReadFloat('Declinaison_critique_Ouest')/Pi*180)
        else
          begin
            SetupTelescope.SpinEdit_Declinaison_critique_Ouest.value :=
            Round(ReadFloat('Pointage', 'Declinaison_critique_Ouest', 0));
            Reg.WriteFloat('Declinaison_critique_Ouest',
              SetupTelescope.SpinEdit_Declinaison_critique_Ouest.value*Pi/180);
          end;

    // Longitude

    if Reg.ValueExists('Longitude') then LongSetup := Reg.ReadFloat('Longitude')
    else
    begin
      LongSetup := 0;
      Reg.WriteFloat('Longitude', LongSetup);
    end;
    if LongSetup<0 then
    begin
      SetupTelescope.CheckBoxEast.Checked := true;
      LongSetup:=-LongSetup;
    end
    else
      SetupTelescope.CheckBoxEast.Checked := false;

    LongSetup:=LongSetup*180/pi;
    SetupTelescope.NbreEditLongDeg.Text := IntToStr(trunc(LongSetup));
    LongSetup:=(LongSetup-int(LongSetup))*60;
    SetupTelescope.NbreEditLongMin.Text := IntToStr(trunc(LongSetup));
    LongSetup:=(LongSetup-int(LongSetup))*60;
    SetupTelescope.NbreEditLongSec.Text := IntToStr(trunc(LongSetup));
    LongSetup:=(LongSetup-int(LongSetup))*100;
    SetupTelescope.NbreEditLongSec100.Text := IntToStr(trunc(LongSetup));

    // Latitude
    If Reg.ValueExists('Latitude') then
      LatSetup := Reg.ReadFloat('Latitude')
    else
    begin
      LatSetup := 0;
      Reg.WriteFloat('Latitude',LatSetup);
    end;

    If (LatSetup<0) then
    begin
      SetupTelescope.CheckBoxSouth.Checked := true;
      LatSetup:=-LatSetup;
    end
    else
      SetupTelescope.CheckBoxSouth.Checked := false;

    LatSetup:=LatSetup*180/pi;
    SetupTelescope.NbreEditLatDeg.Text := IntToStr(trunc(LatSetup));
    LatSetup:=(LatSetup-int(LatSetup))*60;
    SetupTelescope.NbreEditLatMin.Text := IntToStr(trunc(LatSetup));
    LatSetup:=(LatSetup-int(LatSetup))*60;
    SetupTelescope.NbreEditLatSec.Text := IntToStr(trunc(LatSetup));
    LatSetup:=(LatSetup-int(LatSetup))*100;
    SetupTelescope.NbreEditLatSec100.Text := IntToStr(trunc(LatSetup));

    // Altitude

    if Reg.ValueExists('Altitude') then
      AltSetup := Reg.ReadFloat('Altitude')
    else
    begin
      AltSetup := 0;
      Reg.WriteFloat('Altitude',AltSetup);
    end;
    SetupTelescope.NbreEditAltitude.Text:= FloatToStr(AltSetup);

    if Reg.ValueExists('Passage_Interdit')
      then GerMountSetup.Passage_Interdit:=Reg.ReadBool('Passage_Interdit')
    else
    begin
      GerMountSetup.Passage_Interdit :=ReadBool('Pointage', 'Passage_Interdit', false);
      Reg.WriteBool('Passage_Interdit',GerMountSetup.Passage_Interdit);
    end;
    if Reg.ValueExists('External_PEC')
      then ExternalMCMT_PEC.Activated:=Reg.ReadBool('External_PEC')
    else
    begin
      ExternalMCMT_PEC.Activated:=ReadBool('Parameters', 'External_PEC', false);
      Reg.WriteBool('External_PEC',ExternalMCMT_PEC.Activated)
    end;
    SetupTelescope.CheckBoxActivatePEC.Checked:=ExternalMCMT_PEC.Activated;
    if Reg.ValueExists('Periode0') then
      SetupTelescope.NbreEditPeriode_PEC_0.Text := Format('%1.2f',[Reg.ReadFloat('Periode0')])
    else
    begin
      SetupTelescope.NbreEditPeriode_PEC_0.Text := ReadString('Parameters', 'Periode0', '0');
      Reg.WriteFloat('Periode0',StrToFloat(SetupTelescope.NbreEditPeriode_PEC_0.Text));
    end;
    if Reg.ValueExists('Periode1') then
      SetupTelescope.NbreEditPeriode_PEC_1.Text := Format('%1.2f',[Reg.ReadFloat('Periode1')])
    else
    begin
      SetupTelescope.NbreEditPeriode_PEC_1.Text := ReadString('Parameters', 'Periode1', '0');
      Reg.WriteFloat('Periode1',StrToFloat(SetupTelescope.NbreEditPeriode_PEC_1.Text));
    end;
    if Reg.ValueExists('Periode2') then
      SetupTelescope.NbreEditPeriode_PEC_2.Text := Format('%1.2f',[Reg.ReadFloat('Periode2')])
    else
    begin
      SetupTelescope.NbreEditPeriode_PEC_2.Text := ReadString('Parameters', 'Periode2', '0');
      Reg.WriteFloat('Periode2',StrToFloat(SetupTelescope.NbreEditPeriode_PEC_2.Text));
     end;
    if Reg.ValueExists('Periode3') then
      SetupTelescope.NbreEditPeriode_PEC_3.Text := Format('%1.2f',[Reg.ReadFloat('Periode3')])
    else
    begin
      SetupTelescope.NbreEditPeriode_PEC_3.Text := ReadString('Parameters', 'Periode3', '0');
      Reg.WriteFloat('Periode3',StrToFloat(SetupTelescope.NbreEditPeriode_PEC_3.Text));
    end;
    if Reg.ValueExists('Phase0') then
      SetupTelescope.NbreEditPhase_PEC_0.Text := Format('%1.2f',[Reg.ReadFloat('Phase0')])
    else
    begin
      SetupTelescope.NbreEditPhase_PEC_0.Text := ReadString('Parameters', 'Phase0', '0');
      Reg.WriteFloat('Phase0',StrToFloat(SetupTelescope.NbreEditPhase_PEC_0.Text));
    end;
    if Reg.ValueExists('Phase1') then
      SetupTelescope.NbreEditPhase_PEC_1.Text := Format('%1.2f',[Reg.ReadFloat('Phase1')])
    else
    begin
      SetupTelescope.NbreEditPhase_PEC_1.Text := ReadString('Parameters', 'Phase1', '0');
      Reg.WriteFloat('Phase1',StrToFloat(SetupTelescope.NbreEditPhase_PEC_1.Text));
    end;
    if Reg.ValueExists('Phase2') then
      SetupTelescope.NbreEditPhase_PEC_2.Text := Format('%1.2f',[Reg.ReadFloat('Phase2')])
    else
    begin
      SetupTelescope.NbreEditPhase_PEC_2.Text := ReadString('Parameters', 'Phase2', '0');
      Reg.WriteFloat('Phase2',StrToFloat(SetupTelescope.NbreEditPhase_PEC_2.Text));
    end;
    if Reg.ValueExists('Phase3') then
      SetupTelescope.NbreEditPhase_PEC_3.Text := Format('%1.2f',[Reg.ReadFloat('Phase3')])
    else
    begin
      SetupTelescope.NbreEditPhase_PEC_3.Text := ReadString('Parameters', 'Phase3', '0');
      Reg.WriteFloat('Phase3',StrToFloat(SetupTelescope.NbreEditPhase_PEC_3.Text));
    end;
    if Reg.ValueExists('Amplitude0') then
      SetupTelescope.NbreEditAmplitude_PEC_0.Text := Format('%1.2f',[Reg.ReadFloat('Amplitude0')])
    else
    begin
      SetupTelescope.NbreEditAmplitude_PEC_0.Text := ReadString('Parameters', 'Amplitude0', '0');
      Reg.WriteFloat('Amplitude0',StrToFloat(SetupTelescope.NbreEditAmplitude_PEC_0.Text));
    end;
    if Reg.ValueExists('Amplitude1') then
      SetupTelescope.NbreEditAmplitude_PEC_1.Text := Format('%1.2f',[Reg.ReadFloat('Amplitude1')])
    else
    begin
      SetupTelescope.NbreEditAmplitude_PEC_1.Text := ReadString('Parameters', 'Amplitude1', '0');
      Reg.WriteFloat('Amplitude1',StrToFloat(SetupTelescope.NbreEditAmplitude_PEC_1.Text));
    end;
    if Reg.ValueExists('Amplitude2') then
      SetupTelescope.NbreEditAmplitude_PEC_2.Text := Format('%1.2f',[Reg.ReadFloat('Amplitude2')])
    else
    begin
      SetupTelescope.NbreEditAmplitude_PEC_2.Text := ReadString('Parameters', 'Amplitude2', '0');
      Reg.WriteFloat('Amplitude2',StrToFloat(SetupTelescope.NbreEditAmplitude_PEC_2.Text));
    end;
    if Reg.ValueExists('Amplitude3') then
      SetupTelescope.NbreEditAmplitude_PEC_3.Text := Format('%1.2f',[Reg.ReadFloat('Amplitude3')])
    else
    begin
      SetupTelescope.NbreEditAmplitude_PEC_3.Text := ReadString('Parameters', 'Amplitude3', '0');
      Reg.WriteFloat('Amplitude3',StrToFloat(SetupTelescope.NbreEditAmplitude_PEC_3.Text));
    end;
    if Reg.ValueExists('King')
      then KING_Activated:=Reg.ReadBool('King')
    else
    begin
      KING_Activated:=ReadBool('Parameters', 'King', false);
      Reg.WriteBool('King',KING_Activated)
    end;
    SetupTelescope.CheckBoxActivateKing.Checked:=KING_Activated;

    if SetupTelescope.CheckBoxActivatePEC.Checked then
    begin
      SetupTelescope.label62.Enabled := true;
      SetupTelescope.label63.Enabled := true;
      SetupTelescope.label64.Enabled := true;
      SetupTelescope.label65.Enabled := true;
      SetupTelescope.label66.Enabled := true;
      SetupTelescope.label67.Enabled := true;
      SetupTelescope.label68.Enabled := true;
      SetupTelescope.NbreEditAmplitude_PEC_0.Enabled := true;
      SetupTelescope.NbreEditAmplitude_PEC_1.Enabled := true;
      SetupTelescope.NbreEditAmplitude_PEC_2.Enabled := true;
      SetupTelescope.NbreEditAmplitude_PEC_3.Enabled := true;
      SetupTelescope.NbreEditPeriode_PEC_0.Enabled := true;
      SetupTelescope.NbreEditPeriode_PEC_1.Enabled := true;
      SetupTelescope.NbreEditPeriode_PEC_2.Enabled := true;
      SetupTelescope.NbreEditPeriode_PEC_3.Enabled := true;
      SetupTelescope.NbreEditPhase_PEC_0.Enabled := true;
      SetupTelescope.NbreEditPhase_PEC_1.Enabled := true;
      SetupTelescope.NbreEditPhase_PEC_2.Enabled := true;
      SetupTelescope.NbreEditPhase_PEC_3.Enabled := true;
    end
    else
    begin
      SetupTelescope.label62.Enabled := false;
      SetupTelescope.label63.Enabled := false;
      SetupTelescope.label64.Enabled := false;
      SetupTelescope.label65.Enabled := false;
      SetupTelescope.label66.Enabled := false;
      SetupTelescope.label67.Enabled := false;
      SetupTelescope.label68.Enabled := false;
      SetupTelescope.NbreEditAmplitude_PEC_0.Enabled := false;
      SetupTelescope.NbreEditAmplitude_PEC_1.Enabled := false;
      SetupTelescope.NbreEditAmplitude_PEC_2.Enabled := false;
      SetupTelescope.NbreEditAmplitude_PEC_3.Enabled := false;
      SetupTelescope.NbreEditPeriode_PEC_0.Enabled := false;
      SetupTelescope.NbreEditPeriode_PEC_1.Enabled := false;
      SetupTelescope.NbreEditPeriode_PEC_2.Enabled := false;
      SetupTelescope.NbreEditPeriode_PEC_3.Enabled := false;
      SetupTelescope.NbreEditPhase_PEC_0.Enabled := false;
      SetupTelescope.NbreEditPhase_PEC_1.Enabled := false;
      SetupTelescope.NbreEditPhase_PEC_2.Enabled := false;
      SetupTelescope.NbreEditPhase_PEC_3.Enabled := false;
    end;
  end;
  Reg.CloseKey;
  Finally
    DeviceIni.free;
    Reg.Free;
  End;

  SetupTelescope.CheckBoxAutoCOM.Checked:=AutoFindCOM;

  // Mettre a jour avec la vraie config contenue dans MCMT II
  If COM_OK then
    SetupTelescope.Button_LireClick(nil);

  SetupTelescope.creating:=False;
  Form2.TimerCoord.Enabled:=false;
  ///////////////  ///////////////  ///////////////  ///////////////  ///////////////  ///////////////
  if (SetupTelescope.ShowModal = MrOK) then // Uniquement si *Validation* !
  ///////////////  ///////////////  ///////////////  ///////////////  ///////////////  ///////////////
  begin
    DeviceIni   := TIniFile.Create(return_MCMT_StoreIniFilePath);
    Reg         := TRegistry.Create;
    Try
      Reg.RootKey := HKEY_CURRENT_USER;
      if not (Reg.OpenKey('SOFTWARE\'+Version, False)) then Reg.OpenKey('SOFTWARE\'+Version, true);
      with DeviceIni do
      begin
        WriteBool('Parameters','External_PEC',SetupTelescope.CheckBoxActivatePEC.Checked);
        ExternalMCMT_PEC.Activated:=SetupTelescope.CheckBoxActivatePEC.Checked;
        Reg.WriteBool('External_PEC',ExternalMCMT_PEC.Activated);
        WriteString('Parameters', 'Periode0',SetupTelescope.NbreEditPeriode_PEC_0.text);
        ExternalMCMT_PEC.Periode[0]:=strtofloat(SetupTelescope.NbreEditPeriode_PEC_0.text);
        Reg.WriteFloat('Periode0',ExternalMCMT_PEC.Periode[0]);
        WriteString('Parameters', 'Periode1',SetupTelescope.NbreEditPeriode_PEC_1.text);
        ExternalMCMT_PEC.Periode[1]:=strtofloat(SetupTelescope.NbreEditPeriode_PEC_1.text);
        Reg.WriteFloat('Periode1',ExternalMCMT_PEC.Periode[1]);
        WriteString('Parameters', 'Periode2',SetupTelescope.NbreEditPeriode_PEC_2.text);
        ExternalMCMT_PEC.Periode[2]:=strtofloat(SetupTelescope.NbreEditPeriode_PEC_2.text);
        Reg.WriteFloat('Periode2',ExternalMCMT_PEC.Periode[2]);
        WriteString('Parameters', 'Periode3',SetupTelescope.NbreEditPeriode_PEC_3.text);
        ExternalMCMT_PEC.Periode[3]:=strtofloat(SetupTelescope.NbreEditPeriode_PEC_3.text);
        Reg.WriteFloat('Periode3',ExternalMCMT_PEC.Periode[3]);
        WriteString('Parameters', 'Phase0',SetupTelescope.NbreEditPhase_PEC_0.text);
        ExternalMCMT_PEC.Phase[0]:=strtofloat(SetupTelescope.NbreEditPhase_PEC_0.text);
        Reg.WriteFloat('Phase0',ExternalMCMT_PEC.Phase[0]);
        WriteString('Parameters', 'Phase1',SetupTelescope.NbreEditPhase_PEC_1.text);
        ExternalMCMT_PEC.Phase[1]:=strtofloat(SetupTelescope.NbreEditPhase_PEC_1.text);
        Reg.WriteFloat('Phase1',ExternalMCMT_PEC.Phase[1]);
        WriteString('Parameters', 'Phase2',SetupTelescope.NbreEditPhase_PEC_2.text);
        ExternalMCMT_PEC.Phase[2]:=strtofloat(SetupTelescope.NbreEditPhase_PEC_2.text);
        Reg.WriteFloat('Phase2',ExternalMCMT_PEC.Phase[2]);
        WriteString('Parameters', 'Phase3',SetupTelescope.NbreEditPhase_PEC_3.text);
        ExternalMCMT_PEC.Phase[3]:=strtofloat(SetupTelescope.NbreEditPhase_PEC_3.text);
        Reg.WriteFloat('Phase3',ExternalMCMT_PEC.Phase[3]);
        WriteString('Parameters', 'Amplitude0',SetupTelescope.NbreEditAmplitude_PEC_0.text);
        ExternalMCMT_PEC.Amplitude[0]:=strtofloat(SetupTelescope.NbreEditAmplitude_PEC_0.text);
        Reg.WriteFloat('Amplitude0',ExternalMCMT_PEC.Amplitude[0]);
        WriteString('Parameters', 'Amplitude1',SetupTelescope.NbreEditAmplitude_PEC_1.text);
        ExternalMCMT_PEC.Amplitude[1]:=strtofloat(SetupTelescope.NbreEditAmplitude_PEC_1.text);
        Reg.WriteFloat('Amplitude1',ExternalMCMT_PEC.Amplitude[1]);
        WriteString('Parameters', 'Amplitude2',SetupTelescope.NbreEditAmplitude_PEC_2.text);
        ExternalMCMT_PEC.Amplitude[2]:=strtofloat(SetupTelescope.NbreEditAmplitude_PEC_2.text);
        Reg.WriteFloat('Amplitude2',ExternalMCMT_PEC.Amplitude[2]);
        WriteString('Parameters', 'Amplitude3',SetupTelescope.NbreEditAmplitude_PEC_3.text);
        ExternalMCMT_PEC.Amplitude[3]:=strtofloat(SetupTelescope.NbreEditAmplitude_PEC_3.text);
        Reg.WriteFloat('Amplitude3',ExternalMCMT_PEC.Amplitude[3]);
        WriteBool('Parameters','King',SetupTelescope.CheckBoxActivateKing.Checked);
        KING_Activated:=SetupTelescope.CheckBoxActivateKing.Checked;
        Reg.WriteBool('King',KING_Activated);

        AutoFindCOM:= SetupTelescope.CheckBoxAutoCOM.Checked;
        if AutoFindCOM then
          NumCom     := StrToInt(SetupTelescope.ComboBoxNumCOM.text)
        else
          NumCom     := StrToInt(SetupTelescope.NbreEditPortSerieForced.text);

        WriteInteger('Configuration', 'Port', NumCOM);
        Reg.WriteInteger('Port'             , NumCOM);
        Reg.WriteBool   ('FindAutoPort'     , AutoFindCOM);

        WriteString('Configuration', 'FirmwareFile', SetupTelescope.OpenDialog1.filename);
        Reg.WriteString('FirmwareFile', SetupTelescope.OpenDialog1.filename);

        WriteBool('Parameters', 'DIR_AD', SetupTelescope.CheckBox_DIR_AD.Checked);
        Reg.WriteBool('DIR_AD', SetupTelescope.CheckBox_DIR_AD.Checked);

        WriteBool('Parameters', 'DIR_DEC', SetupTelescope.CheckBox_DIR_DEC.checked);
        Reg.WriteBool('DIR_DEC', SetupTelescope.CheckBox_DIR_DEC.checked);

        WriteBool('Parameters', 'PEC_AD', SetupTelescope.CheckBox_PEC.checked);
        Reg.WriteBool('PEC_AD', SetupTelescope.CheckBox_PEC.checked);

        WriteBool('Parameters', 'Raq_Can',(SetupTelescope.RadioGroup_Raquette.ItemIndex=1));
        Reg.WriteBool('Raq_Can',          (SetupTelescope.RadioGroup_Raquette.ItemIndex=1));

        SpeedIntMCMT.Inv_Codeur_AD := SetupTelescope.CheckBox_DIR_AD.checked;
        SpeedIntMCMT.Inv_Codeur_DEC := SetupTelescope.CheckBox_DIR_DEC.checked;

        SpeedIntMCMT.Resol_AD := strtoint(SetupTelescope.Edit_Resol_AD.text);
        WriteInteger('Parameters', 'Resol_AD', SpeedIntMCMT.Resol_AD);
        Reg.WriteInteger('Resol_AD', SpeedIntMCMT.Resol_AD);

        SpeedIntMCMT.Resol_DEC := strtoint(SetupTelescope.Edit_Resol_DEC.text);
        WriteInteger('Parameters', 'Resol_DEC', SpeedIntMCMT.Resol_DEC);
        Reg.WriteInteger('Resol_DEC', SpeedIntMCMT.Resol_DEC);

        WriteString('Parameters', 'ReducAlpha', SetupTelescope.EditResolAlpha.text);
        Reg.WriteFloat('ReducAlpha', StrToFloat(SetupTelescope.EditResolAlpha.text));

        WriteString('Parameters', 'ReducDelta', SetupTelescope.EditResolDelta.text);
        Reg.WriteFloat('ReducDelta', StrToFloat(SetupTelescope.EditResolDelta.text));

        WriteString('Parameters', 'V_Guidage_AD', SetupTelescope.Edit_V_Guidage_AD.text);
        SpeedIntMCMT.V_Guidage_AD := strtofloat(SetupTelescope.Edit_V_Guidage_AD.text);
        Reg.WriteFloat('V_Guidage_AD', SpeedIntMCMT.V_Guidage_AD);
        SpeedIntMCMT.V_Guidage_AD_SID := SpeedIntMCMT.V_Guidage_AD;

        WriteString('Parameters', 'V_Guidage_DEC', SetupTelescope.Edit_V_Guidage_DEC.text);
        SpeedIntMCMT.V_Guidage_DEC := strtofloat(SetupTelescope.Edit_V_Guidage_DEC.text);
        Reg.WriteFloat('V_Guidage_DEC', SpeedIntMCMT.V_Guidage_DEC);

        WriteString('Parameters', 'V_Corr_P_AD', SetupTelescope.Edit_V_Corr_P_AD.text);
        SpeedIntMCMT.V_Corr_P_AD := strtofloat(SetupTelescope.Edit_V_Corr_P_AD.text);
        Reg.WriteFloat('V_Corr_P_AD', SpeedIntMCMT.V_Corr_P_AD);

        WriteString('Parameters', 'V_Corr_P_DEC', SetupTelescope.Edit_V_Corr_P_DEC.text);
        SpeedIntMCMT.V_Corr_P_DEC := strtofloat(SetupTelescope.Edit_V_Corr_P_DEC.text);
        Reg.WriteFloat('V_Corr_P_DEC', SpeedIntMCMT.V_Corr_P_DEC);

        WriteString('Parameters', 'V_Corr_M_AD', SetupTelescope.Edit_V_Corr_M_AD.text);
        SpeedIntMCMT.V_Corr_M_AD := strtofloat(SetupTelescope.Edit_V_Corr_M_AD.text);
        Reg.WriteFloat('V_Corr_M_AD', SpeedIntMCMT.V_Corr_M_AD);

        WriteString('Parameters', 'V_Corr_M_DEC', SetupTelescope.Edit_V_Corr_M_DEC.text);
        SpeedIntMCMT.V_Corr_M_DEC := strtofloat(SetupTelescope.Edit_V_Corr_M_DEC.text);
        Reg.WriteFloat('V_Corr_M_DEC', SpeedIntMCMT.V_Corr_M_DEC);

        WriteString('Parameters', 'V_Point_L_AD', SetupTelescope.Edit_V_Point_L_AD.text);
        SpeedIntMCMT.V_Point_L_AD := strtofloat(SetupTelescope.Edit_V_Point_L_AD.text);
        Reg.WriteFloat('V_Point_L_AD', SpeedIntMCMT.V_Point_L_AD);

        WriteString('Parameters', 'V_Point_L_DEC', SetupTelescope.Edit_V_Point_L_DEC.text);
        SpeedIntMCMT.V_Point_L_DEC := strtofloat(SetupTelescope.Edit_V_Point_L_DEC.text);
        Reg.WriteFloat('V_Point_L_DEC', SpeedIntMCMT.V_Point_L_DEC);

        WriteString('Parameters', 'V_Point_R_AD', SetupTelescope.Edit_V_Point_R_AD.text);
        SpeedIntMCMT.V_Point_R_AD := strtofloat(SetupTelescope.Edit_V_Point_R_AD.text);
        Reg.WriteFloat('V_Point_R_AD', SpeedIntMCMT.V_Point_R_AD);

        WriteString('Parameters', 'V_Point_R_DEC', SetupTelescope.Edit_V_Point_R_DEC.text);
        SpeedIntMCMT.V_Point_R_DEC := strtofloat(SetupTelescope.Edit_V_Point_R_DEC.text);
        Reg.WriteFloat('V_Point_R_DEC', SpeedIntMCMT.V_Point_R_DEC);

        WriteString('Parameters', 'Accel_AD', inttostr(SetupTelescope.ComboBox_Accel_AD.ItemIndex));
        SpeedIntMCMT.Accel_AD := SetupTelescope.ComboBox_Accel_AD.ItemIndex;
        Reg.WriteInteger('Accel_AD', SpeedIntMCMT.Accel_AD);

        WriteString('Parameters', 'Accel_DEC', inttostr(SetupTelescope.ComboBox_Accel_DEC.ItemIndex));
        SpeedIntMCMT.Accel_DEC := SetupTelescope.ComboBox_Accel_DEC.ItemIndex;
        Reg.WriteInteger('Accel_DEC', SpeedIntMCMT.Accel_DEC);

        WriteBool('Parameters', 'Backlash_Activated', SetupTelescope.CheckBoxActivateBacklash.Checked);
        Backlash.Activated := SetupTelescope.CheckBoxActivateBacklash.Checked;
        Reg.WriteBool('Backlash_Activated', Backlash.Activated);

        WriteString('Parameters', 'Backlash', SetupTelescope.NbreEditBacklash.Text);
        Backlash.Value := StrToFloat(SetupTelescope.NbreEditBacklash.Text);
        Reg.WriteFloat('Backlash', Backlash.Value);

        WriteInteger('Parameters', 'Backlash_Delta_Type',SetupTelescope.RadioGroup_Mode_Correction_Delta.ItemIndex);         //JLH
        Backlash.Delta_Mode:=SetupTelescope.RadioGroup_Mode_Correction_Delta.ItemIndex;
        Reg.WriteInteger('Backlash_Delta_Mode', Backlash.Delta_Mode);

        LogEnabled:=SetupTelescope.CheckBoxLogFile.Checked;
        Reg.WriteBool('LogEnabled',LogEnabled);

        TraceEnabled:=SetupTelescope.CheckBoxTrace.Checked;
        Reg.WriteBool('TraceEnabled',TraceEnabled);
{$IFDEF MCMT32}
        FreeSpeed:=SetupTelescope.CheckBoxFreeSpeed.Checked;
        Reg.WriteBool('FreeSpeed',FreeSpeed);
        SimulatorEnabled:=SetupTelescope.CheckBoxSimulator.Checked;
        Reg.WriteBool('SimulatorEnabled',SimulatorEnabled);
{$ELSE}
{$ENDIF}

        WriteInteger('Parameters', 'Backlash_Pause', SetupTelescope.ComboBoxPauseBacklash.ItemIndex);
        Backlash.Pause := SetupTelescope.ComboBoxPauseBacklash.ItemIndex * 1000;
        Reg.WriteInteger('Backlash_Pause', Backlash.Pause);

        // Type de monture ***************************************************
        if SetupTelescope.RadioGroupTypeMonture.ItemIndex=0
          then GerMountSetup.Monture_Allemande:= false
          else GerMountSetup.Monture_Allemande:= true;
        WriteBool('Pointage', 'Monture_Allemande', GerMountSetup.Monture_Allemande);
        Reg.WriteBool('Monture_Allemande', GerMountSetup.Monture_Allemande);


        // Monture  fourche***************************************************
        if SetupTelescope.RadioGroupAlgorithmePointage.ItemIndex=0
          then GerMountSetup.Passage_Interdit:=false
          else GerMountSetup.Passage_Interdit:=true;
        Reg.WriteBool('Passage_Interdit',GerMountSetup.Passage_Interdit);
        WriteBool('Pointage', 'Passage_Interdit', GerMountSetup.Passage_Interdit);


        // Monture allemande ***************************************************
        GerMountSetup.Angle_Horaire_Critique_Est:=SetupTelescope.SpinEdit_Angle_Horaire_Critique_Est.value*Pi/12/60;
        WriteFloat('Pointage', 'Angle_Horaire_Critique_Est',GerMountSetup.Angle_Horaire_Critique_Est);
        Reg.WriteFloat('Angle_Horaire_Critique_Est', GerMountSetup.Angle_Horaire_Critique_Est);

        GerMountSetup.Angle_Horaire_Critique_Ouest:=SetupTelescope.SpinEdit_Angle_Horaire_Critique_Ouest.value*Pi/720;
        WriteFloat('Pointage', 'Angle_Horaire_Critique_Ouest',GerMountSetup.Angle_Horaire_Critique_Ouest);
        Reg.WriteFloat('Angle_Horaire_Critique_Ouest', GerMountSetup.Angle_Horaire_Critique_Ouest);

        GerMountSetup.Angle_horaire_Retournement:=SetupTelescope.SpinEdit_Angle_horaire_Retournement.value*Pi/720;
        WriteFloat('Pointage', 'Angle_horaire_Retournement',GerMountSetup.Angle_horaire_Retournement);
        Reg.WriteFloat('Angle_horaire_Retournement', GerMountSetup.Angle_horaire_Retournement);

        GerMountSetup.Rayon_minimal:=SetupTelescope.SpinEdit_Rayon_minimal.value*Pi/180;
        WriteFloat('Pointage', 'Rayon_minimal',GerMountSetup.Rayon_minimal);
        Reg.WriteFloat('Rayon_minimal', GerMountSetup.Rayon_minimal);

        GerMountSetup.Declinaison_critique_Est:=SetupTelescope.SpinEdit_Declinaison_critique_Est.value*Pi/180;
        WriteFloat('Pointage', 'Declinaison_critique_Est',GerMountSetup.Declinaison_critique_Est);
        Reg.WriteFloat('Declinaison_critique_Est', GerMountSetup.Declinaison_critique_Est);

        GerMountSetup.Declinaison_critique_Ouest:=SetupTelescope.SpinEdit_Declinaison_critique_Ouest.value*Pi/180;
        WriteFloat('Pointage', 'Declinaison_critique_Ouest',GerMountSetup.Declinaison_critique_Ouest);
        Reg.WriteFloat('Declinaison_critique_Ouest', GerMountSetup.Declinaison_critique_Ouest);

        GerMountSetup.Retournement_autorise:=(SetupTelescope.RadioGroup_retournement.ItemIndex=0);
        WriteBool('Pointage', 'Retournement_autorise',GerMountSetup.Retournement_autorise);
        Reg.WriteBool('Retournement_autorise', GerMountSetup.Retournement_autorise);

        GerMountSetup.TelescopeZenithOuest:=(SetupTelescope.RadioGroupEtatTelescope.ItemIndex=1);
        Reg.WriteBool('TelescopeZenithOuest',GerMountSetup.TelescopeZenithOuest);

        King_Activated:= SetupTelescope.CheckBoxActivateKing.Checked;
        PEC_Activated:= SetupTelescope.CheckBoxActivatePEC.Checked;

        LongSetup:=StrToFloat(SetupTelescope.NbreEditLongSec100.Text)/100;
        LongSetup:=(LongSetup+StrToFloat(SetupTelescope.NbreEditLongSec.Text))/60;
        LongSetup:=(LongSetup+StrToFloat(SetupTelescope.NbreEditLongMin.Text))/60;
        LongSetup:=(LongSetup+StrToFloat(SetupTelescope.NbreEditLongDeg.Text))/180*pi;
        if SetupTelescope.CheckBoxEast.Checked = true then LongSetup:=-LongSetup;
        Reg.WriteFloat('Longitude', LongSetup);

        LatSetup:=StrToFloat(SetupTelescope.NbreEditLatSec100.Text)/100;
        LatSetup:=(LatSetup+StrToFloat(SetupTelescope.NbreEditLatSec.Text))/60;
        LatSetup:=(LatSetup+StrToFloat(SetupTelescope.NbreEditLatMin.Text))/60;
        LatSetup:=(LatSetup+StrToFloat(SetupTelescope.NbreEditLatDeg.Text))/180*pi;
        if SetupTelescope.CheckBoxSouth.Checked = true then LatSetup:=-LatSetup;
        Reg.WriteFloat('Latitude', LatSetup);

        Reg.WriteFloat('Altitude', StrToFloat(SetupTelescope.NbreEditAltitude.Text));

        TimeOutMs_ConnectionsWrite:=StrToInt(SetupTelescope.NumberEditWriteTimeOutms.Text);
        Reg.WriteInteger('TimeOutMs_ConnectionsWrite',TimeOutMs_ConnectionsWrite);

        TimeOutMs_ConnectionsRead:=StrToInt(SetupTelescope.NumberEditReadTimeoutMs.Text  );
        Reg.WriteInteger('TimeOutMs_ConnectionsRead',TimeOutMs_ConnectionsRead);

        PurgeCommMethod:=SetupTelescope.CheckBoxPurgCOMM.Checked;
        Reg.WriteBool('PurgeCommMethod',PurgeCommMethod);
      end;
      Reg.CloseKey;
    Finally
      DeviceIni.free;
      Reg.Free;
    End;
  end;
  Form2.TimerCoord.Enabled:=true;
  SetupTelescope.Free; //Destroy...
  Form_Etat_Telescope.Free; //Destroy...
  LogEndProc(procName);
end;

procedure SlewToCoordinates_Core(RightAscension, Declination: Double);
Const
  procName='SlewToCoordinates_Core';
begin
  LogEnterProc(procName);
  if fParked then
  begin
    WriteToTrace('SlewToCoordinates error telescope is parked');
    raise EoleSysError.Create('Telescope is parked', $80040401, 0)
  end
  else
    if not (fTracking) then
    begin
      WriteToTrace('SlewToCoordinates error telescope is not tracking');
      raise EoleSysError.Create('Telescope is not tracking', $80040401, 0)
    end
    else
    begin
      WriteToTrace('SlewToCoordinates '+FloatToStr(RightAscension)+' '+FloatToStr(Declination));
      SlewToCoordinatesAsync_Core(RightAscension, Declination);
      while Get_Slewing_Core do ;
    end;
  LogEndProc(procName);
end;

procedure SlewToCoordinatesAsync_Core(RightAscension, Declination: Double);
Const
  procName='SlewToCoordinatesAsync_Core';
var
  Alpha, Delta: double;
begin
  LogEnterProc(procName);
  if fParked then
  begin
    WriteToTrace('SlewToCoordinatesAsync error telescope is parked');
    raise EoleSysError.Create('Telescope is parked', $80040401, 0)
  end
  else
    if not (fTracking) then
    begin
      WriteToTrace('SlewToCoordinatesAsync error telescope is not tracking');
      raise EoleSysError.Create('Telescope is not tracking', $80040401, 0)
    end
    else
    begin
      if (RightAscension > 24) or (RightAscension < 0) then
      begin
        WriteToTrace('SlewToCoordinatesAsync error right ascension out of [0,24]');
        raise EoleSysError.Create('Right ascension out of [0,24]', $80040401, 0);
      end
      else
      begin
        fTargetRightAscension := RightAscension;
        fTargetAscOK := true;
        Alpha := RightAscension / 12 * PI;
      end;
      if (Declination > 90) or (Declination < -90) then
      begin
        WriteToTrace('SlewToCoordinatesAsync error declination out of [-90,90]');
        raise EoleSysError.Create('Declination out of [-90,90]', $80040401, 0);
      end
      else
      begin
        fTargetDeclination := Declination;
        fTargetDecOK := true;
        Delta := Declination / 180 * PI
      end;
      WriteToTrace('SlewToCoordinatesAsync '+FloatToStr(RightAscension)+' '+FloatToStr(Declination));
      InitTelescopeSeeker(Alpha, Delta,Lieu.Lat,Lieu.Long);
    end;
  LogEndProc(procName);
end;

procedure SlewToTarget_Core;
Const
  procName='SlewToTarget_Core';
begin
  LogEnterProc(procName);
  if fParked then
  begin
    WriteToTrace('SlewToTarget error telescope is parked');
    raise EoleSysError.Create('Telescope is parked', $80040401, 0)
  end
  else
    if not (fTracking) then
    begin
      WriteToTrace('SlewToTarget error telescope is not tracking');
      raise EoleSysError.Create('Telescope is not tracking', $80040401, 0)
    end
    else
    begin
      WriteToTrace('SlewToTarget');
      SlewToTargetAsync_Core;
      while Get_Slewing_Core do ;
    end;
  LogEndProc(procName);
end;

procedure SlewToTargetAsync_Core;
Const
  procName='SlewToTargetAsync_Core';
var
  Alpha, Delta: double;
begin
  LogEnterProc(procName);
  if fParked then
  begin
    WriteToTrace('SlewToTargetAsync error telescope is parked');
    raise EoleSysError.Create('Telescope is parked', $80040401, 0)
  end
  else
    if not (fTracking) then
    begin
      WriteToTrace('SlewToTargetAsync error telescope is not tracking');
      raise EoleSysError.Create('Telescope is not tracking', $80040401, 0)
    end
    else
    begin
      Alpha := fTargetRightAscension / 12 * PI;
      Delta := fTargetDeclination / 180 * PI;
      WriteToTrace('SlewToTargetAsync');
      InitTelescopeSeeker(Alpha, Delta,Lieu.Lat,Lieu.Long);
    end;
  LogEndProc(procName);
end;

procedure SyncToCoordinates_Core(RightAscension, Declination: Double);
Const
  procName='SyncToCoordinates_Core';
begin
  LogEnterProc(procName);
  if fParked then
  begin
    WriteToTrace('SyncToCoordinates error telescope is parked');
    raise EoleSysError.Create('Telescope is parked', $80040401, 0)
  end
  else
    if not (fTracking) then
    begin
      WriteToTrace('SyncToCoordinates error telescope is not tracking');
      raise EoleSysError.Create('Telescope is not tracking', $80040401, 0)
    end
    else
    begin
      if (Declination > 90) or (Declination < -90) or (RightAscension > 24) or (RightAscension < 0) then
      begin
        WriteToTrace('SyncToCoordinates error coordinates not available');
        raise EoleSysError.Create('Coordinates not available', $80040401, 0);
      end;
      fDeclination := Declination;
      fRightAscension := RightAscension;
      fTargetDeclination := Declination;
      fTargetRightAscension := RightAscension;
      if not (Calibrate(RightAscension / 12 * PI, Declination / 180 * PI)) then
      begin
        WriteToTrace('SyncToCoordinates error sync impossible');
        raise EoleSysError.Create('Sync impossible', $80040401, 0);
      end;
      WriteToTrace('SyncToCoordinates '+FloatToStr(RightAscension)+' '+FloatToStr(Declination));
    end;
  LogEndProc(procName);
end;

procedure SyncToTarget_Core;
Const
  procName='SyncToTarget_Core';
begin
  LogEnterProc(procName);
  WriteToTrace('Sync To Target');
  if fParked then
  begin
    WriteToTrace('SyncToTarget error telescope is parked');
    raise EoleSysError.Create('Telescope is parked', $80040401, 0)
  end
  else
    if not (fTracking) then
    begin
      WriteToTrace('SyncToTarget error telescope is not tracking');
      raise EoleSysError.Create('Telescope is not tracking', $80040401, 0)
    end
    else
    begin
      fDeclination := fTargetDeclination;
      fRightAscension := fTargetRightAscension;
      if not (Calibrate(fTargetRightAscension / 12 * PI, fTargetDeclination / 180 * PI)) then
      begin
        WriteToTrace('SyncToTarget error sync impossible');
        raise EoleSysError.Create('Sync impossible', $80040401, 0);
      end;
      WriteToTrace('SyncToTarget');
    end;
  LogEndProc(procName);
end;

procedure Unpark_Core;
Const
  procName='Unpark_Core';
var
  astring   : shortstring;
  Reg       : TRegistry;
  Alpha_Temp,
  Decli_Temp: double;

begin
  LogEnterProc(procName);
  WriteToTrace('Unpark');
  if fParked then
  begin
    Reg := TRegistry.Create;
    Reg.RootKey := HKEY_CURRENT_USER;
    if not (Reg.OpenKey('SOFTWARE\'+Version, False)) then Reg.OpenKey('SOFTWARE\'+Version, True);
    if Reg.ValueExists('Park_AH') then
      Alpha_Temp := Reg.ReadFloat('Park_AH')
    else Alpha_Temp := 0;

    if Reg.ValueExists('Park_Dec') then
      Decli_Temp := Reg.ReadFloat('Park_Dec')
    else Decli_Temp := 0;
    Reg.CloseKey;
    Reg.Free;
    Try
      astring := 'N' + AnsiChar(0) + AnsiChar(0) + AnsiChar(0) + AnsiChar(0); // Commande No Park
      sendcommand(astring, 0, 0);
    finally
    end;
    BackLash.Activated := BackLash.Saved_Park; // restore backlash state
    Alpha_Temp := CalculeTSL(Now+SpeedIntMCMT.GMT/24,Lieu.Long)-Alpha_Temp;
    Remet2Pi(Alpha_Temp);
    Calibrate(Alpha_Temp, Decli_Temp);
    fParked := false;
    fTracking := true;
  end;
  LogEndProc(procName);
end;

end.
